import uuid
from functools import partial
from typing import Any, Dict, List, Optional

import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import auth_policy_id_arg, globus_command
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.auth_client import MissingConsent
from gcs_cli.services.auth import get_client


def format_domains(key: str, policy: Dict[str, Any]) -> str:
    domains: Optional[List[str]] = policy[f"domain_constraints_{key}"]
    if not domains:
        return "None"
    return ", ".join(domains)


@globus_command(
    "show",
    short_help="Display the authentication policy",
    disable_options=["explicit_host"],
)
@requires_login
@auth_policy_id_arg
def auth_policy_show(auth_policy_id: uuid.UUID) -> None:
    """
    Display the authentication policy.
    """
    client = get_client()
    try:
        res = client.get_policy(auth_policy_id)
    except MissingConsent as e:
        raise click.ClickException(
            "Missing consent for the policy API. Please run the following command:\n"
            f"    globus-connect-server session consent {e.scope}"
        )

    formatted_print(
        res["policy"],
        text_format=FORMAT_TEXT_RECORD,
        fields=[
            ("Display Name", "display_name"),
            ("Description", "description"),
            ("ID", "id"),
            ("Project ID", "project_id"),
            ("Include Domains", partial(format_domains, "include")),
            ("Exclude Domains", partial(format_domains, "exclude")),
            ("High Assurance", "high_assurance"),
            ("Timeout", "authentication_assurance_timeout"),
        ],
        non_gcs_response=True,
        list_field_name="id",
    )
