from __future__ import annotations

import re
import typing as t
from datetime import date


def get_last_access(
    obj: t.Dict[str, t.Any], earliest_last_access: date | None = None
) -> str:
    """
    The last_access property was added to collection#1.9.0 but is only set when
    there is a known login attempt against that collection. A null value may mean

    1. The endpoint doesn't track last access
    2. The collection was not accessed since the endpoint was update to support
       tracking last_access. We can say not accessed since <earliest_last_access>
    3. The collection was never accessed

    To resolve the ambiguity, we also check the following:

    1. If collection#1.9.0 or older, we know it's (1)
    2. Endpoint doesn't have earliest_last_access (added in endpoint#1.2.0).
       We don't have any info (1)
    3. Endpoint has earliest_last_access
       A. Collection was created after endpoint's earliest_last_access. In
          this case we know the collection has never been accessed (3).
       B. The collection was created before the endpoint started tracking
          last_access. The best we can say is (2)
    """
    last_access = obj.get("last_access")
    created_at = obj.get("created_at")

    if last_access is None:
        data_type = obj.get("DATA_TYPE", "")
        version_match = re.match(
            r"collection#(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+)",
            data_type,
        )
        if not version_match:
            return "Not supported"
        (major, minor) = (version_match.group("major"), version_match.group("minor"))

        if (
            (int(major) == 1 and int(minor) < 9)
            or not earliest_last_access
            or not created_at
        ):
            return "Not supported"

        created_at_date = date.fromisoformat(created_at)
        if created_at_date > earliest_last_access:
            return "Never"
        else:
            return "*"

    return str(last_access)
