import uuid

import click

from gcs_cli.sdk_client.auth_client import MissingConsent
from gcs_cli.services.auth import AuthClient


def lookup_project(client: AuthClient, option: str) -> uuid.UUID:
    """
    Return the ID of the single project admined by the caller.

    If the admin has no projects, throw ClickException with a description
    on how to create a project.

    If the admin has more than one project, throw ClickException with a
    description of how to use 'option' to specify the project.
    """
    try:
        project_res = client.get_projects()
    except MissingConsent as e:
        raise click.ClickException(
            "Missing consent for the policy API. Please run the following "
            "command:\n"
            f"    globus-connect-server session consent {e.scope}"
        )

    if len(project_res["projects"]) == 0:
        raise click.ClickException(
            "You are not the admin of any projects. Please go to "
            "https://auth.globus.org/developers and create one."
        )
    if len(project_res["projects"]) > 1:
        raise click.ClickException(
            f"You are the admin of multiple projects, please use the {option} "
            "option to specify which project to use. Current projects:\n"
            + "\n".join(
                [f"{d['id']} : {d['project_name']}" for d in project_res["projects"]]
            )
        )
    return uuid.UUID(f"{project_res['projects'][0]['id']}")
