from typing import (
    Any,
    cast,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TYPE_CHECKING,
    Union,
)

import click

from globus_sdk import AuthClient, GlobusError, IdentityMap
from globus_sdk.services.auth.response import GetIdentitiesResponse
from globus_sdk._types import UUIDLike

from ..models import MigrationPlan, MyProxyConfiguration, S3Credential
from ..parsers import parse_identity_mapping
from ..util import get_api, get_unusable_idp_domains
from .decorators import requires_logged_in, requires_migration_plan
from ._fixup_s3_credential import _fixup_s3_credential

if TYPE_CHECKING:
    from ..typing.gcs_documents import IdentityMappingDict  # noqa: F401


class ProvisioningAuthClient(AuthClient):
    """Overload get_identities to default to True"""

    def get_identities(
        self,
        *,
        usernames: Union[Iterable[str], str, None] = None,
        ids: Union[Iterable[UUIDLike], UUIDLike, None] = None,
        provision: bool = True,
        query_params: Optional[Dict[str, Any]] = None,
    ) -> GetIdentitiesResponse:
        return super().get_identities(
            usernames=usernames, ids=ids, provision=provision, query_params=query_params
        )


@click.command("myproxy", help="Configure Authorization for MyProxy")
@click.option(
    "--subject",
    "-s",
    metavar="SUBJECT",
    help="MyProxy CA Subject. Optional if only one MyProxy CA is configured for "
    "your endpoint.",
)
@click.option(
    "--domain",
    "-d",
    multiple=True,
    required=True,
    metavar="DOMAIN",
    help="Domains to allow access to the endpoint.",
)
@click.option(
    "--identity-mapping",
    "-i",
    multiple=True,
    metavar="JSON_OR_FILE",
    help="Either a JSON object or the name of a file that contains a json object. "
    "This object is parsed as an identity mapping document and added to the "
    "endpoint's policies. This is optional if there is only one domain being added, "
    "otherwise one must be defined to disambiguate how users are mapped.",
)
@click.option(
    "--dont-update-guest-collections",
    "-D",
    is_flag=True,
    help="If specified, don't update guest collection configurations to use "
    "this mapping.",
)
@click.option(
    "--provision",
    "-p",
    is_flag=True,
    help="If an identity does not exist for the Globus username, provision a new one",
)
@requires_logged_in
@requires_migration_plan
@click.pass_context
def identity_mapping_myproxy(
    ctx: click.Context,
    subject: Optional[str],
    domain: Tuple[str],
    identity_mapping: Optional[str],
    dont_update_guest_collections: bool,
    provision: bool,
    migration_plan: MigrationPlan,
) -> None:
    if not migration_plan:
        raise click.ClickException("Unable to load migration plan")
    if not migration_plan.gsi_authz_configuration:
        raise click.ClickException("Migration plan has no GSI configuration")
    if not migration_plan.gsi_authz_configuration.myproxy_callouts:
        raise click.ClickException("Migration plan has no MyProxy configuration")
    if not migration_plan.storage_gateway:
        raise click.ClickException("Migration plan has no storage gateway")

    myproxy_configurations: Dict[int, "MyProxyConfiguration"] = {}

    # Load identity mapping if specified on the command-line
    identity_mapping_json: Optional[List["IdentityMappingDict"]] = None
    if identity_mapping:
        identity_mapping_json = []
        for idm in identity_mapping:
            try:
                identity_mapping_json.append(parse_identity_mapping(idm))
            except Exception as e:
                raise click.ClickException(str(e))

    # Extract MyProxy configurations from the authz config, and store them
    # along with their indices so we can check them for default values and then
    # update credentials which use these configurations.
    assert migration_plan.gsi_authz_configuration.callouts is not None
    for idx, callout in enumerate(migration_plan.gsi_authz_configuration.callouts):
        if callout.is_myproxy:
            myproxy_callout = cast("MyProxyConfiguration", callout)
            if not subject or subject == myproxy_callout.dn:
                myproxy_configurations[idx] = myproxy_callout

    # Determine subject to use if not specified on the command-line
    subjects: Set[str] = set()
    if not subject:
        for myproxy_callout in myproxy_configurations.values():
            if myproxy_callout.dn:
                subjects.add(myproxy_callout.dn)
        if len(subjects) > 1:
            raise click.ClickException(
                "Your GCSv4 configuration is configured to use more than one "
                "MyProxy CA. Please use the -s command-line option "
                "to select which one you will be adding the mapping for."
            )
        subject = subjects.pop()
    else:
        # Verify that the MyProxy subject was used by a configuration
        found_subject = False
        for myproxy_callout in myproxy_configurations.values():
            if myproxy_callout.dn == subject:
                found_subject = True
        if not found_subject:
            raise click.ClickException(
                f"Your endpoint has no MyProxy configuration for {subject}"
            )

    # Verify domain with Auth before adding them to the gateway
    unusable_domains = get_unusable_idp_domains(set(domain))

    if unusable_domains:
        click.echo(
            "The following domains are not valid identity providers "
            f"recognized by Globus Auth: {', '.join(unusable_domains)}.",
            err=True,
        )
        ctx.exit(1)

    # Add domain and identity mappings to the storage gateway
    migration_plan.storage_gateway.add_allowed_domains(
        list(domain), identity_mappings=identity_mapping_json
    )
    # Update the myproxy callout to indicate the domains being used
    for callout in migration_plan.gsi_authz_configuration.callouts:
        if isinstance(callout, MyProxyConfiguration) and callout.dn == subject:
            for d in list(domain):
                callout.add_domain(d)

    # Additional mappings to handle the case where cred.username
    # isn't lowercase
    additional_mappings: List[Tuple[str, str]] = []

    if migration_plan.guest_collections or migration_plan.credentials:
        # Update the credentials which match this MyProxy configuration to use
        # names if there is only one domain
        if dont_update_guest_collections:
            click.echo(
                "Not updating guest collection ownership. Use gcs-migration "
                "set-guest-collection-owner to update guest collections with "
                "appropriate ownership."
            )
        elif len(domain) > 1:
            click.echo(
                "Not updating guest collection ownership because more than one "
                "domain is being allowed. Use gcs-migration "
                "set-guest-collection-owner to update guest collections with "
                "appropriate ownership."
            )
            dont_update_guest_collections = True

    if not dont_update_guest_collections:
        cred_domain = domain[0].lower()

        # Create and populate an identity map to batch ID lookups
        if provision:
            auth_client = get_api("auth.globus.org", cls=ProvisioningAuthClient)
        else:
            auth_client = get_api("auth.globus.org")
        identity_map = IdentityMap(auth_client)

        for cred in migration_plan.credentials:
            if cred and cred.gsi_authz_callout_index in myproxy_configurations.keys():
                assert cred.gsi_authz_callout_index is not None
                if cred.identity_username is None:
                    identity_username = f"{cred.username}@{cred_domain}".lower()
                    identity_map.add(identity_username)

        # Resolve identities and update user credentials for guest collections.
        for cred in migration_plan.credentials:
            if cred and cred.gsi_authz_callout_index in myproxy_configurations.keys():
                if cred.identity_username is None:
                    identity_username = f"{cred.username}@{cred_domain}"
                    lowered = identity_username.lower()

                    try:
                        identity = identity_map[lowered]
                        cred.identity_username = lowered
                        cred.identity_id = identity["id"]

                        # For S3 connector the identity mapping is a no-op,
                        # but migration-plan create doesn't know about domains.
                        # If this is a new domain assignment for this credential,
                        # update the local username to match the globus username.
                        if isinstance(cred, S3Credential):
                            _fixup_s3_credential(migration_plan, cred)

                        assert cred.username is not None
                        if identity_username != lowered:
                            additional_mappings.append((lowered, cred.username))
                    except (GlobusError, KeyError):
                        click.echo(
                            f"Unable to resolve {identity_username} to a valid "
                            f"Globus identity, not updating configuration for {cred}"
                        )

    if additional_mappings:
        migration_plan.storage_gateway.add_identity_mappings(additional_mappings)

    migration_plan.dump_storage_gateway()
    migration_plan.dump_credentials()
    migration_plan.dump_guest_collections()
    migration_plan.dump_gsi_authz_configuration()
