import typing as t


class _HPSSEnv:
    """
    Look up HPSS environment variables and store them in attributes.
    - path: path to HPSS environment override file
    - hpss_api_authn_mech = HPSS_API_AUTHN_MECH
    - hpss_primary_authn_mech = HPSS_PRIMARY_AUTHN_MECH
    - hpss_primary_authenticator = HPSS_PRIMARY_AUTHN_MECH
    """

    def __init__(self, path: t.Optional[str] = None) -> None:
        """
        Look up HPSS environment variables using 'path' as the override file.
        """

        # Since we can't call the C function hpss_Getenv() to get the official values,
        # we simulate it by scanning the override file. If any value is not found
        # there, we use the well-known, hard-coded defaults from the HPSS headers.

        # Pass 1: Scan the override file
        if path is None:
            path = "/var/hpss/etc/env.conf"

        self.path: str = path
        self.hpss_api_authn_mech: t.Optional[str] = None
        self.hpss_primary_authn_mech: t.Optional[str] = None
        self.hpss_primary_authenticator: t.Optional[str] = None

        with open(path, "r") as f:
            for line in f:
                line = line.split("#", 1)[0].strip()
                if line and "=" in line:
                    k, v = line.split("=", 1)
                    k = k.strip()
                    v = v.strip()

                    if k == "HPSS_API_AUTHN_MECH":
                        self.hpss_api_authn_mech = v
                    elif k == "HPSS_PRIMARY_AUTHN_MECH":
                        self.hpss_primary_authn_mech = v
                    elif k == "HPSS_PRIMARY_AUTHENTICATOR":
                        self.hpss_primary_authenticator = v

        # Pass 2: Use the well-known, hard-coded defaults from HPSS.

        # HPSS_PRIMARY_AUTHN_MECH defaults to 'unix'
        if self.hpss_primary_authn_mech is None:
            self.hpss_primary_authn_mech = "unix"
        # HPSS_API_AUTHN_MECH defaults to HPSS_PRIMARY_AUTHN_MECH
        if self.hpss_api_authn_mech is None:
            self.hpss_api_authn_mech = self.hpss_primary_authn_mech
        # HPSS_PRIMARY_AUTHENTICATOR defaults to hpss.unix.keytab
        if self.hpss_primary_authenticator is None:
            self.hpss_primary_authenticator = (
                "auth_keytab:/var/hpss/etc/hpss.unix.keytab"
            )


class HPSSConfig:
    """
    Determines the v4 HPSS DSI configuration for the local system. Provides these
    values:
    - path: str
    - login_name: str
    - authentication_mech: str
    - authenticator: str
    - uda_checksum_support: "on" | "off"
    """

    def __init__(self, path: t.Optional[str] = None) -> None:
        """
        Look up the v4 HPSS DSI configuration in these locations, in order of
        precedence:
        - path:
           Defaults to "/var/hpss/etc/gridftp_hpss_dsi.conf". Sites can override it by
           using the $HPSS_DSI_CONFIG_FILE gridftp environment variable.

        - hpss_Getenv():
           HPSS C function which has hard-coded values for HPSS environment variables
           that can be overriden in /var/hpss/etc/env.conf.
        """
        if path is None:
            path = "/var/hpss/etc/gridftp_hpss_dsi.conf"

        self.path = path
        self.login_name: str = "hpssftp"
        # Defaults to HPSS_API_AUTHN_MECH or HPSS_PRIMARY_AUTHN_MECH
        self.authentication_mech: t.Optional[str] = None
        # Defaults to HPSS_PRIMARY_AUTHENTICATOR
        self.authenticator: t.Optional[str] = None
        self.uda_checksum_support: str = "off"

        with open(path, "r") as f:
            for line in f:
                line = line.split("#", 1)[0].strip()
                if line:
                    k, v = line.split(" ", 1)
                    k = k.strip()
                    v = v.strip()

                    if k == "LoginName":
                        self.login_name = v
                    elif k == "AuthenticationMech":
                        self.authentication_mech = v
                    elif k == "Authenticator":
                        self.authenticator = v
                    elif k == "UDAChecksumSupport":
                        self.uda_checksum_support = v

        if not self.authentication_mech or not self.authenticator:
            try:
                hpss_env = _HPSSEnv()

                if not self.authentication_mech:
                    self.authentication_mech = hpss_env.hpss_api_authn_mech
                    if not self.authentication_mech:
                        self.authentication_mech = hpss_env.hpss_primary_authn_mech
                if not self.authenticator:
                    self.authenticator = hpss_env.hpss_primary_authenticator
            except FileNotFoundError:
                # /var/hpss/etc/env.conf almost always exists but it is not an error if
                # it is missing.
                pass
