from __future__ import annotations

import typing as t

import globus.manager.api.errors as errors


class IdentitySet(list):
    """
    This is a wrapper around a list of filtered identities which also includes
    information about identities which were filtered out with detail
    information so that the caller can better determine how to fix the problem.

    It provides list-like behavior as an identity set, but adds a error_details
    property that can be serialized in the details part of an forbidden error.
    """

    def __init__(
        self,
        identity_set: list[dict[str, t.Any]] | None = None,
        error_details: list[errors.ErrorDetail] | None = None,
    ) -> None:
        """
        :param identity_set:
            Either a list of dicts, each containing the properties from one
            Auth identity resource, or an IdentitySet.
        :param error_details:
            A optional list of serializable objects which contain the reason
            why one or more identities were removed from consideration for this
            operation.
        """
        if isinstance(identity_set, IdentitySet):
            super().__init__(identity_set)
            self.error_details: list[errors.ErrorDetail] = []
            self.error_details.extend(identity_set.error_details)
        elif identity_set is None:
            super().__init__()
            self.error_details = []
        else:
            super().__init__(identity_set)
            self.error_details = []

        if error_details:
            self.error_details.extend(error_details)

    @staticmethod
    def __error_detail_sort_key(item) -> int:
        """
        Compute a sort key for an error detail object. Non-authorization error
        details are returned with the lowest sort key, and authorization error
        details are returned with a lower sort key indicating a more easily
        repairable error.
        """
        precedence = [
            errors.AuthenticationTimeout,
            errors.NotResourceOwner,
            errors.IdNotInIdentitySet,
            errors.NotFromAllowedDomain,
            errors.InvalidCredential,
            errors.CredentialNotFound,
            errors.InvalidUser,
            errors.MissingRequiredRole,
            errors.SubscriptionRequired,
            errors.NotPublic,
        ]
        value = [p for p in range(len(precedence)) if isinstance(item, precedence[p])]
        if not value:
            value = [-1]
        return value[0]

    @property
    def error_detail(self) -> errors.ErrorDetail | None:
        if self.error_details:
            return sorted(self.error_details, key=self.__error_detail_sort_key)[0]
        return None

    def __repr__(self) -> str:
        return "<IdentitySet {} {}>".format([i["id"] for i in self], self.error_details)
