import datetime
import enum
from typing import List, Optional

from flask import (
    Response,
    current_app,
    g,
    jsonify,
)

import globus_sdk
from globus_sdk import GroupsClient
from globus_sdk.scopes import GroupsScopes

import globus.manager.api.errors as errors
import globus.manager.api.models as models
from globus.manager.api.authorization import (
    RoleAuthorizer,
    RoleSpec,
)
from globus.manager.api.blueprints.decorators import (
    authenticated,
    document_authorization,
    enum_list_converter,
    expects_json,
    query_parameter,
    requires_endpoint_visibility,
    requires_one_role,
    result_data,
)
from globus.manager.api.blueprints.utils import (
    create_object,
    patch_object,
    put_object,
    update_object,
    validate_unique_custom_domain,
)
from globus.manager.api.models.result import Result
from globus.manager.api.schemas.api import data_type_schema as data_types
from globus.manager.api.utils.log import log_groups_exception, log_transfer_exception
from globus.manager.constants import ConfigurationObjectTag, PrincipalType, RoleType

from .openapi import (
    OpenAPIBlueprint,
    RoleAuthorizer as OpenAPIRoleAuthorizer,
    SubscriptionManagerAuthorizer as OpenAPISubscriptionManagerAuthorizer,
)

blueprint = OpenAPIBlueprint(
    "endpoint",
    __name__,
    tag="Endpoint",
    description="""
    A GCS Endpoint is an entity that represents an access point to data
    storage. In GCSv5, this includes the following types of items:

    * One or more Storage Gateways which define policies to access
      specific data storage rooted in some sort of file system or object
      store.  In this version of the specification, ActiveScale, Azure Blob
      Storage, BlackPearl, Box, Ceph, Google Drive, Google Cloud Storage, HPSS,
      iRODS, OneDrive, POSIX, and S3 Storage Gateways are supported.
    * One or more Collections, which may define either a set of
      data on a Storage Gateway that is shared with other Globus users or
      publicly, or a set of data on a Storage Gateway that Globus users
      access with their own local or cloud accounts.
    * One or more nodes which define physical servers providing access to
      the GCS endpoint resources.

    """,
    menu_weight=3,
    url_prefix="/api/",
)


@blueprint.route(
    "/endpoint",
    methods=["GET"],
    summary="Get endpoint definition",
    operationId="getEndpoint",
)
@result_data(
    schemas=[data_types.Endpoint],
    description="Get endpoint response",
    max_items=1,
)
@authenticated
@requires_endpoint_visibility
def endpoint_get() -> Response:
    """
    Get the endpoint.
    """
    endpoint = models.Endpoint.query.one()

    res = data_types.Result.dump(Result(data=[data_types.Endpoint.dump(endpoint)]))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


class EndpointUpdateInclude(enum.Enum):
    endpoint = "endpoint"


@blueprint.route("/endpoint", methods=["PATCH"], summary="Update an endpoint")
@query_parameter(
    name_spec="include",
    description="List of document types to include in the response",
    converter=enum_list_converter(EndpointUpdateInclude),
    default=[],
)
@result_data(
    schemas=[data_types.Endpoint],
    description="Update endpoint response",
    max_items=1,
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(
    schema=data_types.Endpoint,
    patch_schema=True,
    description="Changes to the Endpoint document",
    required=True,
)
def endpoint_patch(*, include: List[EndpointUpdateInclude]) -> Response:
    """
    Update the Endpoint document, changing only the properties included in the
    input. Items explicitly set to null in the input are removed from the
    endpoint document. This operation optionally returns the endpoint after
    applying the changes in the input if the include=endpoint query parameter
    is passed to this operation.
    """
    include_endpoint = EndpointUpdateInclude.endpoint in include

    current_subscription = current_app.subscription

    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()

        old_contact_email = getattr(endpoint, "contact_email", None)
        input_endpoint_id = g.validated_data.get("endpoint_id")

        if input_endpoint_id and str(input_endpoint_id) != endpoint.endpoint_id:
            raise errors.ValidationError("Invalid id")

        new_subscription_id = g.validated_data.get("subscription_id")

        if "subscription_id" in g.validated_data:
            if new_subscription_id != current_subscription:
                _change_subscription(endpoint, new_subscription_id)

        g.validated_data.pop("subscription_id", None)

        changed_port = False
        old_control_port = endpoint.gridftp_control_channel_port
        if "gridftp_control_channel_port" in g.validated_data:
            new_control_port = g.validated_data["gridftp_control_channel_port"]
            changed_port = old_control_port != new_control_port

        patch_object(endpoint, g.validated_data)

        if not current_subscription and new_subscription_id:
            _add_admin_role_for_endpoint_owner(endpoint)
            _add_admin_role_for_collection_owner(endpoint)
            _update_collections()
        elif changed_port or (current_subscription and not new_subscription_id):
            _update_collections()
        elif not old_contact_email and g.validated_data.get("contact_email"):
            _update_collections(mapped_only=True)

        if include_endpoint:
            result = [data_types.Endpoint.dump(endpoint)]
        else:
            result = []

        res = data_types.Result.dump(
            Result(
                data=result, message="Updated endpoint {}".format(endpoint.endpoint_id)
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(str(endpoint.endpoint_id))

    return response


@blueprint.route("/endpoint", methods=["PUT"], summary="Update an endpoint")
@authenticated
@query_parameter(
    name_spec="include",
    description="List of document types to include in the response",
    converter=enum_list_converter(EndpointUpdateInclude),
    default=[],
)
@result_data(
    schemas=[data_types.Endpoint],
    description="Update endpoint response",
    max_items=1,
)
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(
    schema=data_types.Endpoint,
    description="New endpoint document",
    required=True,
)
def endpoint_put(*, include: List[EndpointUpdateInclude]) -> Response:
    """
    Update the endpoint document, replacing all properties with those in the
    input. This operation optionally returns the Endpoint after the update if
    the include=endpoint query parameter is passed to this operation.
    """
    include_endpoint = EndpointUpdateInclude.endpoint in include

    current_subscription = current_app.subscription

    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()

        old_contact_email = getattr(endpoint, "contact_email", None)
        input_endpoint_id = g.validated_data.get("endpoint_id")

        if input_endpoint_id and str(input_endpoint_id) != endpoint.endpoint_id:
            raise errors.ValidationError("Invalid id")

        new_subscription_id = g.validated_data.get("subscription_id")

        if "subscription_id" in g.validated_data:
            if new_subscription_id != current_subscription:
                _change_subscription(endpoint, new_subscription_id)

        changed_port = False
        old_control_port = endpoint.gridftp_control_channel_port
        new_control_port = g.validated_data.get("gridftp_control_channel_port", 443)
        changed_port = old_control_port != new_control_port

        put_object(endpoint, g.validated_data)

        if not current_subscription and new_subscription_id:
            _add_admin_role_for_endpoint_owner(endpoint)
            _add_admin_role_for_collection_owner(endpoint)
            _update_collections()
        elif changed_port or (current_subscription and not new_subscription_id):
            _update_collections()
        elif not old_contact_email and g.validated_data.get("contact_email"):
            _update_collections(mapped_only=True)

    if include_endpoint:
        result = [data_types.Endpoint.dump(endpoint)]
    else:
        result = []
    res = data_types.Result.dump(
        Result(message="Updated endpoint {}".format(endpoint.endpoint_id), data=result)
    )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(str(endpoint.endpoint_id))

    return response


@blueprint.route("/endpoint/owner", methods=["PUT"], summary="Set endpoint owner")
@result_data(
    schemas=[],
    description="Set endpoint owner response",
)
@authenticated
@requires_one_role(role_names=["endpoint:owner"])
@expects_json(
    schema=data_types.EndpointOwner, desription="New endpoint owner", required=True
)
def endpoint_owner_put() -> Response:
    """
    Assign a new identity to act as the endpoint owner.
    """
    new_owner_id = str(g.validated_data["identity_id"])

    new_owner_id_result = current_app.auth_client.get_identities(
        ids=[new_owner_id]
    ).data.get("identities", [])

    if len(new_owner_id_result) != 1:
        current_app.logger.error("Unable to resolve identity_id=%s", new_owner_id)
        raise errors.BadRequestError(
            "Unable to resolve identity_id to a Globus Auth identity"
        )

    new_owner_id = new_owner_id_result[0]["id"]
    new_owner_username = new_owner_id_result[0]["username"]

    if new_owner_username == f"{current_app.client_id}@clients.auth.globus.org":
        current_app.logger.error("Cannot set the endpoint client as endpoint owner")
        raise errors.BadRequestError(
            "The endpoint client identity can not be used as the endpoint owner"
        )

    with current_app.appsync_client.pause_notifications():
        owner_role = models.Role.query.filter(
            ~models.Role.deleted, models.Role.role == RoleType.endpoint_owner  # type: ignore
        ).one()
        old_owner_id = owner_role.identity_id
        endpoint = models.Endpoint.query.one()

        if new_owner_id != owner_role.identity_id:
            update_object(owner_role, lambda x, y: x.put_owner_role(y), new_owner_id)
            old_admin_role = models.Role.query.filter(
                ~models.Role.deleted,  # type: ignore
                models.Role.principal_type == PrincipalType.identity,
                models.Role.identity_id == old_owner_id,
                models.Role.role == RoleType.endpoint_administrator,
            ).one_or_none()

            if old_admin_role and old_admin_role.primary_role:

                def clear_primary(role):
                    role.primary_role = False

                update_object(old_admin_role, clear_primary)
                g.updated_objects.append(old_admin_role.id)

            if current_app.subscription:
                _add_admin_role_for_endpoint_owner(endpoint)
                _add_admin_role_for_collection_owner(endpoint)

        res = data_types.Result.dump(
            Result(
                message="Updated endpoint owner to {} ({})".format(
                    new_owner_id, new_owner_username
                )
            )
        )
    response = jsonify(res)
    response.status_code = res["http_response_code"]
    g.updated_objects.append(str(owner_role.id))

    return response


@blueprint.route(
    "/endpoint/owner_string", methods=["PUT"], summary="Set endpoint owner string"
)
@result_data(schemas=[], description="Set endpoint owner string response")
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(
    schema=data_types.OwnerString,
    description="New endpoint owner_string",
    required=True,
)
def endpoint_owner_string_put() -> Response:
    """
    Modify the endpoint's advertised owner to match the username of one of the
    caller's linked identities.  The identity must have an administrator role
    on the endpoint.
    """
    new_owner_id = str(g.validated_data["identity_id"])

    # Validate the requested owner_id is in the caller's identity set
    id_set = g.auth_state.identity_set_data
    new_owner_id_data = [i for i in id_set if i["id"] == new_owner_id]
    if not new_owner_id_data:
        raise errors.IdNotInIdentitySetError(id=new_owner_id)

    # Validate the requested owner_id has a role or is in a group with a role
    allowed_roles = [RoleSpec("endpoint:administrator"), RoleSpec("endpoint:owner")]
    role_authorizer = RoleAuthorizer(allowed_roles)
    try:
        role_id = role_authorizer.authorize_identity(new_owner_id, g.auth_state)
    except errors.MissingRequiredRoleError as e:
        raise errors.MissingRequiredRoleError(
            f"The identity {new_owner_id} does not have a valid role on the endpoint",
            detail=e.detail if isinstance(e.detail, errors.ErrorDetail) else None,
            roles=allowed_roles,
        )

    if not role_id:
        raise errors.MissingRequiredRoleError(roles=allowed_roles)

    new_username = new_owner_id_data[0].get("username")
    endpoint = models.Endpoint.query.first()
    endpoint.set_owner_string(new_owner_id)

    res = data_types.Result.dump(
        Result(
            message="Updated endpoint owner_string to "
            f"{new_username} ({new_owner_id})"
        )
    )
    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/endpoint/owner_string",
    methods=["DELETE"],
    summary="Reset advertised owner string",
)
@result_data(schemas=[], description="Reset advertised owner string response")
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def endpoint_owner_string_delete() -> Response:
    """
    Reset the endpoint's advertised owner to the client_id of the endpoint.
    """
    endpoint = models.Endpoint.query.first()
    endpoint.reset_owner_string()

    res = data_types.Result.dump(
        Result(
            message="Reset endpoint owner_string to "
            f"{current_app.client_id}@clients.auth.globus.org"
        )
    )
    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/endpoint/subscription_id",
    methods=["PUT"],
    summary="Set the endpoint subscription id",
)
@result_data(
    schemas=[],
    description="Set endpoint owner response",
)
@authenticated
@expects_json(
    schema=data_types.EndpointSubscription,
    description="Endpoint subscription request",
    required=True,
)
@document_authorization(
    authorization=OpenAPIRoleAuthorizer(
        roles=[RoleType.endpoint_owner.value, RoleType.endpoint_administrator.value]
    )
)
@document_authorization(authorization=OpenAPISubscriptionManagerAuthorizer())
def endpoint_subscription_id_put() -> Response:
    """
    Change the subscription_id of this endpoint. Because subscription
    is enforcement is handled in a separate service than GCS and an
    organization's subscription manager may not be the administrator
    of the endpoint, this API has allows for both role-based
    authorization and subscription manager based authorization.

    The authorization allows the following:

    <dl>
    <dt>Caller has a role but is not subscription manager</dt>
    <dd>Remove an existing subscription from an endpoint, even if the
        caller is not a manager for that subscription.</dd>

    <dt>Caller does not have a role but is a subscription manager</dt>
    <dd>Set the subscription_id to a subscription they manage on a
        currently-unmanaged endpoint or remove the subscription_id from
        the endpoint if it is one that they managed.</dd>

    <dt>Caller has a role and is a subscription manager</dt>
    <dd>Set the subscription_id to a subscription they manage on an endpoint
        even if it is currently managed by a subscription that the caller is
        not a manager of.</dd>
    </dl>
    """
    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()
        old_subscription = getattr(endpoint, "subscription_id", None)
        new_subscription = g.validated_data["subscription_id"]

        _change_subscription(endpoint, new_subscription)

    if old_subscription is None and new_subscription is not None:
        _add_admin_role_for_endpoint_owner(endpoint)
        _add_admin_role_for_collection_owner(endpoint)
        _update_collections()
    elif old_subscription is not None and new_subscription is None:
        _update_collections()

    # Assume assistant will update vhost files when subscription state changes
    result = data_types.Result.dump(
        Result(message="Updated Endpoint {}".format(endpoint.id))
    )

    response = jsonify(result)
    response.status_code = result["http_response_code"]
    g.updated_objects.append(str(endpoint.id))

    return response


@blueprint.route(
    "/endpoint/domain",
    methods=["PUT"],
    summary="Set endpoint domain",
    tags=["Domains"],
)
@result_data(
    schemas=[],
    description="Set endpoint domain response",
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
@expects_json(
    schema=data_types.Domain,
    description="Put domain request",
    required=True,
)
def endpoint_domain_put():
    """
    Register a new custom domain and certificate to to be used to serve
    the endpoint or collection.

    The domain is used for the endpoint itself. If the `wildcard`
    property is set to true, then all all collections associated
    with that endpoint that do not have a custom domain will be
    updated to use subdomains of that domain. Otherwise, only the
    endpoint will use that domain.
    """
    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()

        old_domain = getattr(endpoint, "domain", None)
        old_domain_name = old_domain.domain_name if old_domain else None
        old_wildcard = bool(old_domain and old_domain.wildcard)

        if g.validated_data["domain_name"] != endpoint.domain_name:
            validate_unique_custom_domain(g.validated_data["domain_name"])

        update_object(endpoint, models.Endpoint.put_domain, g.validated_data)
        g.updated_objects.append(str(endpoint.id))

        new_domain = endpoint.domain
        new_wildcard = new_domain.wildcard

        if (old_wildcard or new_wildcard) and old_domain_name != new_domain.domain_name:
            _update_collections_using_endpoint_subdomains(endpoint)

    res = data_types.Result.dump(
        Result(
            message="Updated endpoint to use domain {}".format(new_domain.domain_name)
        )
    )
    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


@blueprint.route(
    "/endpoint/domain",
    methods=["DELETE"],
    summary="Delete endpoint domain",
    tags=["Domains"],
)
@result_data(
    schemas=[],
    description="Delete endpoint domain response",
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def endpoint_domain_delete():
    """
    Delete the custom endpoint domain. This will cause the endpoint
    to revert to using a data.globus.org domain for the GCS Manager
    and any collections which do not have custom domains associated
    with them.

    This requires an `administrator` role on the Endpoint.
    """
    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()
        old_domain = getattr(endpoint, "domain", None)

        if not old_domain:
            raise errors.NotFoundError("No domain is configured for this endpoint.")

        update_object(endpoint, models.Endpoint.delete_domain)
        g.updated_objects.append(str(endpoint.id))

        if old_domain and old_domain.wildcard:
            _update_collections_using_endpoint_subdomains(endpoint)

        result = data_types.Result.dump(
            Result(message="Updated Endpoint {}".format(endpoint.id))
        )

    response = jsonify(result)
    response.status_code = result["http_response_code"]
    g.updated_objects.append(str(endpoint.id))

    return response


@blueprint.route(
    "/endpoint/domain",
    methods=["GET"],
    summary="Get endpoint domain",
    tags=["Domains"],
)
@result_data(
    schemas=[],
    description="Get endpoint domain response",
)
@authenticated
@requires_one_role(role_names=["endpoint:owner", "endpoint:administrator"])
def endpoint_domain_get():
    """
    Get the custom domain document associated with this endpoint.

    This requires an `administrator` role on the Endpoint.
    """
    with current_app.appsync_client.pause_notifications():
        endpoint = models.Endpoint.query.one()
        domain = getattr(endpoint, "domain", None)

        if not domain:
            raise errors.NotFoundError("No domain is configured for this endpoint.")

    res = data_types.Result.dump(Result(data=[data_types.Domain.dump(domain)]))

    response = jsonify(res)
    response.status_code = res["http_response_code"]

    return response


def _change_subscription(endpoint: models.Endpoint, new_subscription: str) -> None:
    old_subscription = endpoint.subscription_id

    groups_token = g.auth_state.dependent_token(
        required_scope=GroupsScopes.view_my_groups_and_memberships,
    )
    if not groups_token:
        current_app.logger.debug("No dependent token for groups is available")
        raise errors.InvalidCredentialError(
            error_description=(
                "Unable to authenticate with the Globus groups service on "
                "your behalf"
            )
        )

    transfer_token = g.auth_state.dependent_token("transfer.api.globus.org")
    if not transfer_token:
        current_app.logger.debug("No dependent token for transfer is available")
        raise errors.InvalidCredentialError(
            error_description=(
                "Unable to authenticate with the globus tranfer service on "
                "your behalf"
            )
        )

    groups_client = GroupsClient(
        authorizer=globus_sdk.AccessTokenAuthorizer(groups_token),
        environment=current_app.environment,
    )
    try:
        groups_response = groups_client.get_my_groups().data
    except globus_sdk.GroupsAPIError as e:
        log_groups_exception(e)
        raise errors.BadGatewayError(
            "Error accessing subscription information: {}".format(e.code)
        )

    # subscription groups I am a member of. I can set the subscription to one
    # of these if I am endpoint:administrator or endpoint:owner
    my_subscriptions = [
        group["subscription_id"]
        for group in groups_response
        if group["subscription_id"] is not None
    ]

    role_authorizer = RoleAuthorizer(["endpoint:administrator", "endpoint:owner"])
    role_identities = role_authorizer.authorize_identities(g.auth_state)

    # subscription groups I am an admin of. I can set the subscription to one
    # of these if the endpoint is unmanaged
    my_managed_subscriptions = [
        group["subscription_id"]
        for group in groups_response
        if group["subscription_id"] is not None
        and (
            "admin" in [m["role"] for m in group["my_memberships"]]
            or "manager" in [m["role"] for m in group["my_memberships"]]
        )
    ]

    if not role_identities:
        # We don't have a role, so we must use a subscription we are an admin of
        use_subscriptions = my_managed_subscriptions
    else:
        # We have a role, we can use any subscription we are a member of
        use_subscriptions = my_subscriptions

    if new_subscription == "DEFAULT":
        if len(use_subscriptions) == 1:
            new_subscription = use_subscriptions[0]
        elif len(my_subscriptions) > 1:
            raise errors.BadRequestError(
                "Unable to use DEFAULT subscription. "
                "Use one of {}".format([s for s in use_subscriptions])
            )
        else:
            raise errors.BadRequestError(
                "Unable to use DEFAULT subscription. "
                "Your identity does not manage any subscriptions"
            )

    if new_subscription is not None and new_subscription not in use_subscriptions:
        raise errors.BadRequestError(
            f"Your identity does not manage subscription {new_subscription}: only {use_subscriptions}"
        )

    current_app.logger.debug(
        "old_subscription=%s new_subscription=%s id_with_role=%s",
        old_subscription,
        new_subscription,
        role_identities,
    )
    if old_subscription is not None:
        if not (role_identities or old_subscription in my_managed_subscriptions):
            raise errors.ForbiddenError(
                "You must have an endpoint owner or administrator role, or "
                "be a subscription manager to remove the subscription "
                "from this endpoint"
            )

    if new_subscription is None:
        if not (role_identities or old_subscription in my_managed_subscriptions):
            raise errors.ForbiddenError(
                "You must have an endpoint owner or administrator role, or "
                "be a subscription manager to remove a subscription "
                "from this endpoint"
            )
    else:
        if not (role_identities or new_subscription in my_managed_subscriptions):
            raise errors.ForbiddenError(
                "You must have an endpoint owner or administrator role, or "
                "be a subscription manager to add a subscription "
                "to this endpoint"
            )

    def update_subscription(endpoint, new_subscription):
        resource = f"/private/endpoint/{endpoint.id}/gcs_subscription_id"

        try:
            current_app.logger.debug("Attempting to set new subscription")
            subscription_doc = {
                "user_token": transfer_token,
                "subscription_id": new_subscription,
            }
            current_app.transfer_client.put(resource, data=subscription_doc)
            endpoint.subscription_id = new_subscription
            current_app.logger.debug("Set subscription successful")
        except globus_sdk.TransferAPIError as e:
            log_transfer_exception(e)

    update_object(endpoint, update_subscription, new_subscription)
    current_app.subscription = new_subscription

    if endpoint.subscription_id != new_subscription:
        raise errors.BadRequestError("Unable to set subscription")


def _add_admin_role_for_endpoint_owner(endpoint: models.Endpoint) -> None:
    """
    Create an endpoint:administrator role on this endpoint for the
    identity which currently has an endpoint:owner role
    """
    owner_role = models.Role.query.filter(
        models.Role.role == RoleType.endpoint_owner
    ).one()

    admin_role = models.Role.query.filter(
        models.Role.role == RoleType.endpoint_administrator,
        models.Role.principal_type == PrincipalType.identity,
        models.Role.identity_id == owner_role.identity_id,
        models.Role.resource.is_(None),  # type: ignore
    ).one_or_none()

    if not admin_role:
        admin_role = models.Role(
            doc=dict(
                DATA_TYPE="role#1.0.0",
                principal=owner_role.principal,
                role="administrator",
            ),
            primary_role=True,
            endpoint=endpoint,
        )
        create_object(admin_role)
        g.created_objects.append(str(admin_role.id))


def _add_admin_role_for_collection_owner(endpoint: models.Endpoint):
    """
    Create a collection:administrator role on each mapped collection where the
    identity_id associated with the collection does not have a collection
    administrator role
    """
    for collection in models.Collection.query.filter(
        ~models.Collection.deleted, models.Collection.is_mapped_collection  # type: ignore
    ):
        admin_role = models.Role.query.filter(
            ~models.Role.deleted,  # type: ignore
            models.Role.role == RoleType.collection_administrator,
            models.Role.principal_type == PrincipalType.identity,
            models.Role.identity_id == collection.identity_id,
            models.Role.resource == collection.id,
        ).one_or_none()
        if not admin_role:
            admin_role = models.Role(
                doc=dict(
                    DATA_TYPE="role#1.0.0",
                    principal="urn:globus:auth:identity:" + collection.identity_id,
                    role="administrator",
                    collection=collection.id,
                ),
                primary_role=True,
                endpoint=endpoint,
            )
            create_object(admin_role)
            g.created_objects.append(str(admin_role.id))


def _update_collections(mapped_only: bool = False) -> None:
    """
    Subscription state, contact_email, or gridftp_control_channel_port changed,
    so update all non deleted collection endpoint documents in transfer so that
    they are consistent with the endpoint configuration.
    """
    query = models.Collection.query.filter(~models.Collection.is_deleted)  # type: ignore
    if mapped_only:
        query = query.filter(models.Collection.is_mapped_collection)  # type: ignore

    ep_serial: int = (
        models.db.session.query(models.ConfigurationObject.serial)
        .filter(models.ConfigurationObject.tag == ConfigurationObjectTag.ENDPOINT)
        .scalar()
    )
    collection: models.Collection

    for collection in query:
        pending_update = models.PendingUpdate(
            object_id=collection.id,
            endpoint_serial=ep_serial,
            change_time=datetime.datetime.now(tz=datetime.timezone.utc),
            update_type=models.PendingUpdate.PATCH,
        )
        models.db.session.add(pending_update)
    models.db.session.commit()


def _update_collections_using_endpoint_subdomains(endpoint: models.Endpoint) -> None:
    """
    Endpoint domain changed, so update all non deleted collections that do not
    have a custom domain. Unfortunately things can change, so we'll check every
    non-deleted collection and use the dependency tracking in appsync to sort out
    things that are updated on us.
    """
    ep_serial: int = (
        models.db.session.query(models.ConfigurationObject.serial)
        .filter(models.ConfigurationObject.tag == ConfigurationObjectTag.ENDPOINT)
        .scalar()
    )
    collection: models.Collection

    for collection in models.Collection.query.filter(
        ~models.Collection.is_deleted,  # type:ignore
    ):
        collection_serial: Optional[int] = None

        if collection.is_guest_collection:
            collection_serial = collection.mapped_collection.serial

        pending_update = models.PendingUpdate(
            object_id=collection.id,
            endpoint_serial=ep_serial,
            change_time=datetime.datetime.now(tz=datetime.timezone.utc),
            update_type=models.PendingUpdate.UPDATE_DOMAIN,
            collection_serial=collection_serial,
        )
        models.db.session.add(pending_update)
    models.db.session.commit()
