import copy
import re
import typing as t
from dataclasses import dataclass
from types import MethodType

from flask import Blueprint, current_app
from typing_extensions import ParamSpec, Protocol

import globus.manager.api.errors as errors


class PublicAuthorizer:
    """
    The PublicAuthorizer allows the operation if the endpoint
    document's `public` property is `true`.
    """

    def json(self) -> t.Any:
        return "PublicAuthorizer"


@dataclass
class RoleAuthorizer:
    """
    A Role authorizer parses the request and the roles assigned to
    the current user. If one of the roles is in the listed set, then
    the operation is authorized. Understood roles are:

    <dl>
    <dt><pre>endpoint:owner</pre></dt>
    <dd>Endpoint owner</dd>
    <dt><pre>endpoint:administrator</pre></dt>
    <dd>Endpoint administrator</dd>
    <dt><pre>endpoint:activity_manager</pre></dt>
    <dd>Endpoint activity_manager</dd>
    <dt><pre>endpoint:activity_monitor</pre></dt>
    <dd>Endpoint activity_monitor</dd>
    <dt><pre>collection:administrator:*</pre></dt>
    <dd>Collection administrator for any collection on this endpoint</dd>
    <dt><pre>collection:activity_manager:*</pre></dt>
    <dd>Collection activity_manager for any collection on this endpoint</dd>
    <dt><pre>collection:activity_monitor:*</pre></dt>
    <dd>Collection activity_monitor for any collection on this endpoint</dd>
    <dt><pre>collection:access_manager:*</pre></dt>
    <dd>Collection access_manager for any collection on this endpoint</dd>
    <dt><pre>collection:administrator:{collection_id}</pre></dt>
    <dd>Collection administrator for the collection being operated on</dd>
    <dt><pre>collection:activity_manager:{collection_id}</pre></dt>
    <dd>Collection activity_manager for the collection being operated on</dd>
    <dt><pre>collection:activity_monitor:{collection_id}</pre></dt>
    <dd>Collection activity_monitor for the collection being operated on</dd>
    <dt><pre>collection:access_manager:{collection_id}</pre></dt>
    <dd>Collection access_manager for the collection being operated on</dd>
    <dt><pre>collection:administrator:{mapped_collection_id}</pre></dt>
    <dd>Collection administrator for the mapped collection associated with the
    guest collection that is being being operated on</dd>
    <dt><pre>collection:activity_manager:{mapped_collection_id}</pre></dt>
    <dd>activity_manager for the mapped collection associated
    the guest collection that is being being operated on</dd>
    <dt><pre>collection:activity_monitor:{mapped_collection_id}</pre></dt>
    <dd>activity_monitor for the mapped collection associated
    the guest collection that is being being operated on</dd>
    <dt><pre>collection:access_manager:{mapped_collection_id}</pre></dt>
    <dd>access_manager for the mapped collection associated
    the guest collection that is being being operated on</dd>
    </dl>
    """

    roles: t.List[str]

    def json(self) -> t.Dict[str, t.List[str]]:
        return {"RoleAuthorizer": [role for role in self.roles]}


@dataclass
class StorageGatewayAuthorizer:
    """
    The StorageGatewayAuthorizer allows the operation if the caller's identity
    set is allowed by the policies of a storage gateway. The parameter to this
    authorizer is one of the following:

    <dl>
    <dt><pre>*</pre></dt>
    <dd>Any storage gateway</dd>
    <dt><pre>{storage_gateway_id}</pre></dt>
    <dd>Either the storage gateway that was passed in to this operation, or the
    storage gateway that the resource (either a user credential or collection)
    is associated with.</dd>
    </dl>
    """

    scopes: t.List[str]

    def json(self) -> t.Dict[str, t.List[str]]:
        return {"StorageGatewayAuthorizer": [scope for scope in self.scopes]}


@dataclass
class SubscriptionManagerAuthorizer:
    """
    The SubscriptionManagerAuthorizer allows the operation if the caller's
    identity set is a Globus subscription manager. This allows an endpoint to
    be set as managed without the subscription manager being the endpoint
    administrator.
    """

    def json(self) -> str:
        return "SubscriptionManagerAuthorizer"


@dataclass
class UserCredentialOwnerAuthorizer:
    """
    The UserCredentialOwnerAuthorizer allows the operation if the caller's
    identity set owns a credential. The parameter to this authorizer is one of
    the following:

    <dl>
    <dt><pre>*</pre></dt>
    <dd>Any User Credential</dd>
    <dt><pre>{user_credential_id}</pre></dd>
    <dd>The user credential passed in to the operation</dd>
    </dl>
    """

    scopes: t.List[str]

    def json(self) -> t.Dict[str, t.List[str]]:
        return {"UserCredentialOwnerAuthorizer": [scope for scope in self.scopes]}


AuthorizerExtension = t.Union[
    PublicAuthorizer,
    RoleAuthorizer,
    StorageGatewayAuthorizer,
    SubscriptionManagerAuthorizer,
    UserCredentialOwnerAuthorizer,
]


class OpenAPIMetadata:
    method: t.Optional[str]
    path: t.Optional[str]
    operationId: t.Optional[str]
    summary: t.Optional[str]
    tags: t.List[str]
    parameters: t.List[t.Dict[str, t.Any]]
    request_body: t.Dict[str, t.Any]
    responses: t.Dict[str, t.Any]
    authorization: t.List[AuthorizerExtension]
    extra_metadata: t.Dict[str, t.Any]

    def __init__(self) -> None:
        self.method = None
        self.path = None
        self.operationId = None
        self.summary = None
        self._description: t.Optional[str] = None
        self.tags = []
        self.parameters = []
        self.request_body = {}
        self.responses = {}
        self.extra_metadata = {}
        self.authorization = []

    @property
    def description(self) -> t.Optional[str]:
        return self._description

    @description.setter
    def description(self, value: t.Optional[str]) -> None:
        if value is None:
            self._description = None
        else:
            if value.startswith("\n"):
                value = value[1:]
            match = re.match("^( *)", value)
            if match:
                leading_space = len(match[1])
                value = "\n".join(
                    [
                        line[leading_space:]
                        if line.startswith(" " * leading_space)
                        else line
                        for line in value.split("\n")
                    ]
                )
            self._description = value

    def returns_error(
        self,
        error_type: t.Type[errors.APIBaseError],
        error_detail: t.Optional[t.List[t.Type[errors.ErrorDetail]]] = None,
        response_id: t.Optional[str] = None,
    ) -> None:
        http_status = error_type.status_code or 400

        status_description = {
            400: "Bad Request",
            401: "Unauthorized",
            403: "Permission denied",
            404: "Not found",
            405: "Method not allowed",
            409: "Conflict",
            415: "Unsupported media type",
            422: "Unprocessable entity",
        }[http_status]

        detail_ref: t.Dict[str, t.Any]
        if error_detail:
            detail_ref = {
                "detail": {
                    "oneOf": [{"type": "string"}]
                    + [
                        {"$ref": f"#/components/schemas/{d.__name__}"}
                        for d in error_detail
                    ],
                }
            }
        else:
            detail_ref = {}

        response_id_prop = {"x-responseId": response_id} if response_id else {}

        response_schema_specialization = {
            "type": "object",
            "properties": {
                "code": {"type": "string", "enum": [error_type.error]},
                "http_response_code": {
                    "enum": [http_status],
                },
                **detail_ref,
            },
        }

        if str(http_status) not in self.responses:
            self.responses[str(http_status)] = {
                "description": status_description,
                "content": {"application/json": {"schema": {"oneOf": []}}},
                **response_id_prop,
            }
        schema = self.responses[str(http_status)]["content"]["application/json"][
            "schema"
        ]
        new_entry = {
            "allOf": [
                response_schema_specialization,
                {"$ref": "#/components/schemas/Result"},
            ]
        }
        for entry in schema["oneOf"]:
            if entry == new_entry:
                return

        schema["oneOf"].append(new_entry)

    def responds_with_redirect(self, http_status, description):
        status_description = {
            302: "Found",
            303: "See Other",
        }[http_status]

        if str(http_status) not in self.responses:
            self.responses[str(http_status)] = {
                "description": description or status_description,
            }

    def json(self) -> t.Dict[str, t.Any]:
        data: t.Dict[str, t.Any] = {}
        if self.operationId:
            data["operationId"] = self.operationId
        if self.summary:
            data["summary"] = self.summary
        if self.tags:
            data["tags"] = self.tags
        if self.description:
            data["description"] = self.description
        if self.parameters:
            data["parameters"] = self.parameters
        if self.request_body:
            data["requestBody"] = self.request_body
        if self.responses:
            data["responses"] = self.responses
        if self.authorization:
            data["x-authorizers"] = [auth.json() for auth in self.authorization]

        data.update(self.extra_metadata)

        return data


T = t.TypeVar("T")
CT = t.TypeVar("CT", covariant=True)
P = ParamSpec("P")


class OpenAPIAnnotated(Protocol[P, CT]):
    metadata: OpenAPIMetadata
    operation_metadata: t.Dict[str, OpenAPIMetadata]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> CT:
        ...


def ensure_openapi(fn: t.Callable[P, T]) -> OpenAPIAnnotated[P, T]:
    fn = t.cast(OpenAPIAnnotated[P, T], fn)
    if isinstance(fn, MethodType):
        fnobj = fn.__self__
        method = None
        for f in dir(fnobj):
            try:
                if getattr(fnobj, f) == fn:
                    method = getattr(fnobj.__class__, f)
                    break
            except Exception:
                pass
        if method:
            method.metadata = getattr(method, "metadata", OpenAPIMetadata())
            method.operation_metadata = getattr(method, "operation_metadata", {})
    else:
        fn.metadata = getattr(fn, "metadata", OpenAPIMetadata())
        fn.operation_metadata = getattr(fn, "operation_metadata", {})

    return fn


class OpenAPIBlueprint(Blueprint):
    def __init__(
        self,
        name,
        *args,
        tag: str,
        description: str,
        menu_weight: t.Optional[int] = None,
        schema_name: t.Optional[str] = None,
        private: bool = False,
        **kwargs,
    ):
        self.tag = tag
        self.schema_name = schema_name
        if description.startswith("\n"):
            description = description[1:]
        match = re.match("^( *)", description)
        if match:
            leading_space = len(match[1])
            description = "\n".join(
                [
                    line[leading_space:]
                    if line.startswith(" " * leading_space)
                    else line
                    for line in description.split("\n")
                ]
            )
        self.description = description
        self.menu_weight = menu_weight
        super().__init__(name, *args, **kwargs)

    def add_url_rule(  # type: ignore
        self,
        rule: str,
        endpoint: t.Optional[str],
        view_func: t.Callable,
        private: bool = False,
        **kwargs,
    ) -> None:
        # Normalize path, removing duplicate / and replacing flask convert
        # <type:name> with openapi {type:name}
        url_path = re.sub("//*", "/", f"{self.url_prefix}/{rule}")

        templated_parameters = {}
        for templated in re.finditer("<(([^:]+):)?([^>]+)>", url_path):
            templated_parameters[templated.group(3)] = templated.group(2)

        url_path = re.sub("<[^:]*:([^>]*)>", r"{\1}", url_path)

        view_func = ensure_openapi(view_func)

        if templated_parameters:
            view_func.metadata.returns_error(
                error_type=errors.NotFoundError, response_id="NotFound"
            )

        methods = kwargs.get("methods", [])
        for meth in methods:
            metadata = view_func.metadata
            if "summary" in kwargs:
                metadata.summary = kwargs.pop("summary")
            if "tags" in kwargs:
                metadata.tags = kwargs.pop("tags")
            else:
                metadata.tags = [self.tag]
            if "operationId" in kwargs:
                operation_id = kwargs.pop("operationId")
            else:
                operation_id = meth.lower() + "".join(
                    [
                        word.capitalize()
                        for word in re.split(
                            "[_-]",
                            re.sub("<([^:]*:)?([^>]+)>", r"\2", rule).replace("/", "-"),
                        )
                    ]
                )
            metadata.operationId = operation_id
            if metadata.description is None:
                metadata.description = view_func.__doc__
            operation_metadata = view_func.operation_metadata[
                operation_id
            ] = copy.deepcopy(metadata)
            operation_metadata.method = meth
            operation_metadata.path = url_path

            for param, fmt in templated_parameters.items():
                param_spec: t.Dict[str, t.Any] = {
                    "name": param,
                    "in": "path",
                    "required": True,
                    "schema": {"type": "string"},
                }
                if fmt:
                    param_spec["schema"]["format"] = fmt
                if "path_parameters" in kwargs and param in kwargs["path_parameters"]:
                    param_spec["description"] = kwargs["path_parameters"][param]

                operation_metadata.parameters.append(param_spec)
            if "deprecated" in kwargs:
                operation_metadata.extra_metadata["deprecated"] = kwargs["deprecated"]
            if private:
                operation_metadata.extra_metadata["x-private"] = True
        kwargs.pop("deprecated", None)
        kwargs.pop("path_parameters", None)
        return super().add_url_rule(rule, endpoint, view_func, **kwargs)

    def openapi(self) -> t.Dict[str, t.Any]:
        data: t.Dict[str, t.Any] = {
            "tags": [
                {
                    "name": self.tag,
                    "description": self.description,
                }
            ],
            "paths": {},
            "components": {
                "parameters": {},
                "responses": {},
                "schemas": {},
            },
        }
        if self.menu_weight:
            data["tags"][0]["x-menu-weight"] = self.menu_weight

        for rule in current_app.url_map.iter_rules():
            blueprint_name = rule.endpoint.split(".", 1)[0]
            if blueprint_name != self.name:
                continue

            url_path = re.sub("//*", "/", rule.rule)
            templated_parameters = {}
            for templated in re.finditer("<(([^:]+):)?([^>]+)>", url_path):
                templated_parameters[templated.group(3)] = templated.group(2)

            url_path = re.sub("<[^:]*:([^>]+)>", r"{\1}", url_path)
            view_func = ensure_openapi(current_app.view_functions[rule.endpoint])
            for metadata in view_func.operation_metadata.values():
                method = metadata.method
                assert method is not None
                if method not in rule.methods:
                    continue
                if url_path != metadata.path:
                    continue
                # Move parameters with titles to the components section and
                # replace with a ref
                for idx, param in enumerate(metadata.parameters):
                    if "title" in param:
                        title = param["title"]
                        data["components"]["parameters"][title] = {
                            k: v for k, v in param.items() if k != "title"
                        }
                        metadata.parameters[idx] = {
                            "$ref": f"#/components/parameters/{title}"
                        }
                # Move responses with x-responseId to the components section and
                # replace with a ref
                for status, resp in metadata.responses.items():
                    if "x-responseId" in resp:
                        response_id = resp["x-responseId"]
                        data["components"]["responses"][response_id] = {**resp}
                        metadata.responses[status] = {
                            "$ref": f"#/components/responses/{response_id}"
                        }

                if url_path not in data["paths"]:
                    data["paths"][url_path] = {}
                json_metadata = metadata.json()
                data["paths"][url_path][method.lower()] = json_metadata

        return data
