from types import MethodType
from typing import Any, Dict

from oauth2client import client as oauth_client

import globus.manager.api.models as m
import globus.manager.api.schemas.connectors.dropbox  # noqa

from .oauth_credential_api import OAuthCredentialAPI


class DropboxConnector:
    CONNECTOR_ID = "49b00fd6-63f1-48ae-b27f-d8af4589f876"
    DISPLAY_NAME = "Dropbox"
    model = m.DropboxStorageGateway
    SCOPES = (
        "profile "
        "openid "
        "email "
        "account_info.read "
        "files.metadata.write "
        "files.metadata.read "
        "files.content.write "
        "files.content.read"
    )

    def __init__(self, app):
        self.oauth_credential_api = OAuthCredentialAPI(app)
        self.oauth_credential_api.add_connector(self.CONNECTOR_ID, self)

    def initiate_flow(
        self,
        storage_gateway: m.StorageGateway,
        authorized_identity: Dict[str, Any],
        **kwargs
    ) -> oauth_client.OAuth2WebServerFlow:
        flow = oauth_client.OAuth2WebServerFlow(
            client_id=storage_gateway.client_id,
            client_secret=storage_gateway.secret,
            scope=self.SCOPES,
            prompt="consent",
            token_access_type="offline",
            redirect_uri=self.callback_url,
            auth_uri="https://www.dropbox.com/oauth2/authorize",
            token_uri="https://api.dropboxapi.com/oauth2/token",
            revoke_uri="https://api.dropboxapi.com/2/auth/token/revoke",
            **kwargs
        )
        if not getattr(storage_gateway, "allow_any_account", False):
            mapped_to = storage_gateway.map_identity(authorized_identity)
            flow.params["login_hint"] = mapped_to

        flow.step2_exchange = MethodType(step2_exchange, flow)
        return flow

    @property
    def callback_url(self) -> str:
        return self.oauth_credential_api.authcallback_url


def step2_exchange(self, code=None, http=None, device_flow_info=None):
    """
    Custom impl of the step2_exchange method because Dropbox doesn't
    like the 'scope' param being sent in this step by the builtin method.
    """
    post_data = {
        "client_id": self.client_id,
        "code": code,
        "client_secret": self.client_secret,
        "grant_type": "authorization_code",
        "redirect_uri": self.redirect_uri,
    }
    body = oauth_client.urllib.parse.urlencode(post_data)
    headers = {
        "content-type": "application/x-www-form-urlencoded",
    }

    http = oauth_client.transport.get_http_object()

    resp, content = oauth_client.transport.request(
        http, self.token_uri, method="POST", body=body, headers=headers
    )
    d = oauth_client._parse_exchange_token_response(content)
    if resp.status == oauth_client.http_client.OK and "access_token" in d:
        access_token = d["access_token"]
        refresh_token = d.get("refresh_token", None)
        token_expiry = None
        if "expires_in" in d:
            delta = oauth_client.datetime.timedelta(seconds=int(d["expires_in"]))
            token_expiry = delta + oauth_client._UTCNOW()

        extracted_id_token = None
        id_token_jwt = None
        if "id_token" in d:
            extracted_id_token = oauth_client._extract_id_token(d["id_token"])
            id_token_jwt = d["id_token"]

        oauth_client.logger.info("Successfully retrieved access token")
        return oauth_client.OAuth2Credentials(
            access_token,
            self.client_id,
            self.client_secret,
            refresh_token,
            token_expiry,
            self.token_uri,
            self.user_agent,
            revoke_uri=self.revoke_uri,
            id_token=extracted_id_token,
            id_token_jwt=id_token_jwt,
            token_response=d,
            scopes=self.scope,
            token_info_uri=self.token_info_uri,
        )
    else:
        oauth_client.logger.info("Failed to retrieve access token: %s", content)
        if "error" in d:
            # you never know what those providers got to say
            error_msg = str(d["error"]) + str(d.get("error_description", ""))
        else:
            error_msg = "Invalid response: {0}.".format(str(resp.status))
        raise oauth_client.FlowExchangeError(error_msg)
