from typing import Any, Dict

from oauth2client import client as oauth_client

import globus.manager.api.errors as errors
import globus.manager.api.models as m
import globus.manager.api.schemas.connectors.onedrive  # noqa

from .oauth_credential_api import OAuthCredentialAPI


class OneDriveConnector:
    CONNECTOR_ID = "28ef55da-1f97-11eb-bdfd-12704e0d6a4d"
    DISPLAY_NAME = "OneDrive"
    model = m.OneDriveStorageGateway
    SCOPES = "openid email profile offline_access user.read files.readwrite.all sites.readwrite.all"

    def __init__(self, app):
        self.oauth_credential_api = OAuthCredentialAPI(app)
        self.oauth_credential_api.add_connector(self.CONNECTOR_ID, self)

    def initiate_flow(
        self,
        storage_gateway: m.StorageGateway,
        authorized_identity: Dict[str, Any],
        **kwargs
    ) -> oauth_client.OAuth2WebServerFlow:
        auth_uri_base = "https://login.microsoftonline.com/{}/oauth2/v2.0/".format(
            getattr(storage_gateway, "tenant", "common")
        )
        flow = oauth_client.OAuth2WebServerFlow(
            client_id=storage_gateway.client_id,
            client_secret=storage_gateway.secret,
            scope=self.SCOPES,
            redirect_uri=self.callback_url,
            auth_uri=auth_uri_base + "authorize",
            token_uri=auth_uri_base + "token",
            **kwargs,
        )
        if not getattr(storage_gateway, "allow_any_account", False):
            mapped_to = storage_gateway.map_identity(authorized_identity)
            if not mapped_to:
                raise errors.InvalidUserError()
            split_username = mapped_to.split("@")
            flow.params["login_hint"] = "@".join(split_username[0:2])

        return flow

    @property
    def callback_url(self) -> str:
        return self.oauth_credential_api.authcallback_url

    def identity_username(
        self, storage_gateway: m.StorageGateway, id_token: Dict[str, Any]
    ) -> str:
        # "upn" is an optional claim. if they enable it, use that.  next
        # use "preferred_username" as long as it looks like an email address.
        # finally use "email".
        username = id_token.get("upn", None)
        if not username:
            username = id_token.get("preferred_username", "")
        if "@" not in username:
            username = id_token["email"]
        return username
