from __future__ import annotations

import typing as t
from uuid import UUID

from globus.manager.api.errors import ErrorDetail, ForbiddenError

if t.TYPE_CHECKING:
    from globus.manager.api.authorization.role import RoleSpec
    from globus.manager.api.models.account import Account


StringOrUUID = t.TypeVar("StringOrUUID", str, UUID)


class MissingRequiredRole(ErrorDetail):
    """
    Error details implementing missing_required_role#1.0.0
    """

    DATA_TYPE = "missing_required_role#1.0.0"

    def __init__(self, roles: list[RoleSpec] | None = None) -> None:
        self.roles: list[t.Mapping[str, t.Any]] = []
        for r in roles or []:
            role: dict[str, t.Any] = {"role": r.role_name}

            if r.collection:
                role["collection"] = r.resource
            self.roles.append(role)


class MissingRequiredRoleError(ForbiddenError):
    """
    ForbiddenError with information about missing role needed for
    authorization.
    """

    error_description = (
        "None of your identities have been granted a role to access this resource"
    )

    def __init__(
        self,
        *args,
        roles: list[RoleSpec] | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None:
            detail = MissingRequiredRole(roles=roles)

        if detail:
            kwargs["detail"] = detail

        super().__init__(*args, **kwargs)


class NotResourceOwner(ErrorDetail):
    """
    Error details implementing not_resource_owner#1.0.0
    """

    DATA_TYPE = "not_resource_owner#1.0.0"

    def __init__(self, id: StringOrUUID | None = None) -> None:
        if id:
            if not isinstance(id, UUID):
                self.id = UUID(id)
            else:
                self.id = id


class NotResourceOwnerError(ForbiddenError):
    """
    ForbiddenError with information about the resource the caller does
    not own
    """

    error_description = "None of your identities are the owner of this resource"

    def __init__(
        self,
        *args,
        id: StringOrUUID | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None:
            detail = NotResourceOwner(id=id)

        if detail:
            kwargs["detail"] = detail

        super().__init__(*args, **kwargs)


class NotFromAllowedDomain(ErrorDetail):
    """
    Error details implementing not_from_allowed_domain#1.0.0
    """

    DATA_TYPE = "not_from_allowed_domain#1.0.0"

    def __init__(self, allowed_domains: list[str] | None = None) -> None:
        self.allowed_domains = allowed_domains


class NotFromAllowedDomainError(ForbiddenError):
    """
    ForbiddenError with information about the domains the caller's
    identity must be issue by.
    """

    error_description = (
        "None of your identities are from domains allowed by resource policies"
    )

    def __init__(
        self,
        *args,
        allowed_domains: list[str] | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None:
            detail = NotFromAllowedDomain(allowed_domains=allowed_domains)

        if detail:
            kwargs["detail"] = detail

        super().__init__(*args, **kwargs)

        if isinstance(self.detail, NotFromAllowedDomain):
            allowed_domains = self.detail.allowed_domains
            if allowed_domains:
                self.authorization_parameters = {
                    "session_required_single_domain": allowed_domains
                }


class AuthenticationTimeout(ErrorDetail):
    """
    Error details implementing authentication_timeout#1.1.0
    """

    DATA_TYPE = "authentication_timeout#1.1.0"

    def __init__(
        self,
        identities: list[t.Mapping[str, t.Any]] | None = None,
        high_assurance: bool = False,
        require_mfa: bool = False,
    ) -> None:
        if identities is not None:
            self.identities = [i["id"] for i in identities]
            self.identity_names = [i["username"] for i in identities]
        else:
            self.identities = []
            self.identity_names = []

        self.high_assurance = high_assurance
        self.require_mfa = require_mfa


class AuthenticationTimeoutError(ForbiddenError):
    """
    ForbiddenError with information about authentication timeout policies
    """

    def __init__(
        self,
        *args,
        identities: list[t.Mapping[str, t.Any]] | None = None,
        high_assurance: bool = False,
        require_mfa: bool = False,
        detail: AuthenticationTimeout | None = None,
        error_description: str | None = None,
        **kwargs,
    ) -> None:
        if detail is None:
            detail = AuthenticationTimeout(
                identities, high_assurance=high_assurance, require_mfa=require_mfa
            )
        if not error_description:
            if not identities and isinstance(detail, AuthenticationTimeout):
                identities = [{"username": i} for i in detail.identity_names]

            error_description = "You must reauthenticate one of your identities "
            if identities:
                error_description += "({}) ".format(
                    ", ".join([i["username"] for i in identities])
                )
            if high_assurance:
                error_description += "in this session "
            if require_mfa:
                error_description += "using multi-factor authentication "
            error_description += "in order to access this resource"

        if detail:
            kwargs["detail"] = detail

        kwargs["error_description"] = error_description

        super().__init__(*args, **kwargs)

        if not identities and isinstance(self.detail, AuthenticationTimeout):
            identities = self.detail.identities

        self.authorization_parameters: dict[str, t.Any] = {
            "session_required_mfa": require_mfa,
            "session_message": kwargs["error_description"],
        }
        if identities:
            self.authorization_parameters["session_required_identities"] = identities
        else:
            self.authorization_parameters["prompt"] = "login"


class NotPublic(ErrorDetail):
    """
    Error details implementing not_public#1.0.0
    """

    DATA_TYPE = "not_public#1.0.0"

    def __init__(self, resource: StringOrUUID | None = None) -> None:
        self.resource: UUID | None = None
        if isinstance(resource, UUID):
            self.resource = resource
        elif isinstance(resource, str):
            self.resource = UUID(resource)


class NotPublicError(ForbiddenError):
    def __init__(
        self,
        *args,
        resource: StringOrUUID | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None:
            detail = NotPublic(resource=resource)
        if detail:
            kwargs["detail"] = detail
        super().__init__(*args, **kwargs)


class IdNotInIdentitySet(ErrorDetail):
    """
    Error details implementing id_not_in_identity_set#1.0.0
    """

    DATA_TYPE = "id_not_in_identity_set#1.0.0"

    def __init__(self, id: StringOrUUID | None = None) -> None:
        if id:
            if not isinstance(id, UUID):
                self.id = UUID(id)
            else:
                self.id = id


class IdNotInIdentitySetError(ForbiddenError):
    """
    ForbiddenError with information about the required identity id needed
    to perform the requested operation.
    """

    error_description = "Identity is not available in the current identity_set"

    def __init__(
        self,
        *args,
        id: StringOrUUID | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None:
            detail = IdNotInIdentitySet(id=id)
        if detail:
            kwargs["detail"] = detail
        super().__init__(*args, **kwargs)


class InvalidCredential(ErrorDetail):
    """
    Error details implementing invalid_credential#1.0.0
    """

    DATA_TYPE = "invalid_credential#1.0.0"

    def __init__(self, user_credential_id: StringOrUUID) -> None:
        if not isinstance(user_credential_id, UUID):
            self.user_credential_id = UUID(user_credential_id)
        else:
            self.user_credential_id = user_credential_id


class InvalidCredentialError(ForbiddenError):
    """
    ForbiddenError with information about a credential which would have allowed
    data access, but is requires some user intervention to repair the credential.
    """

    error_description = (
        "Identity set maps to a credential, but it is in an invalid state "
        "that requires user action to repair it"
    )
    error = "invalid_credential"

    def __init__(
        self,
        *args,
        user_credential_id: StringOrUUID | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None and user_credential_id:
            detail = InvalidCredential(user_credential_id=user_credential_id)

        if detail:
            kwargs["detail"] = detail

        super().__init__(*args, **kwargs)


class CredentialNotFound(ErrorDetail):
    """
    Error details when a user has attempted to use a credential when creating a
    collection or logging in, but there are multiple mapped identities and none
    of them have a valid credential.
    """

    DATA_TYPE = "credential_not_found#1.0.0"

    def __init__(
        self,
        accounts: list[Account] | None = None,
    ) -> None:
        if not accounts:
            accounts = []
        self.accounts = accounts.copy()


class CredentialNotFoundError(ForbiddenError):
    """
    ForbiddenError with information about the list of account mappings which
    don't have any credential information associated with them.
    """

    error_description = (
        "Identity set maps to one or more accounts, but none are associated "
        "with a valid credential"
    )

    def __init__(
        self,
        *args,
        accounts: list[Account] | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None:
            detail = CredentialNotFound(accounts=accounts)

        if detail:
            kwargs["detail"] = detail

        super().__init__(*args, **kwargs)


class InvalidUser(ErrorDetail):
    """
    Error details implementing invalid_user#1.0.0
    """

    DATA_TYPE = "invalid_user#1.0.0"

    def __init__(self, usernames: list[str] | None = None) -> None:
        if usernames:
            self.usernames = usernames


class InvalidUserError(ForbiddenError):
    """
    ForbiddenError with information that indicates that the caller has an
    identity from an allowed domain but it does not map to a valid username for
    a connector.
    """

    error_description = (
        "Identity set contains an identity from an allowed domain, but it "
        "does not map to a valid username for this connector"
    )

    def __init__(
        self,
        *args,
        usernames: list[str] | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None:
            detail = InvalidUser(usernames=usernames)

        if detail:
            kwargs["detail"] = detail

        super().__init__(*args, **kwargs)


class SubscriptionRequired(ErrorDetail):
    """
    Error details implementing subscription_required#1.0.0
    """

    DATA_TYPE = "subscription_required#1.0.0"

    def __init__(self, subscription_level=None, add_ons=None) -> None:
        self.subscription_level = subscription_level or "standard"
        self.add_ons = add_ons or []


class SubscriptionRequiredError(ForbiddenError):
    """
    ForbiddenError with information about the required subscription features
    to perform the requested operation.
    """

    error = "subscription_required"
    error_description = (
        "This feature requires the endpoint to be managed with a Globus subscription"
    )

    def __init__(
        self, *args, subscription_level=None, add_ons=None, detail=None, **kwargs
    ) -> None:
        if detail is None:
            detail = SubscriptionRequired(
                subscription_level=subscription_level, add_ons=add_ons
            )
        kwargs["detail"] = detail
        super().__init__(*args, **kwargs)


class Batch(ErrorDetail):
    """
    Error details implementing batch#1.0.0
    """

    DATA_TYPE = "batch#1.0.0"

    def __init__(self, ids: list[UUID]) -> None:
        self.ids = ids


class InvalidBatchItemsError(ForbiddenError):
    """
    ForbiddenError with information about the batch items which could not be
    processed
    """

    error_description = (
        "One or more items in the batch were not valid for this operation"
    )

    def __init__(self, *args, ids=None, detail=None, **kwargs) -> None:
        if detail is None and ids:
            detail = Batch(ids)
        kwargs["detail"] = detail
        super().__init__(*args, **kwargs)


def AuthorizationError(*args, **kwargs) -> ForbiddenError:
    detail: ErrorDetail | None = kwargs.get("detail")

    if isinstance(detail, AuthenticationTimeout):
        return AuthenticationTimeoutError(*args, **kwargs)
    elif isinstance(detail, CredentialNotFound):
        return CredentialNotFoundError(*args, **kwargs)
    elif isinstance(detail, IdNotInIdentitySet):
        return IdNotInIdentitySetError(*args, **kwargs)
    elif isinstance(detail, InvalidCredential):
        return InvalidCredentialError(*args, **kwargs)
    elif isinstance(detail, InvalidUser):
        return InvalidUserError(*args, **kwargs)
    elif isinstance(detail, MissingRequiredRole):
        return MissingRequiredRoleError(*args, **kwargs)
    elif isinstance(detail, NotFromAllowedDomain):
        return NotFromAllowedDomainError(*args, **kwargs)
    elif isinstance(detail, NotPublic):
        return NotPublicError(*args, **kwargs)
    elif isinstance(detail, NotResourceOwner):
        return NotResourceOwnerError(*args, **kwargs)
    elif isinstance(detail, SubscriptionRequired):
        return SubscriptionRequiredError(*args, **kwargs)
    else:
        return ForbiddenError(*args, **kwargs)
