from __future__ import annotations

import typing as t

import marshmallow as ma

import globus_sdk


class ErrorDetail:
    DATA_TYPE: str


class APIBaseError(Exception):
    status_code: int | None
    error: str | None
    error_description: str | None = None

    def __init__(
        self,
        error_description: str | None = None,
        error: str | None = None,
        status_code: int = 0,
        errors: list[t.Any] | None = None,
        detail: ErrorDetail | str | None = None,
    ) -> None:
        if error_description is not None:
            self.error_description = error_description

        Exception.__init__(self, self.error_description)

        if status_code:
            self.status_code = status_code
        if error is not None:
            self.error = error

        self.errors = errors or []
        self.detail = detail or ""

    def to_dict(self):
        return {
            "error": self.error,
            "error_description": self.error_description,
            "errors": self.errors,
            "detail": self.detail,
        }


class BadRequestError(APIBaseError):
    status_code = 400
    error = "bad_request"
    error_description = "Bad request"


class InternalServerError(APIBaseError):
    status_code = 500
    error = "internal_server_error"


class UnauthorizedError(APIBaseError):
    status_code = 401
    error = "not_authorized"
    error_description = "Missing or invalid authorization header"


class ForbiddenError(APIBaseError):
    status_code = 403
    error = "permission_denied"


class MissingRequiredScopes(ErrorDetail):
    """
    Error details implementing missing_required_scopes#1.0.0
    """

    DATA_TYPE = "missing_required_scopes#1.0.0"

    def __init__(self, scopes: t.Iterable[str | globus_sdk.MutableScope]) -> None:
        self.required_scopes: list[str] = [str(s) for s in scopes]


class MissingRequiredScopesError(UnauthorizedError):
    error_description = (
        "This operation requires a token with one or more additional OAuth scopes."
    )

    def __init__(
        self,
        *args,
        scopes: t.Iterable[str | globus_sdk.MutableScope] | None = None,
        detail: ErrorDetail | None = None,
        **kwargs,
    ) -> None:
        if detail is None and scopes:
            detail = MissingRequiredScopes(scopes)
        kwargs["detail"] = detail
        super().__init__(*args, **kwargs)
        if isinstance(self.detail, MissingRequiredScopes):
            self.error_description = (
                "This operation requires a token with the following scopes: "
                f"{self.detail.required_scopes}"
            )
            scopes = self.detail.required_scopes
        self.authorization_parameters: dict[str, t.Any] = {
            "session_message": self.error_description,
        }
        if scopes:
            self.authorization_parameters["required_scopes"] = scopes


class NotFoundError(APIBaseError):
    status_code = 404
    error = "not_found"


class RouteNotFoundError(NotFoundError):
    error_description = (
        "The route is not known. Either your client is misconfigured or the "
        "service requires an update to support that route."
    )


class DataTypeVersionError(APIBaseError):
    status_code = 422
    error = "unsupported_datatype_version"
    error_description = (
        "The endpoint does not recognize the document version. Either your "
        "client is misconfigured or the service requires an update to support "
        "it."
    )


class MethodNotAllowedError(APIBaseError):
    status_code = 405
    error = "method_not_allowed"


class UnsupportedMediaError(APIBaseError):
    status_code = 415
    error = "unsupported_media_type"
    error_description = "Request is in an invalid format"


class InvalidInput(ErrorDetail):
    """
    Error details for implementing invalid_input#1.0.0"
    """

    DATA_TYPE = "invalid_input#1.0.0"

    def __init__(self, *, exc: ma.ValidationError) -> None:
        assert isinstance(exc, ma.ValidationError)
        self.errors = []

        if isinstance(exc.messages, list):
            self.errors.extend([{"message": m} for m in exc.messages])
        elif isinstance(exc.messages, dict):
            for p, m in exc.messages.items():
                if not isinstance(m, list):
                    mlist = [m]
                else:
                    mlist = m

                for mi in mlist:
                    message = {"message": mi}
                    if p != "_schema":
                        message["property"] = p
                    self.errors.append(message)
        else:
            raise InternalServerError()

    def __str__(self) -> str:
        invalid_properties: dict[str, list[str]] = {}
        general_errors = []
        for e in self.errors:
            prop = e.get("property")
            message = e.get("message")

            if prop is not None:
                if prop not in invalid_properties:
                    invalid_properties[prop] = []
                invalid_properties[prop].append(f"{message}")
            else:
                general_errors.append(message)
        result = ["Data Validation Error."]
        for message in general_errors:
            result.append(f"{message}")
        for prop, messages in invalid_properties.items():
            if len(messages) == 1:
                result.append(f'Error processing the "{prop}" property: {messages[0]}')
            else:
                result.append(f'Errors processing the "{prop}" property: {messages}')

        return " ".join(result)


class ValidationError(APIBaseError):
    status_code = 422
    error = "unprocessable_entity"
    error_description = "Data Validation Error"

    def __init__(
        self,
        error_description: str | None = None,
        detail: ErrorDetail | None = None,
        exc: ma.ValidationError | None = None,
        **kwargs,
    ) -> None:
        if detail is None and isinstance(exc, ma.ValidationError):
            detail = InvalidInput(exc=exc)

        if error_description is None and isinstance(detail, InvalidInput):
            error_description = str(detail)

        if error_description is None and exc:
            error_description = str(exc)

        super().__init__(error_description=error_description, detail=detail, **kwargs)


class NotImplementedError(APIBaseError):
    status_code = 501
    error = "not_implemented"
    error_description = (
        "The server does not support the functionality "
        "required to fulfill this request"
    )


class BadGatewayError(APIBaseError):
    status_code = 502
    error = "bad_gateway"


class ServiceUnavailableError(APIBaseError):
    status_code = 503
    error = "service_unavailable"


class GatewayTimeout(APIBaseError):
    status_code = 504
    error = "gateway_timeout"


class CredentialDeletedError(BadRequestError):
    error = "credential_deleted"


class ResourceConflict(ErrorDetail):
    DATA_TYPE = "resource_conflict#1.0.0"

    def __init__(self, resource_ids: list[str] | None = None) -> None:
        if resource_ids:
            self.resources = resource_ids.copy()


class ConflictError(APIBaseError):
    status_code = 409
    error = "conflict"

    def __init__(
        self,
        error_description: str | None = None,
        detail: ErrorDetail | None = None,
        resource_ids: list[str] | None = None,
        **kwargs,
    ) -> None:
        if detail is None and resource_ids:
            detail = ResourceConflict(resource_ids=resource_ids)

        super().__init__(error_description=error_description, detail=detail, **kwargs)


class ResourceExistsError(ConflictError):
    error = "exists"
    error_description = "Attempt to create a resource that already exists."


class ResourceInUseError(ConflictError):
    error_description = "Attempt to delete a resource that is in use."


class LimitExceeded(ErrorDetail):
    """
    Error details for implementing limit_exceeded#1.0.0"
    """

    DATA_TYPE = "limit_exceeded#1.0.0"


class LimitExceededError(ForbiddenError):
    """
    ForbiddenError with information about a resource limit which was hit
    attempting to process this request.
    """

    error_description = "Unable to process this request due to resource limits"

    def __init__(self, *args, detail: ErrorDetail | None = None, **kwargs) -> None:
        if detail is None:
            detail = LimitExceeded()
        super().__init__(*args, detail=detail, **kwargs)  # type: ignore # mypy issue 4335
