from __future__ import annotations

import datetime

import requests
from flask import current_app

import globus.manager.api.errors as errors
from globus.manager.constants import ConfigurationObjectTag

from .oidc_user_credential import OIDCUserCredential


class AzureBlobUserCredential(OIDCUserCredential):
    POLICY_DATA_TYPE = "azure_blob_user_credential_policies#1.0.0"
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]
    SERVICE_SCOPES = "https://storage.azure.com/.default"
    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name
            + "9436da0c-a444-11eb-af93-12704e0d6a4d"
        )
    }

    @property
    def connector_properties(self) -> list[str]:
        return super().connector_properties + ["tid"]

    def _credentials_to_patch(self, credentials):
        patch = super()._credentials_to_patch(credentials)
        if credentials:
            id_token = credentials.id_token
            if "policies" not in patch:
                patch["policies"] = {"DATA_TYPE": self.POLICY_DATA_TYPE}
            if "tid" in id_token:
                patch["policies"]["tid"] = id_token["tid"]
        return patch

    def _get_service_access_token(self, client_id, secret):
        try:
            data = {
                "grant_type": "client_credentials",
                "scope": self.SERVICE_SCOPES,
                "client_id": client_id,
                "client_secret": secret,
            }
            current_app.logger.debug("Token uri: {}".format(self.token_uri))
            response = requests.post(self.token_uri, data=data).json()
            access_token = response.get("access_token")
            expires_in = response.get("expires_in", 0)
            delta = datetime.timedelta(seconds=int(expires_in))
            token_expiry = delta + datetime.datetime.utcnow()

        except Exception as e:
            current_app.logger.error(
                "Unable to make token request. Exception: {}".format(e)
            )
            raise errors.BadGatewayError(
                error_description="Unable to obtain an access token. Contact the endpoint administrator."
            )

        if access_token is None:
            current_app.logger.error(
                "Response has no access token: {}".format(response)
            )
            raise errors.BadGatewayError(
                error_description="Unable to obtain an access token. Contact the endpoint administrator."
            )

        return access_token, token_expiry

    def __init__(
        self, appsync_data=None, doc=None, storage_gateway=None, credentials=None
    ):
        if doc:
            self.username = doc["username"]
            new_doc = {**doc}
            new_doc.pop("display_name", None)
            # Policies must be either empty (for an unprovisioned credential)
            # or passed in as a credentials object from the microsoft callback
            policies = new_doc.get("policies", None)
            if policies is not None and policies != {
                "DATA_TYPE": self.POLICY_DATA_TYPE
            }:
                raise ValueError("policies")
            doc = new_doc

        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )

        if doc and self.uses_service_auth:
            client_id = self.storage_gateway.client_id
            secret = self.storage_gateway.secret
            access_token, expiry = self._get_service_access_token(client_id, secret)
            p = doc.get("policies", {})
            policies = {
                **p,
                **{
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                    "access_token": access_token,
                    "token_expiry": expiry,
                    "refresh_token": "service",
                    "tid": getattr(self.storage_gateway, "tenant", "common"),
                    "sub": client_id,
                    "email": self.username,
                    "scopes": self.SERVICE_SCOPES,
                },
            }
            doc = {
                **doc,
                **{
                    "display_name": client_id,
                    "invalid": False,
                    "deleted": False,
                    "provisioned": True,
                    "policies": policies,
                },
            }
            self._patch(**doc)
        elif doc:
            patch = self._credentials_to_patch(credentials)
            patch["deleted"] = False
            patch["provisioned"] = bool(credentials)
            self._patch(**patch)

    tid = OIDCUserCredential.connector_property("tid")

    @property
    def connector_id(self):
        return "9436da0c-a444-11eb-af93-12704e0d6a4d"

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def token_uri(self) -> str:
        return "https://login.microsoftonline.com/{}/oauth2/v2.0/token".format(
            getattr(self.storage_gateway, "tenant", "common")
        )

    @property
    def supports_revoke(self) -> bool:
        return False

    @property
    def uses_service_auth(self) -> bool:
        return not self.storage_gateway.user_credential_required

    def refresh(self):
        if self.uses_service_auth:
            client_id = self.storage_gateway.client_id
            secret = self.storage_gateway.secret
            access_token, expiry = self._get_service_access_token(client_id, secret)
            self.refresh_token = "service"
            self.access_token = access_token
            self.token_expiry = expiry
            self.tid = getattr(self.storage_gateway, "tenant", "common")
            self.sub = client_id
            self.email = self.username
            self.scopes = self.SERVICE_SCOPES
            self.invalid = False
            self.provisioned = True
            self.configuration_hash.changed()
            current_app.logger.debug("Token refreshed: {}".format(self.token_expiry))
        else:
            super().refresh()

    def __repr__(self):
        return "<AzureBlobUserCredential {}>".format(self.id)
