from __future__ import annotations

import ipaddress
import time
import typing as t

import dns.resolver
from flask import current_app
from sqlalchemy import orm

import globus_sdk

import globus.manager.api.errors as errors
import globus.manager.api.models as m
from globus.manager.api.schemas.appsync import EndpointDomain as endpoint_domain_schema
from globus.manager.constants import ConfigurationObjectTag, NodeStatus

from .configuration_object import ConfigurationObject


class GlobusDnsDomain(ConfigurationObject):
    __mapper_args__ = {
        "polymorphic_identity": ConfigurationObjectTag.WILDCARD_DOMAIN.name
    }
    _DATA_TYPE = endpoint_domain_schema._DATA_TYPE
    configuration_schema = endpoint_domain_schema

    def __init__(self, appsync_data=None) -> None:
        self._nodes: dict[str, m.Node] = {}

        if appsync_data is not None:
            args = {k: v for k, v in appsync_data.items()}
            args["id"] = ConfigurationObjectTag.WILDCARD_DOMAIN.name
            args["tag"] = ConfigurationObjectTag.WILDCARD_DOMAIN
            super().__init__(**args)
            self.init_on_load()
        else:
            super().__init__(
                id=ConfigurationObjectTag.WILDCARD_DOMAIN.name,
                tag=ConfigurationObjectTag.WILDCARD_DOMAIN,
            )
            self.register_domain()
            self.refresh_configuration_hash()

    @orm.reconstructor
    def init_on_load(self):
        ch = self.configuration_hash
        if ch:
            self._nodes = {n["id"]: m.Node(**n) for n in ch.get("nodes", [])}

    def _check_status(self):
        node_id = m.Configuration.query.filter(
            m.Configuration.name == "node_id"
        ).one_or_none()

        if node_id:
            node = self._nodes.get(node_id.value)
            if node is not None and node.status != current_app.node_status:
                current_app.node_status = node.status

    def register_domain(self):
        RETRIES = 15

        auth_client = current_app.auth_client
        dns_client = current_app.dns_client

        parsed_addresses = self.ip_addresses(active_only=True)

        a = list(
            [str(p) for p in parsed_addresses if isinstance(p, ipaddress.IPv4Address)]
        )
        aaaa = list(
            [str(p) for p in parsed_addresses if isinstance(p, ipaddress.IPv6Address)]
        )
        # Order so testing can assert more easily
        a.sort()
        aaaa.sort()

        domain_response = dns_client.create_domain(a=a, aaaa=aaaa)
        domain = domain_response.data["domain"]

        current_app.logger.debug("Waiting for domain propagation")
        for attempt in range(1, RETRIES):
            time.sleep(30)
            try:
                # Verify DNS has propagated
                current_app.logger.debug("Attempt {0}".format(attempt))
                dns.resolver.query("client_id." + domain, "TXT")
                break
            except dns.resolver.NXDOMAIN as e:
                if attempt + 1 < RETRIES:
                    current_app.logger.warning(
                        "DNS for {} not resolving, sleeping and trying again".format(
                            domain
                        )
                    )
                else:
                    raise e

            current_app.logger.debug(
                "DNS Propagated to us: registering domain with auth"
            )
            for attempt in range(1, RETRIES):
                try:
                    auth_response = auth_client.post(
                        "/v2/api/clients/{client_id}/fqdns".format(
                            client_id=auth_client.client_id
                        ),
                        data={"fqdn": domain},
                    )
                    current_app.logger.debug(
                        "Registered dns with auth returned {0}".format(
                            auth_response.http_status
                        )
                    )
                    assert auth_response.http_status == 201
                except globus_sdk.GlobusAPIError as e:
                    current_app.logger.debug(
                        "Error registering domain: {0} {1}".format(
                            e.http_status, str(e)
                        )
                    )
                    time.sleep(30)
                except Exception as e:
                    current_app.logger.debug("Error registering domain: {0}".format(e))
                    if attempt + 1 < RETRIES:
                        current_app.logger.debug(str(e))
                        time.sleep(30)
                    else:
                        raise e
        self.configuration_hash["domain_name"] = domain
        return domain

    def ip_addresses(
        self,
        active_only: bool = True,
    ) -> list[ipaddress.IPv4Address | ipaddress.IPv6Address]:
        return [
            ip
            for n in self._nodes.values()
            for ip in n.ip_addresses
            if n.status == NodeStatus.active or not active_only
        ]

    def update_domain(self):
        if not self.domain_name:
            raise RuntimeError("No domain configured for update")

        dns_client = current_app.dns_client

        parsed_addresses = self.ip_addresses(active_only=True)
        a = list(
            [str(p) for p in parsed_addresses if isinstance(p, ipaddress.IPv4Address)]
        )
        aaaa = list(
            [str(p) for p in parsed_addresses if isinstance(p, ipaddress.IPv6Address)]
        )

        a.sort()
        aaaa.sort()

        dns_client.update_domain(self.domain_name, a=a, aaaa=aaaa)

    def delete(self):
        if hasattr(self, "domain_name"):
            try:
                dns_client = current_app.dns_client
                dns_client.delete_domain(self.domain_name)
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.error("delete_domain: {}".format(str(e)))
                if e.http_status != 404:
                    raise errors.BadGatewayError("Error updating DNS")
            self.configuration_hash["domain_name"] = None
            self._nodes = {}
        super().delete()

    @property
    def nodes(self) -> list[m.Node]:
        if self._nodes:
            return list(self._nodes.values())
        else:
            return []

    @nodes.setter
    def nodes(self, value: list[m.Node | dict]):
        nodes = {}

        if value:
            for n in value:
                if not isinstance(n, m.Node):
                    node = m.Node(**n)
                else:
                    node = n
                nodes[node.id] = node

        self._nodes = nodes
        try:
            if hasattr(self, "domain_name"):
                self.update_domain()
            else:
                self.register_domain()
        finally:
            pass
        self.refresh_configuration_hash()

    @property
    def domain_name(self) -> str:
        value = self.configuration_hash.get("domain_name")
        if not value:
            raise AttributeError(
                "The domain_name attribute is not defined for this object"
            )
        return value

    @domain_name.setter
    def domain_name(self, value: str):
        self.configuration_hash["domain_name"] = value.lower()

    def add_node(self, node: m.Node | dict):
        if not isinstance(node, m.Node):
            node = m.Node(**node)

        if node.id in self._nodes:
            raise ValueError("Node already exists")

        ip_addresses = self.ip_addresses(active_only=False)
        for ip in node.ip_addresses:
            if ip in ip_addresses:
                raise ValueError("Node already exists")

        self._nodes[node.id] = node
        try:
            if hasattr(self, "domain_name"):
                self.update_domain()
            else:
                self.register_domain()
            self.refresh_configuration_hash()
        except globus_sdk.GlobusAPIError as e:
            current_app.logger.error("add_node: {}: {}".format(node.id, str(e)))
            del self._nodes[node.id]
            self.refresh_configuration_hash()
            raise errors.BadGatewayError("Error updating DNS")

    def get_node(
        self,
        id: t.Optional[str] = None,
        ip_address: None | str | ipaddress.IPv4Address | ipaddress.IPv6Address = None,
    ) -> t.Optional[m.Node]:
        potential_nodes = [n for n in self._nodes.values()]
        if ip_address:
            potential_nodes = [
                n
                for n in potential_nodes
                if ipaddress.ip_address(ip_address) in n.ip_addresses
            ]

        if id:
            potential_nodes = [n for n in potential_nodes if n.id == id]

        if len(potential_nodes) > 1:
            raise ValueError("Multiple nodes match")

        elif len(potential_nodes) == 1:
            return potential_nodes.pop()
        else:
            return None

    def put_node(self, new_node: m.Node | dict):
        if not isinstance(new_node, m.Node):
            node = None
            node_id_str = str(new_node["id"])
            node = self._nodes.get(node_id_str)
        else:
            node_id_str = new_node.id
            node = self._nodes.get(new_node.id)

        if node is None:
            raise errors.NotFoundError("Unknown node {}".format(node_id_str))

        if node.put(new_node):
            if hasattr(self, "domain_name"):
                self.update_domain()
            else:
                self.register_domain()
            self.refresh_configuration_hash()
        self._check_status()

    def patch_node(self, node_changes: dict):
        node_id_str = str(node_changes["id"])
        node = self._nodes.get(node_id_str)

        if node is None:
            raise errors.NotFoundError("Unknown node {}".format(node_id_str))

        if node.patch(node_changes):
            if hasattr(self, "domain_name"):
                self.update_domain()
            else:
                self.register_domain()
            self.refresh_configuration_hash()
        self._check_status()

    def remove_node(
        self,
        id: t.Optional[str] = None,
        node: t.Optional[m.Node] = None,
        ip_address: None | str | ipaddress.IPv4Address | ipaddress.IPv6Address = None,
    ):
        potential_nodes = list(self._nodes.values())
        if id:
            potential_nodes = [n for n in potential_nodes if n.id == id]
        if node:
            potential_nodes = [n for n in potential_nodes if n.id == node.id]
        if ip_address:
            potential_nodes = [
                n
                for n in potential_nodes
                if ipaddress.ip_address(ip_address) in n.ip_addresses
            ]

        if len(potential_nodes) == 0:
            raise ValueError("Node does not exist")
        elif len(potential_nodes) > 1:
            raise ValueError("Multiple nodes match")
        else:
            old_node = potential_nodes[0]

        self._nodes.pop(old_node.id)

        if self.domain_name:
            self.update_domain()
            self.refresh_configuration_hash()

    def rollback(self, data=None):
        """
        Rolls back state to a previous version of the configuration_hash,
        including updating DNS. This will fail if the domain name has been
        deleted.
        """
        if data:
            self.update(data["items"][0])
        self.init_on_load()

        if hasattr(self, "domain_name"):
            try:
                self.update_domain()
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.error("rollback: {}".format(str(e)))
                raise errors.BadGatewayError("Error updating DNS")

    def __repr__(self):
        return "<Domain {}>".format(getattr(self, "domain_name", None))
