from __future__ import annotations

import typing as t

from flask import current_app

from globus.manager.constants import ConfigurationObjectTag

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway


class GoogleCloudStorageGateway(StorageGateway):
    CONNECTOR_ID = "56366b96-ac98-11e9-abac-9cb6d0d9fd63"
    POLICY_DATA_TYPE = "google_cloud_storage_policies#1.1.0"
    CONNECTOR_POLICIES = {
        "google_cloud_storage_policies#1.0.0": [
            "client_id",
            "secret",
            "buckets",
            "projects",
            "service_account_key",
        ]
    }
    CONNECTOR_POLICIES["google_cloud_storage_policies#1.1.0"] = CONNECTOR_POLICIES[
        "google_cloud_storage_policies#1.0.0"
    ] + ["allow_any_account"]

    POLICY_DATA_TYPES = list(CONNECTOR_POLICIES.keys())
    connector_policy_names = CONNECTOR_POLICIES[POLICY_DATA_TYPE]

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }

    connector = Connector(
        id=CONNECTOR_ID, display_name="Google Cloud Storage", can_post_credential=False
    )

    client_id = StorageGateway.connector_property("client_id", required=True)

    @checks_policy_property(prop="client_id", required=True)
    def check_client_id(self, value: str):
        if not isinstance(value, str):
            raise ValueError("Invalid client_id")

    secret = StorageGateway.connector_property("secret", required=True)

    @checks_policy_property(prop="secret", required=True)
    def checks_secret(self, value: str):
        if not isinstance(value, str):
            raise ValueError("Invalid secret")

    service_account_key = StorageGateway.connector_property("service_account_key")

    @checks_policy_property(prop="service_account_key")
    def checks_service_account_key(self, value: dict):
        if not isinstance(value, dict):
            raise ValueError("Invalid service_account_key")

    buckets = StorageGateway.connector_property("buckets")

    @checks_policy_property(prop="buckets")
    def check_buckets(self, value: t.Optional[list[str]]):
        if value:
            if not isinstance(value, t.Sequence) or isinstance(value, str):
                raise ValueError()
            for v in value:
                if not isinstance(v, str):
                    raise ValueError()

    projects = StorageGateway.connector_property("projects")

    @checks_policy_property(prop="projects")
    def check_projects(self, value: t.Optional[list[str]]):
        if value:
            if not isinstance(value, t.Sequence) or isinstance(value, str):
                raise ValueError("Invalid projects")
            for v in value:
                if not isinstance(v, str):
                    raise ValueError("Invalid projects")

    allow_any_account = StorageGateway.connector_property("allow_any_account")

    @checks_policy_property(prop="allow_any_account")
    def check_allow_any_account(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError("Invalid allow_any_account")

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    def check_user_allowed(self, username):
        if self.client_id is None:
            current_app.logger.warning(
                "storage_gateway=%s client_id is not configured", self
            )
            return False

        if self.secret is None:
            current_app.logger.warning(
                "storage_gateway=%s secret is not configured", self
            )
            return False

        return super().check_user_allowed(username)

    @property
    def user_credential_required(self):
        return getattr(self, "service_account_key", None) is None

    @property  # type: ignore
    def load_dsi_module(self) -> str:
        return "google_cloud"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        pass

    @property
    def auth_callback(self):
        """
        URL of the auth callback that must be registered on the Google API
        console for the application client_id in order to process Google
        credentials.
        """
        try:
            return current_app.config["connectors"][self.CONNECTOR_ID].callback_url
        except Exception:
            raise AttributeError(
                "This storage gateway does not have a value for the auth_callback property"
            )

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        pass
