from __future__ import annotations

import pwd
import typing as t

from flask import current_app

from globus.manager.constants import ConfigurationObjectTag

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway

if t.TYPE_CHECKING:
    from collections import abc

    from globus.manager.api.connectors.hpss import HPSSConnector


def default_value(prop: property, default_value: t.Any):
    """
    Wrap a property's fget to return a default value if the value is None or if
    it raises AttributeError.
    """

    class _DefaultValue(property):
        """
        Property subclass that hooks __get__ to replace None and AttributeError with
        default_value.
        """

        def __get__(self, instance, owner=None):
            try:
                value = super().__get__(instance, owner)
                return value if value is not None else default_value
            except AttributeError:
                return default_value

    return _DefaultValue(fget=prop.fget, fset=prop.fset, fdel=prop.fdel)


class HPSSStorageGateway(StorageGateway):
    CONNECTOR_ID = "fb656a17-0f69-4e59-95ff-d0a62ca7bdf5"

    # Current policy data_type, used for new docs and dumping
    POLICY_DATA_TYPE = "hpss_storage_policies#1.1.0"

    CONNECTOR_POLICIES = {}
    CONNECTOR_POLICIES["hpss_storage_policies#1.0.0"] = [
        "authentication_mech",
        "authenticator",
        "uda_checksum",
    ]
    CONNECTOR_POLICIES["hpss_storage_policies#1.1.0"] = [
        *CONNECTOR_POLICIES["hpss_storage_policies#1.0.0"],
        "login_name",
    ]

    # Supported policy data_types, current and legacy
    POLICY_DATA_TYPES = list(CONNECTOR_POLICIES.keys())
    # Our properties used by StorageGateway to validate input
    connector_policy_names = CONNECTOR_POLICIES[POLICY_DATA_TYPE]

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }

    connector = Connector(id=CONNECTOR_ID, display_name="HPSS")

    authentication_mech = StorageGateway.connector_property(
        "authentication_mech", required=True
    )

    @checks_policy_property(prop="authentication_mech", required=True)
    def _check_authentication_mech(self, value: str) -> None:
        if value not in ["unix", "krb5"]:
            raise ValueError()

    authenticator = StorageGateway.connector_property("authenticator", required=True)

    @checks_policy_property(prop="authenticator", required=True)
    def _check_authenticator(self, value) -> None:
        if not isinstance(value, (bytes, str)):
            raise ValueError()

    uda_checksum = StorageGateway.connector_property("uda_checksum", required=True)

    @checks_policy_property(prop="uda_checksum", required=True)
    def _check_uda_checksum(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError()

    login_name = StorageGateway.connector_property("login_name")
    login_name = default_value(login_name, "hpssftp")

    @checks_policy_property(prop="login_name")
    def _check_login_name(self, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError()

    def default_mapping(self, identity_data: abc.Mapping[str, t.Any]) -> str | None:
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return user
        return None

    def home_directory(self, username: str) -> str | None:
        """
        Look up the user's home directory by calling a helper function built into
        the HPSS DSI. Used for expansion of '~' and  $HOME in path restrictions. The
        call into the DSI is handled by HPSSConnector.
        """

        # Lookup the HPSSConnector instance
        hpss_connector: HPSSConnector = current_app.config["connectors"][
            self.CONNECTOR_ID
        ]

        # Lookup the user's home directory and cache it
        return hpss_connector.home_directory(
            self.authentication_mech,
            self.authenticator,
            self.login_name,
            username,
        )

    def lookup_unix_info(self, username: str) -> pwd.struct_passwd | None:
        """
        Helper method to resolve the given username to a passwd entry.

        Although we rely on the HPSS API for the user's home directory,
        everything else can come from the system's accounts. This is
        possible because we have always required HPSS sites to have user
        accounts in /etc/passwd / nis / ldap for use with GridFTP server
        gfs_stat_t translations (back when we cared about that).
        """

        try:
            passwd_entry: pwd.struct_passwd | None = pwd.getpwnam(username)
        except KeyError:
            passwd_entry = None

        if not passwd_entry:
            current_app.logger.info(
                "lookup_unix_info: no local account for user=%s", username
            )

        return passwd_entry

    def get_user_display_name(self, username: str) -> str:
        """
        Helper used to standardize display names for this connector.
        """
        pwd = self.lookup_unix_info(username)

        if pwd:
            return pwd.pw_gecos.split(",")[0] or username
        else:
            return username

    @property  # type: ignore
    def load_dsi_module(self) -> str:
        return "hpss"

    @load_dsi_module.setter
    def load_dsi_module(self, value: str) -> None:
        pass

    @property  # type: ignore
    def process_user(self) -> str:
        return "gcsweb"

    @process_user.setter
    def process_user(self, value: t.Any) -> None:
        pass

    def repr(self) -> str:
        return "<HPSSStorageGateway {}>".format(self.id)
