from __future__ import annotations

import ipaddress
import typing as t
import uuid

from flask import current_app

from globus.manager.constants import NodeStatus


class Node:
    """
    Model for a GCS Node.

    Each Node object in the API consists of:
      - [required] an ID of type UUID
      - [required] one or more IP addresses
      - [optional] a status field of type NodeStatus, defaults to active
      - [optional] incoming port range
      - [optional] outgoing port range
      - [optional] data interface (ipv4 and/or ipv6)
    """

    _DATA_TYPE = "node#1.1.0"
    _DATA_TYPES = ["node#1.0.0", "node#1.1.0", "node#1.2.0"]

    def __init__(
        self,
        *,
        DATA_TYPE: t.Optional[str] = None,
        ip_addresses: t.Iterable[str | ipaddress.IPv4Address | ipaddress.IPv6Address]
        | None = None,
        id: t.Optional[str] = None,
        status: t.Optional[NodeStatus | str] = None,
        incoming_port_range: t.Optional[tuple[int, int]] = None,
        outgoing_port_range: t.Optional[tuple[int, int]] = None,
        data_interface: t.Optional[ipaddress.IPv4Address | str] = None,
        data_interface6: t.Optional[ipaddress.IPv6Address | str] = None,
        **kwargs: t.Any,
    ) -> None:
        if not DATA_TYPE or not DATA_TYPE.startswith("node#1."):
            raise ValueError("Invalid DATA_TYPE for node")
        if DATA_TYPE not in self._DATA_TYPES:
            current_app.logger.warning(
                f"Unsupported DATA_TYPE {DATA_TYPE} with extra properties {kwargs}"
            )
        if id is None:
            self._id = str(uuid.uuid4())
        else:
            self._id = id
        if not ip_addresses:
            raise ValueError("Invalid ip_addresses")
        try:
            self.ip_addresses = [ipaddress.ip_address(i) for i in ip_addresses]
        except ValueError:
            raise ValueError("Invalid ip_addresses") from None

        if not status:
            self.status = NodeStatus.active
        elif isinstance(status, NodeStatus):
            self.status = status
        else:
            try:
                self.status = getattr(NodeStatus, status)
            except AttributeError:
                raise ValueError("Invalid node status {}".format(status))

        self._incoming_port_range: t.Optional[tuple[int, int]] = None
        self._outgoing_port_range: t.Optional[tuple[int, int]] = None
        self._data_interface: t.Optional[ipaddress.IPv4Address] = None
        self._data_interface6: t.Optional[ipaddress.IPv6Address] = None

        if incoming_port_range:
            self.incoming_port_range = incoming_port_range
        if outgoing_port_range:
            self.outgoing_port_range = outgoing_port_range
        if data_interface:
            try:
                di = ipaddress.ip_address(data_interface)
            except ValueError:
                di = None

            if not isinstance(di, ipaddress.IPv4Address):
                raise ValueError("Invalid data_interface")

            self.data_interface = di
        if data_interface6:
            try:
                di = ipaddress.ip_address(data_interface6)
            except ValueError:
                di = None

            if not isinstance(di, ipaddress.IPv6Address):
                raise ValueError("Invalid data_interface")

            self.data_interface6 = di

    @property
    def id(self) -> str:
        return self._id

    @property
    def ip_addresses(self) -> list[ipaddress.IPv4Address | ipaddress.IPv6Address]:
        return self._ip_addresses

    @ip_addresses.setter
    def ip_addresses(
        self, value: list[ipaddress.IPv4Address | ipaddress.IPv6Address]
    ) -> None:
        if value is None or len(value) == 0:
            raise ValueError("Invalid ip_addresses")
        self._ip_addresses = list(value)

    @property
    def incoming_port_range(self) -> tuple[int, int]:
        if not self._incoming_port_range:
            raise AttributeError(
                "The incoming_port_range property is not defined for this node"
            )
        return self._incoming_port_range

    @incoming_port_range.setter
    def incoming_port_range(
        self, value: t.Optional[tuple[int, int] | list[int]]
    ) -> None:
        if value:
            self._incoming_port_range = self._parse_port_range(value)
        else:
            self._incoming_port_range = None

    @property
    def outgoing_port_range(self) -> tuple[int, int]:
        if not self._outgoing_port_range:
            raise AttributeError(
                "The outgoing_port_range property is not defined for this node"
            )
        return self._outgoing_port_range

    @outgoing_port_range.setter
    def outgoing_port_range(
        self, value: t.Optional[tuple[int, int] | list[int]]
    ) -> None:
        if value:
            self._outgoing_port_range = self._parse_port_range(value)
        else:
            self._outgoing_port_range = None

    @property
    def status(self) -> NodeStatus:
        return self._status

    @status.setter
    def status(self, value: str | NodeStatus) -> None:
        self._status = NodeStatus(value)

    @property
    def data_interface(self) -> ipaddress.IPv4Address:
        if not self._data_interface:
            raise AttributeError(
                "The data_interface property is not defined for this node"
            )
        return self._data_interface

    @data_interface.setter
    def data_interface(self, value: ipaddress.IPv4Address | None) -> None:
        self._data_interface = None
        if value is not None:
            self._data_interface = value

    @property
    def data_interface6(self) -> ipaddress.IPv6Address:
        if not self._data_interface6:
            raise AttributeError(
                "The data_interface property is not defined for this node"
            )
        return self._data_interface6

    @data_interface6.setter
    def data_interface6(self, value: ipaddress.IPv6Address | None) -> None:
        self._data_interface6 = None
        if value is not None:
            self._data_interface6 = value

    @property
    def data_interfaces(self) -> str | None:
        res: list[ipaddress.IPv4Address | ipaddress.IPv6Address] = []
        if self._data_interface:
            res.append(self._data_interface)
        if self._data_interface6:
            res.append(self._data_interface6)

        if res:
            return ",".join([str(s) for s in res])
        return None

    def put(self, node: Node | t.Mapping[str, t.Any]) -> bool:
        changed = False

        incoming_port_range: t.Optional[tuple[int, int]] = None
        outgoing_port_range: t.Optional[tuple[int, int]] = None
        data_interface: t.Optional[ipaddress.IPv4Address] = None

        if not isinstance(node, Node):
            if node.get("DATA_TYPE") not in Node._DATA_TYPES:
                raise ValueError("Unknown DATA_TYPE")

            node_id_str = str(node["id"])
            ip_addresses = [ipaddress.ip_address(ip) for ip in node["ip_addresses"]]
            status = NodeStatus(node["status"])
            pr = node.get("incoming_port_range")
            if pr:
                incoming_port_range = self._parse_port_range(pr)
            pr = node.get("outgoing_port_range")
            if pr:
                outgoing_port_range = self._parse_port_range(pr)
            data_interface = node.get("data_interface")
        else:
            node_id_str = str(node.id)
            ip_addresses = [ipaddress.ip_address(ip) for ip in node.ip_addresses]
            status = NodeStatus(node.status)
            incoming_port_range = getattr(node, "incoming_port_range", None)
            outgoing_port_range = getattr(node, "outgoing_port_range", None)
            data_interface = getattr(node, "data_interface", None)

        if node_id_str != self._id:
            raise ValueError("The id property is immutable")

        if sorted(ip_addresses, key=str) != sorted(self._ip_addresses, key=str):
            self.ip_addresses = ip_addresses
            changed = True

        if status != self._status:
            self._status = status
            changed = True

        if incoming_port_range != self._incoming_port_range:
            self._incoming_port_range = incoming_port_range
            changed = True

        if outgoing_port_range != self._outgoing_port_range:
            self._outgoing_port_range = outgoing_port_range
            changed = True

        if data_interface != self._data_interface:
            self._data_interface = data_interface
            changed = True

        return changed

    def patch(self, node_changes: t.Mapping[str, t.Any]) -> bool:
        changed = False
        for n in node_changes:
            if n == "DATA_TYPE":
                if node_changes[n] not in Node._DATA_TYPES:
                    raise ValueError("Unknown DATA_TYPE")
            elif n == "id":
                if str(node_changes[n]) != self.id:
                    raise ValueError("id is an immutable property")
            else:
                old_value = getattr(self, n, None)
                new_value = node_changes[n]
                if old_value != new_value:
                    setattr(self, n, new_value)
                    changed = True
        return changed

    @staticmethod
    def _parse_port_range(value: t.Any) -> tuple[int, int]:
        if (
            isinstance(value, (tuple, list))
            and isinstance(value[0], int)
            and isinstance(value[1], int)
            and len(value) == 2
        ):
            start, end = (value[0], value[1])
            if (
                start < end
                and start > 1
                and start < 2**16 - 1
                and end > 1
                and end < 2**16 - 1
            ):
                return (start, end)
        raise ValueError(f"Invalid port range {value}")

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, Node):
            return False
        return (
            self._id == other._id
            and self._ip_addresses == other._ip_addresses
            and self._status == other._status
            and self._incoming_port_range == other._incoming_port_range
            and self._outgoing_port_range == other._outgoing_port_range
            and self._data_interface == other._data_interface
        )

    def __repr__(self) -> str:
        return "<Node {}>".format(self.id)
