from __future__ import annotations

import grp
import pwd
import typing as t
from collections.abc import Sequence

from flask import current_app

from globus.manager.constants import ConfigurationObjectTag

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway


class PosixStorageGateway(StorageGateway):
    CONNECTOR_ID = "145812c8-decc-41f1-83cf-bb2a85a2a70b"
    POLICY_DATA_TYPE = "posix_storage_policies#1.0.0"
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }

    connector_policy_names = ["groups_allow", "groups_deny"]

    connector = Connector(id=CONNECTOR_ID, display_name="POSIX")

    @staticmethod
    def __validate_string_list(name, value: t.Optional[list[str]]):
        if value:
            if not isinstance(value, Sequence) or isinstance(value, str):
                raise ValueError("Invalid value for the {} property.".format(name))
            for v in value:
                if not isinstance(v, str):
                    raise ValueError("Invalid value for the {} property.".format(name))

    groups_allow = StorageGateway.connector_property("groups_allow")

    @checks_policy_property(prop="groups_allow")
    def _check_groups_allow(self, value: t.Optional[list[str]]) -> None:
        PosixStorageGateway.__validate_string_list("groups_allow", value)

    groups_deny = StorageGateway.connector_property("groups_deny")

    @checks_policy_property(prop="groups_deny")
    def _check_groups_deny(self, value: t.Optional[list[str]]):
        PosixStorageGateway.__validate_string_list("groups_deny", value)

    def __repr__(self):
        return "<PosixStorageGateway {}>".format(self.display_name)

    def home_directory(self, username):
        pwd = self.lookup_unix_info(username)

        if pwd is not None:
            return pwd.pw_dir

    def get_user_display_name(self, username):
        pwd = self.lookup_unix_info(username)

        if pwd:
            return pwd.pw_gecos.split(",")[0] or username
        else:
            return username

    def lookup_unix_info(self, username):
        """
        Helper method to resolve the given username to a passwd entry.
        """

        try:
            passwd_entry = pwd.getpwnam(username)
        except KeyError:
            passwd_entry = None

        if not passwd_entry:
            current_app.logger.info(
                "lookup_unix_info: no local account for user=%s", username
            )

        return passwd_entry

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return user

    def check_user_allowed(self, username):
        return self._check_user_allowed(
            username,
            getattr(self, "users_allow", None),
            getattr(self, "users_deny", None),
            getattr(self, "groups_allow", None),
            getattr(self, "groups_deny", None),
        )

    def _check_user_allowed(
        self, username, users_allow, users_deny, groups_allow, groups_deny
    ):
        """
        Check whether a user is allowed to create or access a collection,
        based on a storage_gateway's allow/deny policies. User policies are
        applied first since they are more specific then group policies.
        """
        if username == "gcsweb":
            return False

        # User must have a local account
        passwd_entry = self.lookup_unix_info(username)
        if not passwd_entry:
            return False

        #
        # check user against allow/deny list.
        #

        # if the deny list includes the user, he is immediately denied.
        if users_deny and username in users_deny:
            return False

        # if the allow list includes the user, he is immediately allowed.
        if users_allow and username in users_allow:
            return True

        if not users_allow:
            # if the allow list is unset, he is allowed pending the group check.
            implied_user_response = True
        else:
            # if the allow list is set and does not include the user, he is denied
            # pending the group check.
            implied_user_response = False

        #
        # check groups against allow/deny list.
        #

        try:
            primary_group = grp.getgrgid(passwd_entry.pw_gid)
            primary_group_name = primary_group.gr_name
        except KeyError:
            primary_group_name = None

        # if the deny list includes a group it is immediately denied.
        if groups_deny:
            for g in groups_deny:
                if primary_group_name and primary_group_name == g:
                    return False
                try:
                    gr = grp.getgrnam(g)
                    members = gr.gr_mem
                except KeyError:
                    members = []

                if username in members:
                    return False

        # if the allow list includes a group, it is allowed.
        if groups_allow:
            for g in groups_allow:
                if primary_group_name and primary_group_name == g:
                    return True
                try:
                    gr = grp.getgrnam(g)
                    members = gr.gr_mem
                except KeyError:
                    members = []

                if username in members:
                    return True

        # if the allow list is set and does not include a group, it is denied.
        if groups_allow and len(groups_allow) > 0:
            return False

        # if the allow list is unset, the pending user list result stands.
        return implied_user_response
