import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)

DROPBOX_CONNECTOR_ID = "49b00fd6-63f1-48ae-b27f-d8af4589f876"


@storage_gateway_policy("dropbox_storage_policies#1.0.0")
class DropboxStoragePolicies_1_0_0(PolicySchema):
    """
    Connector-specific storage gateway policies for the Dropbox connector
    """

    client_id = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Client ID (App key) of the app created in the Dropbox App Console
        """,
    )
    secret = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        App secret of the app from the Dropbox App Console
        policy.
        """,
    )
    user_api_rate_limit = ma.fields.Int(
        required=False,
        description="""
        User API Rate Limit associated with this client ID in operations per
        second per user.
        """,
    )
    allow_any_account = ma.fields.Bool(
        required=False,
        load_default=False,
        description="""
        If true, allow users to access personal or external Dropbox accounts.
        If false (the default), users must use the Dropbox account which
        matches the username their Globus credential maps to.
        """,
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be registered on the Dropbox App
        Console for the associated client_id in order to process Dropbox
        credentials.
        """,
    )

    private = ["client_id", "secret", "user_api_rate_limit"]
    connector_id = DROPBOX_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


@collection_policy("dropbox_collection_policies#1.0.0")
class DropboxCollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the Dropbox connector
    """

    connector_id = DROPBOX_CONNECTOR_ID


@user_credential_policy("dropbox_user_credential_policies#1.0.0")
class DropboxUserCredentialPolicies_1_0_0(PolicySchema):
    """
    Connector-specific user credential policies for the Dropbox connector
    """

    sub = ma.fields.Str(allow_none=True, description="OAuth subject identifier claim")
    email = ma.fields.Email(allow_none=True, description="OAuth email claim")
    access_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="OAuth access token",
    )
    refresh_token = ma.fields.Str(
        required=False, allow_none=True, description="OAuth refresh token"
    )
    token_expiry = ma.fields.AwareDateTime(
        description="OAuth access token expiration time"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="OAuth scopes associated with this access token",
        example=[
            "profile",
            "openid",
            "email",
            "account_info.read",
            "files.metadata.write",
            "files.metadata.read",
            "files.content.write",
            "files.content.read",
        ],
    )
    root_info = ma.fields.Raw(
        description="Root path namespace for Dropbox API requests"
    )

    private = ["access_token", "refresh_token", "root_info"]
    connector_id = DROPBOX_CONNECTOR_ID


DropboxCollectionPolicies = data_type_schema.DropboxCollectionPolicies
DropboxStoragePolicies = data_type_schema.DropboxStoragePolicies
DropboxUserCredentialPolicies = data_type_schema.DropboxUserCredentialPolicies
