import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)

GOOGLE_DRIVE_CONNECTOR_ID = "976cf0cf-78c3-4aab-82d2-7c16adbcc281"


@storage_gateway_policy("google_drive_storage_policies#1.0.0")
class GoogleDriveStoragePolicies_1_0_0(PolicySchema):
    """
    Connector-specific storage gateway policies for the Google Drive connector
    """

    client_id = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Client ID registered with the Google Application console to access
        Google Drive.
        """,
    )
    secret = ma.fields.Str(
        required=True,
        allow_none=True,
        description="""
        Secret created to access access Google Drive with the client_id in this
        policy.
        """,
    )
    user_api_rate_quota = ma.fields.Int(
        required=False, description="User API Rate quota associated with this client ID"
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be registered on the Google API
        console for the application client_id in order to process Google
        credentials.
        """,
    )

    private = ["client_id", "secret", "user_api_rate_quota"]
    connector_id = GOOGLE_DRIVE_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data


@storage_gateway_policy("google_drive_storage_policies#1.1.0")
class GoogleDriveStoragePolicies_1_1_0(GoogleDriveStoragePolicies_1_0_0):
    """
    Connector-specific storage gateway policies for the Google Drive connector
    """

    allow_any_account = ma.fields.Bool(
        required=False,
        load_default=False,
        description="""
        If true, allow users to access personal or external Google accounts.
        If false (the default), users must use the Google account which
        matches the username their Globus credential maps to.
        """,
    )

    connector_id = GOOGLE_DRIVE_CONNECTOR_ID


@collection_policy("google_drive_collection_policies#1.0.0")
class GoogleDriveCollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the Google Drive connector
    """

    connector_id = GOOGLE_DRIVE_CONNECTOR_ID


@user_credential_policy("google_drive_user_credential_policies#1.0.0")
class GoogleDriveUserCredentialPolicies_1_0_0(PolicySchema):
    """
    Connector-specific user credential policies for the Google Drive connector
    """

    sub = ma.fields.Str(allow_none=True, description="OAuth subject identifier claim")
    email = ma.fields.Email(allow_none=True, description="OAuth email claim")
    access_token = ma.fields.Str(
        required=False,
        allow_none=True,
        description="OAuth access token",
    )
    refresh_token = ma.fields.Str(
        required=False, allow_none=True, description="OAuth refresh token"
    )
    token_expiry = ma.fields.AwareDateTime(
        description="OAuth access token expiration time"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="OAuth scopes associated with this access token",
        example=[
            "email",
            "profile",
            "https://www.googleapis.com/auth/drive",
            "https://www.googleapis.com/auth/drive.appfolder",
        ],
    )

    private = ["access_token", "refresh_token"]
    connector_id = GOOGLE_DRIVE_CONNECTOR_ID


GoogleDriveCollectionPolicies = data_type_schema.GoogleDriveCollectionPolicies
GoogleDriveStoragePolicies = data_type_schema.GoogleDriveStoragePolicies
GoogleDriveUserCredentialPolicies = data_type_schema.GoogleDriveUserCredentialPolicies
