import marshmallow as ma


class NonFunctionalHostEndpointUpdateSchema(ma.Schema):
    """
    Transfer's view of a non_functional host endpoint
    """

    DATA_TYPE = ma.fields.Constant("endpoint")
    display_name = ma.fields.String(
        required=True, validate=[ma.validate.Length(max=128)]
    )
    public = ma.fields.Boolean(default=True)
    organization = ma.fields.String(
        default=None, validate=[ma.validate.Length(max=1024)]
    )
    department = ma.fields.String(default=None, validate=[ma.validate.Length(max=1024)])
    keywords = ma.fields.Method(
        serialize="dump_keywords", allow_none=True, default=None
    )
    description = ma.fields.Str(default=None, validate=[ma.validate.Length(max=4096)])
    contact_email = ma.fields.Email(default=None)
    contact_info = ma.fields.Str(default=None, validate=[ma.validate.Length(max=4096)])
    info_link = ma.fields.Url(default=None)
    network_use = ma.fields.String(
        validate=[ma.validate.OneOf(["normal", "minimal", "aggressive", "custom"])]
    )

    max_concurrency = ma.fields.Int(validate=[ma.validate.Range(min=1)])
    preferred_concurrency = ma.fields.Int(validate=[ma.validate.Range(min=1)])

    max_parallelism = ma.fields.Int(validate=[ma.validate.Range(min=1)])
    preferred_parallelism = ma.fields.Int(validate=[ma.validate.Range(min=1)])
    gcs_private_version_json = ma.fields.Raw(attribute="fingerprint")

    def dump_keywords(self, obj):
        keywords = ma.utils.get_value(obj, "keywords")
        if keywords is None or keywords == [] or keywords is ma.missing:
            return None
        else:
            return ",".join(keywords)


class NonFunctionalHostEndpointCreateSchema(NonFunctionalHostEndpointUpdateSchema):
    """
    When creating, we set gcs_private_uuid so that we can use the
    Auth Client ID as the Endpoint ID
    """

    gcs_private_uuid = ma.fields.UUID(required=True, attribute="id")
    non_functional = ma.fields.Constant(True)
    subscription_id = ma.fields.UUID()


non_functional_host_endpoint_update_schema = NonFunctionalHostEndpointUpdateSchema()
non_functional_host_endpoint_create_schema = NonFunctionalHostEndpointCreateSchema()
