"""
globus-connect-server endpoint setup
"""
import datetime
import os
import textwrap
import typing as t
import uuid

import click
import jwcrypto
from flask import current_app

import globus_sdk
from gcs_cli import version as gcs_cli_version
from gcs_cli.config import internal_auth_client
from gcs_cli.parsing.types import EXPLICIT_NULL
from globus_sdk.scopes import AuthScopes, TransferScopes

import globus.manager.api.errors as errors
from globus.manager.api.app import make_app
from globus.manager.api.models import (
    AcmeAccount,
    Endpoint,
    GlobusDnsDomain,
    LetsEncryptCertificate,
    Role,
    db,
)
from globus.manager.appsync.client import ConfigurationObjectTag
from globus.manager.appsync.query import GraphQLError
from globus.manager.constants import RoleType
from globus.manager.utils.keychain import (
    DataKeyError,
    NodeKey,
)
from globus.manager.version import __version_fingerprint__

from .utils.deployment_key import DeploymentKey, store_deployment_key


def endpoint_setup(
    display_name: str,
    client_id: t.Optional[str],
    secret: t.Optional[str],
    deployment_key_path: str,
    deployment_node_key: t.Optional[NodeKey],
    project_id: t.Optional[str],
    project_admin: t.Optional[str],
    project_name: t.Optional[str],
    always_create_project: bool,
    owner: str,
    agree_to_letsencrypt_tos: bool,
    dont_set_advertised_owner: bool,
    contact_email: str,
    development_appsync: t.Optional[t.Tuple[str, str]] = None,
    **kwargs,
) -> None:
    ctx = click.get_current_context()

    if not agree_to_letsencrypt_tos:
        click.echo(
            "\n".join(
                textwrap.wrap(
                    "Globus Connect Server uses the Let's Encrypt service to "
                    "manage certificates for the web services it installs "
                    "on your data transfer nodes. These certificates are "
                    "issued for DNS domain names operated by the Globus "
                    "Project."
                )
                + [""]
                + textwrap.wrap("Please read the Terms of Service at:")
                + ["https://letsencrypt.org/repository/"]
            )
        )

        if not click.confirm("Do you agree to these Terms of Service?"):
            ctx.exit(1)

    app = make_app(app_config="globus.manager.api.config.endpoint_cli")
    db.init_app(app)

    endpoint: t.Dict[str, t.Any] = {
        "DATA_TYPE": "endpoint#1.0.0",
        "contact_email": contact_email,
    }
    for k, v in kwargs.items():
        if v is None:
            continue
        if v == EXPLICIT_NULL:
            endpoint[k] = None
        elif isinstance(v, uuid.UUID):
            endpoint[k] = str(v)
        else:
            endpoint[k] = v

    endpoint["display_name"] = display_name
    environment = os.environ.get("GLOBUS_SDK_ENVIRONMENT", "production")

    # Avoid confusing admins as to why some combinations of project options do not
    # throw errors.
    if client_id is not None and (
        project_id is not None
        or project_admin is not None
        or project_name is not None
        or always_create_project is True
    ):
        click.echo(
            "A client_id and secret already exist for this new endpoint so project "
            "options will be ignored\n"
        )

    with app.app_context():
        db.create_all()
        db.session.commit()

        app.config["CLIENT_ID"] = client_id
        app.config["NODE_ID"] = str(client_id)
        app.config["CLIENT_SECRET"] = secret
        app.config["ENVIRONMENT"] = environment
        if development_appsync:
            app.config["GCS_APPSYNC_URL"] = development_appsync[0]
            app.config["GCS_APPSYNC_CLIENT_ID"] = development_appsync[1]

        endpoint_setup = EndpointSetup(
            ctx,
            client_id,
            secret,
            deployment_key_path,
            deployment_node_key,
            project_id,
            project_admin,
            project_name,
            always_create_project,
            owner,
            endpoint,
            environment,
            contact_email,
            dont_set_advertised_owner,
        )
        endpoint_setup.run()
        click.echo(
            "Created endpoint {}".format(endpoint_setup.endpoint_model.endpoint_id)
        )

        click.echo("Endpoint domain_name {}".format(endpoint_setup.domain))

        if not endpoint_setup.subscription_set:
            click.echo(
                "No subscription is set on this endpoint, so only basic "
                "features are enabled.\n"
            )

            click.echo(
                "\nTo enable subscription features on this endpoint, have "
                "the\nGlobus subscription manager for your organization "
                "assign a subscription to this endpoint.\n"
            )

        click.echo(
            "\nIf you plan on using the Google Drive or Google Cloud Storage\n"
            "connectors, use\n "
            "    https://{}/api/v1/authcallback_google\n"
            "as the Authorized redirect URI for this endpoint".format(
                endpoint_setup.domain
            )
        )


def _using_client_credentials() -> bool:
    return (
        os.getenv("GCS_CLI_CLIENT_ID") is not None
        and os.getenv("GCS_CLI_CLIENT_SECRET") is not None
    )


def _is_client_identity(id: str) -> bool:
    return id.endswith("@clients.auth.globus.org")


class EndpointSetup:
    def __init__(
        self,
        ctx,
        client_id: t.Optional[str],
        secret: t.Optional[str],
        deployment_key_path,
        deployment_node_key: t.Optional[NodeKey],
        project_id: t.Optional[str],
        project_admin: t.Optional[str],
        project_name: t.Optional[str],
        always_create_project: bool,
        owner,
        endpoint,
        environment,
        contact_email: t.Optional[str],
        dont_set_advertised_owner: bool,
    ) -> None:
        self.ctx = ctx
        self.client_id = client_id
        self.secret = secret
        self.deployment_key_path = deployment_key_path
        self.deployment_node_key = deployment_node_key
        self.project_id = project_id
        self.project_admin = project_admin if project_admin is not None else owner
        self.project_name = project_name
        self.always_create_project = always_create_project
        self.endpoint = endpoint
        self.subscription_id = endpoint.pop("subscription_id", None)
        self.subscription_set = False
        self.environment = environment
        self.domain = None
        self.owner = owner
        self.contact_email = contact_email
        self.dont_set_advertised_owner = dont_set_advertised_owner

    def run(self):
        tasks = [
            # START: Bootstrap phase, possibly no client ID or secret
            self.verify_owner_id,
            self.verify_project_admin_id,
            self.do_auth_flow,
            self.verify_client_credentials,
            self.get_client_credentials,
            # END: Bootstrap phase, possibly no client ID or secret
            self.create_deployment_key,
            self.check_for_existing_keychain,
            self.create_initial_endpoint,
            self.add_keys_to_keychain,
            self.download_existing_configuration,
            self.create_transfer_endpoint,
            self.check_transfer_subscription,
            self.create_roles,
            self.pre_register_dns,
            self.register_dns,
            self.create_acme_client,
            self.pre_provision_certificate,
            self.provision_certificate,
            self.register_server,
        ]

        if not self.dont_set_advertised_owner:
            tasks.append(self.set_owner)

        def show(obj):
            if obj and obj.__doc__:
                return obj.__doc__.strip().split("\n")[0]
            return ""

        with click.progressbar(tasks, item_show_func=show, width=25) as bar:
            for task in bar:
                task()

    # START: Bootstrap phase, possibly no endpoint client ID or secret

    def _verify_identity(self, username: str) -> str:
        """
        Use the GCS CLI's internal (templated) auth client to verify that the
        username is valid.
        """
        auth_client = internal_auth_client()
        response = auth_client.get_identities(usernames=[username])
        identities = response.data.get("identities", [])

        if len(identities) != 1:
            click.echo("\nUnable to verify that Globus user {} exists".format(username))
            self.ctx.exit(1)

        identity = identities[0]
        current_app.logger.debug(
            "verify_identity: username=%s id=%s status=%s",
            username,
            identity["id"],
            identity["status"],
        )
        if identity["status"] not in ["used", "private"]:
            click.echo(
                f"\n{username} exists but does not have a validated account: {identity['status']}"
            )
            self.ctx.exit(1)

        if identity["id"] == str(self.client_id):
            click.echo(
                "\nThe parameter to the --owner and --project-admin command-line "
                "options must be different than the endpoint's client id",
                err=True,
            )
            self.ctx.exit(1)

        if identity["identity_type"] == "email":
            click.echo(
                "\n"
                + "\n".join(
                    textwrap.wrap(
                        "The parameter to the --owner and --project-admin command-line "
                        "options must be accounts with a non-email identity_type",
                    )
                ),
                err=True,
            )
            self.ctx.exit(1)

        return identity["id"]

    def verify_owner_id(self):
        """
        Verify owner ID.

        Verifies self.owner and sets self.owner_id (str)
        """
        current_app.logger.info("verify_owner_id: start")
        self.owner_id = self._verify_identity(self.owner)
        current_app.logger.info("verify_owner_id: end")

    def verify_project_admin_id(self):
        """
        Verify Auth project admin ID.

        Verifies self.project_admin and sets self.project_admin_id (str)
        and self.project_admin if not already set.
        """
        current_app.logger.info("verify_project_admin_id: start")
        if self.project_admin != self.owner:
            self.project_admin_id = self._verify_identity(self.project_admin)
        else:
            self.project_admin_id = self.owner_id
        current_app.logger.info("verify_project_admin_id: end")

    def do_auth_flow(self) -> None:
        """
        Perform any necessary Auth flows. Outputs self.transfer_token and self.auth_token.
        """
        if _using_client_credentials():
            self._do_client_auth_flow()
        else:
            self._do_link_auth_flow()

    def _do_client_auth_flow(self) -> None:
        """
        Get the necessary tokens for the remainder of the endpoint setup operations.

        Client flows do not prompt for authentication and so do not allow us to enforce
        the use of linked identities through session_required_identities. So mimic those
        requirements explicitly.
        """
        if not self.client_id:
            # We can't prevent using an unlinked project admin identity at this point
            # because we may not end up with an actual project creation. So defer the
            # check until project creation below.
            self.auth_token = globus_sdk.ClientCredentialsAuthorizer(
                globus_sdk.ConfidentialAppAuthClient(
                    t.cast(str, os.getenv("GCS_CLI_CLIENT_ID")),
                    t.cast(str, os.getenv("GCS_CLI_CLIENT_SECRET")),
                    environment=self.environment,
                ),
                ["urn:globus:auth:scope:auth.globus.org:manage_projects"],
            ).access_token

        if self.dont_set_advertised_owner is False:
            # Prevent attempting to set the advertised owner to an unlinked identity.
            if self.owner_id != os.getenv("GCS_CLI_CLIENT_ID"):
                raise click.ClickException(
                    f"Can not set the advertised owner to {self.owner} when using "
                    "client credentials. Rerun with the --dont-set-advertised-owner "
                    "option."
                )

            self.transfer_token = globus_sdk.ClientCredentialsAuthorizer(
                globus_sdk.ConfidentialAppAuthClient(
                    t.cast(str, os.getenv("GCS_CLI_CLIENT_ID")),
                    t.cast(str, os.getenv("GCS_CLI_CLIENT_SECRET")),
                    environment=self.environment,
                ),
                [TransferScopes.all],
            ).access_token

    def _do_link_auth_flow(self) -> None:
        """
        Perform any necessary Auth flows.

        Starts interactive flow for tokens to be used with set_owner() and
        get_client_credentials(). Outputs self.transfer_token and self.auth_token.

        TODO: We use internal_auth_client which uses the CLI's template client which
        acts on its own behalf (not the callers). We should try to reuse tokens that
        are already in token storage.
        """
        current_app.logger.info("do_link_auth_flow: start")

        scopes = []
        session_params = {"prompt": "login"}
        session_messages = []
        session_required_identities = []

        require_auth_token = False
        if not self.client_id:
            require_auth_token = True
            scopes += ["urn:globus:auth:scope:auth.globus.org:manage_projects"]
            session_messages.append("register the new endpoint to get credentials ")
            # Globus Auth has a 30 min, in-session authentication requirement on
            # the projects API. This also guarantees that the caller is an admin on
            # the new project so that we can create a client later.
            session_required_identities.append(self.project_admin_id)

        require_transfer_token = False
        if self.dont_set_advertised_owner is False:
            require_transfer_token = True
            scopes += [TransferScopes.all]
            session_messages.append("set the advertised owner of the endpoint")
            # Tranfer does not have an authentication requirement however this helps
            # guarantee that owner_id is in the caller's linked identity set.
            if self.owner_id not in session_required_identities:
                session_required_identities.append(self.owner_id)

        if session_required_identities:
            session_params["session_required_identities"] = ",".join(
                session_required_identities
            )

        if scopes:
            auth_client = internal_auth_client()
            try:
                auth_client.oauth2_start_flow(
                    redirect_uri=auth_client.base_url + "v2/web/auth-code",
                    refresh_tokens=False,
                    requested_scopes=scopes,
                )
                session_message = "and ".join(session_messages)
                session_params["session_message"] = session_message.capitalize()
                linkprompt = f"Please authenticate with Globus to {session_message}"
                click.echo(
                    "\n{0}:\n{1}\n{2}\n{1}\n".format(
                        linkprompt,
                        "-" * len(linkprompt),
                        auth_client.oauth2_get_authorize_url(
                            query_params=session_params
                        ),
                    )
                )
                auth_code = click.prompt(
                    "Enter the resulting Authorization Code here", default=""
                ).strip()
                if not auth_code:
                    raise Exception("No auth code")
                token_response = auth_client.oauth2_exchange_code_for_tokens(auth_code)
            except Exception as e:
                raise click.ClickException(f"Authorization flow failed: {e}")

            if require_transfer_token:
                token_resp = token_response.by_resource_server.get(
                    TransferScopes.resource_server
                )
                assert token_resp is not None  # mypy kludge
                self.transfer_token = token_resp["access_token"]

            if require_auth_token:
                token_resp = token_response.by_resource_server.get(
                    AuthScopes.resource_server
                )
                assert token_resp is not None  # mypy kludge
                self.auth_token = token_resp["access_token"]

        current_app.logger.info("do_link_auth_flow: end")

    def verify_client_credentials(self) -> None:
        """
        Verify client credentials
        """
        if self.client_id:
            try:
                appsync_client = current_app.appsync_client
                if appsync_client is None:
                    click.echo("Unable to contact Globus services", err=True)
                    self.ctx.exit(1)
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.error(e)
                click.echo(
                    "\n".join(
                        textwrap.wrap(
                            "Unable to contact Globus services with provided client "
                            "id and secret. Please verify that those values are "
                            "correct."
                        )
                    ),
                    err=True,
                )
                self.ctx.exit(1)

    def _get_username(self) -> str:
        """
        Returns the ID of the identity being use for project management. If we
        are being run with client credentials, then is the client username. Otherwise
        it is the project admin (which defaults to the owner).
        """
        if _using_client_credentials():
            return f"{os.getenv('GCS_CLI_CLIENT_ID')}@clients.auth.globus.org"

        return self.project_admin

    def lookup_auth_project(self) -> None:
        """
        Search for an existing Auth project for this endpoint's client registration.

        Filter the available project list by project-id, project-admin and project-name
        if given. In order to make use of an existing project, it must match all of the
        options provided.

        Outputs:
        - self.project_id (if there is an existing project)
        """
        current_app.logger.info("lookup_auth_project: start")
        if not self.client_id and not self.always_create_project:
            assert self.auth_token

            client = globus_sdk.AuthClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(
                    self.auth_token,
                ),
                app_name=gcs_cli_version.app_name,
            )

            # GET /v2/api/projects does not have any session requirements, this will
            # return all projects for which any linked identity that is a project admin.
            # That may include projects connected to identities other than owner or
            # project_admin which may not be in session.
            try:
                projects_response = client.get("/v2/api/projects")
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.warning(e)
                raise click.ClickException(f"Failed to get the project list: {e}")

            # Filter by project_id. If not given, include all projects.
            filtered_projects = projects_response["projects"]
            if self.project_id is not None:
                filtered_projects = [
                    p
                    for p in projects_response["projects"]
                    if p["id"] == self.project_id
                ]

            if self.project_name is not None:
                filtered_projects = [
                    p
                    for p in filtered_projects
                    if self.project_name == p["display_name"]
                ]

            filtered_projects = [
                p
                for p in filtered_projects
                if self.project_admin_id
                in [id["id"] for id in p["admins"]["identities"]]
            ]

            if len(filtered_projects) == 0:
                if self.project_id is not None:
                    if len(projects_response["projects"]) == 0:
                        raise click.ClickException(
                            f"Project {self.project_id} does not exist or "
                            f"{self._get_username()} is not an admin on the project "
                            "and can not view it. Rerun the command without the "
                            "--project-id option and a project will be created for you "
                            "or go to https://app.globus.org/settings/developers and "
                            "create a project before continuing."
                        )
                    else:
                        raise click.ClickException(
                            f"Project {self.project_id} does not exist or "
                            f"{self._get_username()} is not an admin on the project "
                            "and can not view it. Please rerun this command and use "
                            "--project-id to choose an existing project.\n"
                            "Available projects:\n"
                            + "\n".join(
                                [
                                    f"{p['id']} - {p['display_name']}"
                                    for p in projects_response["projects"]
                                ]
                            )
                        )

                if len(projects_response["projects"]) > 0:
                    raise click.ClickException(
                        "You have existing projects but none match the project "
                        "options used with this command. Please rerun this command and "
                        "use --project-id to choose an existing project.\n"
                        "Available projects:\n"
                        + "\n".join(
                            [
                                f"{p['id']} - {p['display_name']}"
                                for p in projects_response["projects"]
                            ]
                        )
                    )

                # A project will be created
                current_app.logger.info("lookup_auth_project: end")

            if len(filtered_projects) == 1:
                self.project_id = list(filtered_projects)[0]["id"]
                current_app.logger.info("lookup_auth_project: end")

            if len(filtered_projects) > 1:
                raise click.ClickException(
                    "You have multiple existing projects that match the project "
                    "options used with this command. Please rerun this command and "
                    "use --project-id to choose an existing project.\n"
                    "Available projects:\n"
                    + "\n".join(
                        [
                            f"{p['id']} - {p['display_name']}"
                            for p in projects_response["projects"]
                        ]
                    )
                )

        current_app.logger.info("lookup_auth_project: end")

    def create_auth_project(self) -> None:
        """
        Create an Auth project for this endpoint's client registration.

        Outputs:
        - self.project_id
        """
        current_app.logger.info("create_auth_project: start")
        if self.always_create_project or not self.project_id:
            assert self.auth_token

            client = globus_sdk.AuthClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(
                    self.auth_token,
                ),
                app_name=gcs_cli_version.app_name,
            )

            # Refuse to create a new project if the sole project_admin is a client
            # identity.
            if _is_client_identity(self.project_admin):
                raise click.ClickException(
                    f"The identity {self.project_admin} does not have any existing "
                    "Auth projects and one will not be automatically created with an "
                    "Auth client as the sole project admin. Visit "
                    "https://app.globus.org/settings/developers to assign "
                    f"{self.project_admin} to a new or existing project."
                )

            # Create a default project
            display_name = self.project_name
            if display_name is None:
                display_name = "Globus Connect Server Endpoints"

            admin_ids: list[str] = [str(self.project_admin_id)]

            # When we are running with client credentials, we need to add the client
            # as an admin on the project so that we can create the client/id/secret
            # later on.
            if _using_client_credentials():
                admin_ids.append(t.cast(str, os.getenv("GCS_CLI_CLIENT_ID")))

            project_doc = {
                "project": {
                    "display_name": display_name,
                    "contact_email": self.contact_email,
                    "admin_ids": admin_ids,
                }
            }

            try:
                response = client.post("/v2/api/projects", data=project_doc)
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.warning(e)
                raise click.ClickException(f"Failed to get create a project: {e}")

            self.project_id = response["project"]["id"]

        current_app.logger.info("create_auth_project: end")

    def get_client_credentials(self) -> None:
        """
        Register the endpoint with Globus Auth

        Outputs:
        - self.client_id
        - self.secret
        - app.config["CLIENT_ID"] = str(client_id)
        - app.config["NODE_ID"] = str(client_id)
        - app.config["CLIENT_SECRET"] = secret
        """
        current_app.logger.info("get_client_credentials: start")

        if not self.client_id:
            self.lookup_auth_project()
            self.create_auth_project()

            assert self.auth_token

            client = globus_sdk.AuthClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(
                    self.auth_token,
                ),
                app_name=gcs_cli_version.app_name,
            )

            # Create the client
            client_doc = {
                "client": {
                    "name": self.endpoint["display_name"],
                    "project": str(self.project_id),
                    "public_client": False,
                    "client_type": "globus_connect_server",
                }
            }

            try:
                response = client.post("/v2/api/clients", data=client_doc)
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.warning(e)
                raise click.ClickException(f"Failed to create a client: {e}")
            client_id = response["client"]["id"]

            # Create the credentials
            credential_doc = {
                "credential": {
                    "name": f"Created by {gcs_cli_version.app_name} at {datetime.datetime.now()}",
                }
            }

            try:
                response = client.post(
                    f"/v2/api/clients/{client_id}/credentials", data=credential_doc
                )
            except globus_sdk.GlobusAPIError as e:
                current_app.logger.warning(e)
                raise click.ClickException(
                    f"Failed to get create client credentials: {e}"
                )
            secret = response["credential"]["secret"]

            self.client_id = str(client_id)
            self.secret = secret
            current_app.config["CLIENT_ID"] = str(client_id)
            current_app.config["NODE_ID"] = str(client_id)
            current_app.config["CLIENT_SECRET"] = secret

        current_app.logger.info("get_client_credentials: end")

    # END: Bootstrap phase, possibly no client ID or secret

    def create_deployment_key(self):
        """
        Configure deployment key

        If the output deployment key exists, load it and use that as the
        deployment key. Otherwise create a new key and serialize it to that
        file.
        """
        if self.deployment_node_key is None:
            self.deployment_node_key = NodeKey.generate()

        store_deployment_key(
            self.deployment_key_path,
            DeploymentKey(self.deployment_node_key, self.client_id, self.secret),
        )

        current_app.config["DEPLOYMENT_KEY"] = self.deployment_node_key

    def check_for_existing_keychain(self):
        """
        Configure keychain
        """
        current_app.logger.info("check_for_existing_keychain: start")
        try:
            items = current_app.appsync_client.query_all(
                id=ConfigurationObjectTag.KEY_CHAIN.value
            )
            for item in items:
                current_app.appsync_client.process_item(item, catch_exceptions=False)
        except GraphQLError as e:
            current_app.logger.info(str(e))
        except jwcrypto.jwe.InvalidJWEData as e:
            current_app.logger.error(str(e))
            click.echo(
                "\n"
                + "\n".join(
                    textwrap.wrap(
                        "Error decrypting keychain. This usually occurs when an "
                        "endpoint was previously set up with the same client ID "
                        "but a different deployment key. If you still have the "
                        "old deployment key, you can rerun this command with the "
                        "--deployment-key parameter to use it. If not, you should "
                        "run globus-connect-server endpoint cleanup, "
                        "create a new endpoint Auth client, and then rerun this "
                        "command."
                    )
                ),
                err=True,
            )
            self.ctx.exit(1)

        current_app.logger.info("check_for_existing_keychain: end")

    def create_initial_endpoint(self):
        """
        Initialize Endpoint
        """
        current_app.logger.info("create_initial_endpoint: start")
        try:
            current_app.logger.debug("create_initial_endpoint: query_appsync")
            items = current_app.appsync_client.query_all(
                tag=ConfigurationObjectTag.ENDPOINT
            )
            if items:
                endpoint_model = current_app.appsync_client.process_item(items[0])
            else:
                endpoint_model = None
            current_app.logger.debug(
                "create_initial_endpoint: query_result=%s", endpoint_model
            )
        except GraphQLError as e:
            current_app.logger.info(str(e))
            endpoint_model = None

        if not endpoint_model:
            current_app.logger.debug("create_initial_endpoint: create_stub_endpoint")
            endpoint_model = Endpoint(doc={})
            try:
                current_app.logger.debug(
                    "create_initial_endpoint: register_with_appsync"
                )
                current_app.appsync_client.create_object(endpoint_model)
            except GraphQLError as e:
                if e.data:
                    res = e.data
                    for o in res.get("configuration_objects", {}).get("items", []):
                        if (
                            ConfigurationObjectTag(o["tag"])
                            == ConfigurationObjectTag.ENDPOINT
                            and o["id"] == self.client_id
                        ):
                            endpoint_model = Endpoint(appsync_data=o)
        self.endpoint_model = endpoint_model
        current_app.logger.debug("create_initial_endpoint: add_to_database")
        db.session.add(endpoint_model)
        current_app.logger.debug("create_initial_endpoint: commit_to_database")
        db.session.commit()
        current_app.logger.info("create_initial_endpoint: end")

    def add_keys_to_keychain(self):
        """
        Add keys to keychain
        """
        current_app.logger.info("add_keys_to_keychain: start")
        if self.deployment_node_key.thumbprint() not in [
            k.thumbprint() for k in current_app.keychain.node_keys
        ]:
            current_app.logger.info(
                "add_keys_to_keychain: add key_id=%s",
                self.deployment_node_key.thumbprint(),
            )
            try:
                current_app.keychain.add_node_key(self.deployment_node_key)
            except errors.LimitExceededError as e:
                click.echo(f"Error adding deployment node key: {e}", err=True)
                self.ctx.exit(1)

        if not current_app.keychain.data_keys:
            dk = current_app.keychain.add_data_key()
            current_app.logger.info(
                "add_keys_to_keychain: create_data_key key_id=%s", dk.thumbprint()
            )
        try:
            current_app.appsync_client.create_object(current_app.keychain)
        except GraphQLError as e:
            conflict_error = False
            for errobj in e.errors:
                if errobj["errorType"] == "CONFLICT":
                    conflict_error = True
            if not conflict_error:
                current_app.logger.error(str(e))
                self.ctx.exit(1)
        current_app.logger.info("add_keys_to_keychain: end")

    @staticmethod
    def __endpoint_config_filter(items):
        return [
            i
            for i in items
            if (
                i["tag"] == ConfigurationObjectTag.ENDPOINT
                or i["tag"] == ConfigurationObjectTag.WILDCARD_DOMAIN
                or i["tag"] == ConfigurationObjectTag.ACME_ACCOUNT
                or i["tag"] == ConfigurationObjectTag.CERTIFICATE
                or (i["tag"] == ConfigurationObjectTag.ROLE and not i.get("resource"))
            )
        ]

    def download_existing_configuration(self):
        """
        Download configuration data
        """

        current_app.logger.info("download_existing_configuration: start")
        # Download endpoint configuration if present
        current_app.logger.info("download_existing_configuration: query_models")
        items = current_app.appsync_client.query_all(
            item_filter=self.__endpoint_config_filter
        )
        current_app.logger.info(
            "download_existing_configuration: commit_models_to_session"
        )
        current_app.appsync_client.commit_items_to_session(db.session, items)
        current_app.logger.info("download_existing_configuration: end")

    def create_transfer_endpoint(self):
        """
        Create Endpoint in Globus
        """
        current_app.logger.info("create_transfer_endpoint: start")
        endpoint_model = Endpoint.query.one()

        if not hasattr(endpoint_model, "endpoint_id"):
            try:
                endpoint_id = self.endpoint.get("id", self.client_id)
                res = current_app.transfer_client.get_endpoint(endpoint_id)
                if res:
                    current_app.logger.debug(
                        "create_transfer_endpoint: endpoint already exists, "
                        "but is not in gcs.globus.org"
                    )
                    endpoint_model.endpoint_id = endpoint_id
            except globus_sdk.TransferAPIError as e:
                if e.code == "EndpointDeleted":
                    click.echo(
                        "\nAn endpoint with this ID has already been deleted and "
                        "this ID may not be reused.\nMake sure you are not reusing "
                        "an older deployment key and do not specify the\n--client-id "
                        "option",
                        err=True,
                    )
                    self.ctx.exit(1)
                elif e.http_status != 404:
                    current_app.logger.error(
                        "create_transfer_endpoint: error checking if "
                        "endpoint already exists: %s",
                        str(e),
                    )
                    self.ctx.exit(1)

        if not hasattr(endpoint_model, "endpoint_id"):
            try:
                current_app.logger.debug(
                    "download_existing_configuration: initializing endpoint"
                )
                endpoint_model.fingerprint = {"GCS": __version_fingerprint__}
                endpoint_model.put(**self.endpoint)
                current_app.logger.debug(
                    "download_existing_configuration: saving endpoint"
                )
                endpoint_model = current_app.appsync_client.update_object(
                    endpoint_model
                )
                current_app.logger.debug(
                    "download_existing_configuration: add to session"
                )
                db.session.add(endpoint_model)
                current_app.logger.debug(
                    "download_existing_configuration: commit to database"
                )
                db.session.commit()
            except DataKeyError as e:
                current_app.logger.error(
                    "graphql operation cannot be processed: %s", str(e)
                )
                self.ctx.exit(1)
        else:
            current_app.logger.debug(
                "download_existing_configuration: updating endpoint"
            )

            res = current_app.transfer_client.get_endpoint(endpoint_model.endpoint_id)
            if "subscription_id" in res.data:
                self.endpoint["subscription_id"] = res.data["subscription_id"]

            endpoint_model.put(endpoint_id=endpoint_model.endpoint_id, **self.endpoint)

            # record endpoint document in appsync
            current_app.appsync_client.update_object(endpoint_model)
            current_app.logger.debug("download_existing_configuration: saving endpoint")
            current_app.logger.debug("download_existing_configuration: add to session")
            db.session.add(endpoint_model)
            current_app.logger.debug(
                "download_existing_configuration: commit to database"
            )
            db.session.commit()

        current_app.config["HOST_ENDPOINT_ID"] = endpoint_model.endpoint_id
        self.endpoint_model = endpoint_model
        current_app.logger.info("create_transfer_endpoint: start")

    def check_transfer_subscription(self):
        """
        Check subscription status
        """
        current_app.logger.info("check_transfer_subscription: start")

        ep_doc = current_app.transfer_client.get_endpoint(
            self.endpoint_model.endpoint_id
        )

        subscription_id = ep_doc.get("subscription_id")

        if subscription_id:
            # If the endpoint already has a subscription, update the model
            self.endpoint_model.patch(
                DATA_TYPE="endpoint#1.0.0",
                subscription_id=subscription_id,
            )
            current_app.appsync_client.update_object(self.endpoint_model)
            db.session.add(self.endpoint_model)
            db.session.commit()
            if self.subscription_id and self.subscription_id != subscription_id:
                current_app.logger.warn(
                    "check_transfer_subscription: subscription_id=%s " "requested=%s",
                    subscription_id,
                    self.subscription_id,
                )
                click.echo(
                    "Endpoint subscription_id does not match requested value", err=True
                )
            self.subscription_set = True
        current_app.logger.info("check_transfer_subscription: end")

    def create_roles(self) -> None:
        """
        Create owner role
        """
        current_app.logger.info("create_roles: start")
        role = Role.query.filter(
            ~Role.deleted,  # type: ignore
            Role.role == RoleType.endpoint_owner,
        ).one_or_none()

        principal = "urn:globus:auth:identity:" + self.owner_id
        if role:
            if role.identity_id != self.owner_id:
                current_app.logger.error(
                    "Owner role already exists for %s", role.identity_id
                )
                old_admin = current_app.auth_client.get_identities(
                    ids=[role.identity_id]
                ).data.get("identities", [])
                if len(old_admin) == 1:
                    old_user = old_admin[0]["username"]
                    click.echo("Owner role exists for {}".format(old_user))
                else:
                    click.echo("Owner role exists for {}".format(role.identity_id))

                if not click.confirm("Replace owner role?"):
                    self.ctx.exit(1)

                role.put_owner_role(self.owner_id)
                try:
                    role = current_app.appsync_client.update_object(role)
                    db.session.add(role)
                    db.session.commit()
                except (globus_sdk.exc.GlobusAPIError, errors.BadGatewayError) as e:
                    current_app.logger.error(
                        "Unable to update owner role: {}".format(str(e))
                    )
                    self.ctx.exit(1)
            else:
                current_app.logger.info("create_roles: reusing_existing_role")
        else:
            try:
                role = Role(
                    doc=dict(DATA_TYPE="role#1.0.0", principal=principal, role="owner"),
                    endpoint=self.endpoint_model,
                )
                role.primary_role = True
                role = current_app.appsync_client.create_object(role)
                db.session.add(role)
                db.session.commit()
            except (globus_sdk.exc.GlobusAPIError, errors.BadGatewayError) as e:
                current_app.logger.info(
                    "Unable to create owner role: {}".format(str(e))
                )

        if self.subscription_set:
            found = False
            for admin_role in Role.query.filter(
                Role.role == RoleType.endpoint_administrator,
            ):
                if admin_role.principal == principal:
                    found = True
                    if not admin_role.primary_role:
                        admin_role.primary_role = True
                        admin_role = current_app.appsync_client.update_object(
                            admin_role
                        )
                        db.session.add(admin_role)

                elif admin_role.principal != principal and admin_role.primary_role:
                    admin_role.primary_role = False
                    admin_role = current_app.appsync_client.update_object(admin_role)
                    db.session.add(admin_role)

            if not found:
                admin_role = Role(
                    doc=dict(
                        DATA_TYPE="role#1.0.0",
                        principal=principal,
                        role="administrator",
                    ),
                    endpoint=self.endpoint_model,
                    primary_role=True,
                )
                admin_role = current_app.appsync_client.create_object(admin_role)
                db.session.add(admin_role)
            db.session.commit()
        current_app.logger.info("create_roles: end")

    def pre_register_dns(self):
        """
        DNS registration (may take a few minutes)
        """
        pass

    def register_dns(self):
        """
        DNS registration
        """
        current_app.logger.info("register_dns: start")
        domain = GlobusDnsDomain.query.one_or_none()

        if not domain:
            domain = GlobusDnsDomain()
            current_app.appsync_client.create_object(domain)
            db.session.add(domain)
            db.session.commit()
        self.domain = domain.domain_name
        current_app.logger.info("register_dns: end")

    def create_acme_client(self):
        """
        Create Let's Encrypt account
        """
        current_app.logger.info("create_acme_account: start")
        acme_account = AcmeAccount.query.one_or_none()
        if not acme_account:
            acme_account = AcmeAccount()
            try:
                acme_account = current_app.appsync_client.create_object(acme_account)
            except GraphQLError as e:
                if len(e.args) > 1:
                    acme_account = AcmeAccount(appsync_data=e.args[1])
                else:
                    current_app.logger.error("Creating acme account: {}".format(e))
                    self.ctx.exit(1)

            db.session.add(acme_account)
            db.session.commit()
        self.acme_account = acme_account
        current_app.logger.info("create_acme_account: end")

    def pre_provision_certificate(self):
        """
        Get Certificate (may take a few minutes)
        """

    def provision_certificate(self):
        """
        Get Certificate
        """
        current_app.logger.info("provision_certificate: start")
        certificate = LetsEncryptCertificate.query.one_or_none()
        if not certificate:
            acme_account = AcmeAccount.query.one()
            acme_account.acquire_lease()

            certificate = LetsEncryptCertificate.create_certificate(self.domain)
            certificate = current_app.appsync_client.create_object(certificate)
            acme_account.release_lease()
            db.session.add(certificate)
            db.session.commit()
        current_app.logger.info("provision_certificate: end")

    def register_server(self):
        """
        Register with Globus
        """
        current_app.logger.info("register_server: start")
        self.endpoint_model.update_endpoint_server()
        current_app.logger.info("register_server: end")

    def set_owner(self):
        """
        Set endpoint advertised owner
        """
        assert self.transfer_token

        try:
            self.endpoint_model.set_owner_string(
                owner_id=self.owner_id,
                token=self.transfer_token,
            )
        except Exception:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "Unable to set the advertised owner id. Have the "
                        "endpoint owner or an endpoint administrator "
                        "log in to the endpoint and use the command "
                        "globus-connect-server endpoint set-owner-string "
                        "to set this."
                    )
                )
            )
