"""
globus-connect-server endpoint setup
"""
import typing as t
import uuid

import click

from gcs_cli.parsing.command_decorator import globus_command
from gcs_cli.parsing.types import CommaDelimitedList, StringOrNull

from .utils import options as options
from .utils.deployment_key import DeploymentKey


@globus_command(
    "setup",
    short_help="Setup a Globus Connect Server endpoint",
    help="Setup a Globus Connect Server endpoint",
    disable_options=["format", "explicit_host"],
)
@click.argument("display_name")
@click.option(
    "--public/--private",
    "public",
    default=True,
    help="Set the endpoint to be public or private [default: public]",
    show_default=False,
)
@click.option(
    "--description",
    type=StringOrNull(),
    help="Description of the endpoint",
)
@click.option(
    "--info-link",
    type=StringOrNull(),
    help="Link to a web page with more information about the endpoint",
)
@click.option(
    "--contact-info",
    type=StringOrNull(),
    help=(
        "Other non-email contact information for the endpoint, e.g.  phone and "
        + "mailing address. This is visible to end users for support."
    ),
)
@click.option(
    "--contact-email",
    type=StringOrNull(),
    help=(
        "Email address of the support contact for this endpoint. This is visible "
        + "to end users so that they may contact your organization for support."
    ),
    required=True,
)
@click.option(
    "--organization",
    type=StringOrNull(),
    help="Organization that owns this endpoint",
    required=True,
)
@click.option(
    "--department",
    type=StringOrNull(),
    help="Department within the organization that runs the endpoint",
)
@click.option(
    "--keywords",
    type=CommaDelimitedList(),
    help="Comma separated list of keywords to help searches for the endpoint",
)
@options.deployment_key_option(
    click_path=click.Path(
        dir_okay=False,
        writable=True,
    ),
    path_arg_name="deployment_key_path",
)
@options.deprecated_client_id_option
@options.deprecated_secret_option
@click.option(
    "--project-id",
    metavar="PROJECT_ID",
    type=click.UUID,
    help=(
        "Auth project ID where the new endpoint client will be stored. "
        "Only required if you are an admin on multiple Auth projects. "
        "Ignored if using --always-create-project."
    ),
)
@click.option(
    "--project-admin",
    help=(
        "Globus username of the admin of the Auth project where the new endpoint "
        "client will be registered. Only required if the project admin identity "
        "is different from the endpoint owner. This identity must belong to the "
        "caller. If not given, the value of --owner is used. Unless "
        "--dont-set-advertised-owner is used, the value of --owner must be an "
        "identity linked to the project admin identity."
    ),
)
@click.option(
    "--project-name",
    help=(
        "Name of the Auth project where the new endpoint client will be registered. "
        "If the project does not exist, it will be created with this value as "
        'its display name. Defaults to "Globus Connect Server Endpoints."'
    ),
)
@click.option(
    "--always-create-project",
    help=(
        "Create a new auth project for this endpoint, even if you are already the admin of "
        "an existing project. Without this option, only your first project can be auto-created. "
    ),
    is_flag=True,
)
@click.option(
    "--owner",
    "-o",
    help=(
        "Globus username of the owner of this endpoint. Unless --dont-set-advertised-owner "
        "is used, the owner identity must belong to the caller. This user will be granted the "
        "owner role for this endpoint."
    ),
    required=True,
)
@click.option(
    "--agree-to-letsencrypt-tos",
    help="Agree to the Let's Encrypt Terms of Service: "
    "https://letsencrypt.org/repository/",
    is_flag=True,
)
@click.option(
    "--dont-set-advertised-owner",
    help="Skip the interactive step to set the endpoint advertised owner",
    is_flag=True,
)
@click.option(
    "--development-appsync",
    help="Use the specified URL and client_id for the appsync service",
    metavar="APPSYNC_URL APPSYNC_CLIENT_ID",
    nargs=2,
    hidden=True,
)
def endpoint_setup(
    display_name: str,
    deployment_key_path: str,
    deployment_key: DeploymentKey,
    project_id: t.Optional[uuid.UUID],
    project_admin: t.Optional[str],
    project_name: t.Optional[str],
    always_create_project: bool,
    owner: str,
    agree_to_letsencrypt_tos: bool,
    dont_set_advertised_owner: bool,
    contact_email: str,
    development_appsync: t.Optional[t.Tuple[str, str]] = None,
    **kwargs
) -> None:
    from globus.manager.scripts.endpoint_setup import endpoint_setup as es

    es(
        display_name,
        str(deployment_key.client_id) if deployment_key.client_id else None,
        deployment_key.secret,
        deployment_key_path,
        deployment_key.node_key,
        str(project_id) if project_id else None,
        project_admin,
        project_name,
        always_create_project,
        owner,
        agree_to_letsencrypt_tos,
        dont_set_advertised_owner,
        contact_email,
        development_appsync,
        **kwargs
    )
