from __future__ import annotations

import click
import jinja2

import gcs_cli.parsing.types


class Option:
    def __init__(self, opt: click.Parameter) -> None:
        self.opt = opt

        self.names: list[str] = opt.opts
        if opt.secondary_opts:
            self.names += opt.secondary_opts

    @property
    def has_completable_param(self) -> bool:
        return bool(
            self.choices or self.directory or self.file or self.identity_mapping_type
        )

    @property
    def choices(self) -> list[str]:
        if isinstance(self.opt.type, click.Choice):
            return list(self.opt.type.choices)
        else:
            return []

    @property
    def directory(self) -> bool:
        return (
            isinstance(self.opt.type, click.Path)
            and self.opt.type.exists
            and self.opt.type.dir_okay
        )

    @property
    def file(self) -> bool:
        return (
            isinstance(self.opt.type, click.Path)
            and not self.opt.type.dir_okay
            and self.opt.type.file_okay
        )

    @property
    def identity_mapping_type(self) -> bool:
        return isinstance(self.opt.type, gcs_cli.parsing.types.IdentityMappingType)

    def __repr__(self) -> str:
        return f"<Option {self.names[0]}>"


class Command:
    def __init__(self, parent: Command | None, cmd: click.Command) -> None:
        self.parent: Command | None = parent
        self.name: str = cmd.name
        self.hidden: bool = getattr(cmd, "hidden", False)
        self.short_help: str = getattr(cmd, "short_help", "")

        if hasattr(cmd, "commands"):
            self.commands: list[Command] = [
                Command(self, subcmd) for subcmd in cmd.commands.values()
            ]
        else:
            self.commands = []

        if hasattr(cmd, "params"):
            self.options: list[Option] = [
                Option(p) for p in cmd.params if p.param_type_name != "argument"
            ]
        else:
            self.options = []

    @property
    def conflicting_commands(self) -> str | None:
        if self.parent:
            cmds = [
                c.name for c in self.parent.commands if c.name.startswith(self.name)
            ]
            if len(cmds) > 1:
                return " ".join(cmds)
        return None

    @property
    def children_names(self) -> list[str]:
        return [c.name for c in self.commands if not c.hidden]

    @property
    def option_names(self) -> list[str]:
        return [oname for o in self.options for oname in o.names]

    @property
    def command_list(self) -> list[Command]:
        res = []
        res.append(self)
        res.extend([sub for c in self.commands for sub in c.command_list])
        return res

    @property
    def completion_cmd(self) -> str:
        if self.parent:
            return f"{self.parent.completion_cmd} {self.name}"
        else:
            return self.name

    def __getattr__(self, name: str) -> Command | Option:
        check = name.replace("_", "-")
        for cmd in self.commands:
            if cmd.name == check:
                return cmd
        for opt in self.options:
            if check in [n.replace("--", "") for n in opt.names]:
                return opt
        raise AttributeError

    def __repr__(self) -> str:
        return f"<Command {self.name}>"


@click.command(
    "generate-completions", help="Generate bash completions for GCS CLI", hidden=True
)
def generate_completions_command() -> None:
    environment = jinja2.Environment(
        loader=jinja2.PackageLoader("globus.manager.api", "templates")
    )
    template = environment.get_template("bash-completions.jinja2")

    from .main import main

    cmd_tree = Command(None, main)
    print(template.render(commands=cmd_tree.command_list))
