"""
globus-connect-server node setup

The globus-connect-server node setup command configures the GCS services on the
local machine to run to serve data for an endpoint.
"""
from __future__ import annotations

import ipaddress
import os
import pwd
from typing import IO, Any, Optional, Tuple

import click

from gcs_cli.parsing.command_decorator import globus_command

from .utils import options
from .utils.deployment_key import DeploymentKey
from .utils.ip_address_types import (
    IPAddressParamType,
    IPv4AddressParamType,
    IPv6AddressParamType,
)


def verify_log_dirs(root: str) -> None:
    """
    Verify log directory permissions
    """
    log_base = os.path.join(root, "var/log/globus-connect-server")
    if not os.path.exists(log_base):
        os.makedirs(log_base, mode=0o0755)
    gcs_logdir = os.path.join(log_base, "gcs-manager")
    if not os.path.exists(gcs_logdir):
        os.mkdir(gcs_logdir, mode=0o0700)

    pwent = pwd.getpwnam("gcsweb")
    os.chmod(gcs_logdir, 0o0700)
    os.chown(gcs_logdir, pwent.pw_uid, pwent.pw_gid)

    # Handle prior `node cleanup`
    setup_log = os.path.join(gcs_logdir, "setup.log")
    if os.path.exists(setup_log):
        os.chown(setup_log, pwent.pw_uid, pwent.pw_gid)


def validate_port_range(
    ctx: click.Context, param: Any, value: Tuple[int, int]
) -> Tuple[int, int]:
    if value and len(value) == 2:
        min, max = value
        if min > max:
            raise click.BadParameter("LOW_PORT must be <= HIGH_PORT")
    return value


@globus_command(
    "setup",
    short_help="Set up a Globus Connect Server node",
    help="Set up a Globus Connect Server node",
    disable_options=["format", "explicit_host"],
)
@options.deployment_key_option(
    click_path=click.Path(
        dir_okay=False,
        exists=True,
        readable=True,
    ),
    required_fields=(
        ("node_key", options.MISSING_NODE_KEY_MSG),
        ("client_id", options.MISSING_CLIENT_ID_MSG),
        ("secret", options.MISSING_SECRET_MSG),
    ),
)
@options.deprecated_client_id_option
@options.deprecated_secret_option
@click.option(
    "--incoming-port-range",
    help="Allowed port range for incoming TCP data connections",
    metavar="LOW_PORT HIGH_PORT",
    type=click.IntRange(1, 2**16 - 1),
    nargs=2,
    default=(
        50000,
        51000,
    ),
    show_default=True,
    callback=validate_port_range,
)
@click.option(
    "--outgoing-port-range",
    help="Port range used as the source for outgoing TCP data connections",
    metavar="LOW_PORT HIGH_PORT",
    type=click.IntRange(1, 2**16 - 1),
    nargs=2,
    callback=validate_port_range,
)
@click.option(
    "--ip-address",
    "-i",
    "ip_addresses",
    multiple=True,
    type=IPAddressParamType(),
    help=(
        "IP address of the GCS Node. Use this option multiple times to set multiple IPs"
    ),
)
@click.option(
    "--data-interface",
    "data_interface",
    type=IPv4AddressParamType(),
    help="IP interface of the GCS Node used for Globus data transfers",
)
@click.option(
    "--data-interface6",
    "data_interface6",
    type=IPv6AddressParamType(),
    help="IPv6 interface of the GCS Node used for Globus data transfers",
    hidden=True,
)
@click.option(
    "--import-node",
    type=click.File(),
    metavar="NODE_INFO_JSON",
    help="File to read node configuration from (created by --export-node)",
)
@click.option(
    "--export-node",
    type=click.Path(dir_okay=False, writable=True),
    metavar="NODE_INFO_JSON",
    help="File to write node configuration to for restoring later",
)
@click.option(
    "--root",
    hidden=True,
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default="/",
)
@click.option("--verbose", "-v", is_flag=True)
@click.option(
    "--development-appsync",
    help="Use the specified URL and client_id for the appsync service",
    metavar="APPSYNC_URL APPSYNC_CLIENT_ID",
    hidden=True,
    nargs=2,
)
@click.pass_context
def node_setup(
    ctx: click.Context,
    deployment_key: DeploymentKey,
    incoming_port_range: Tuple[int, int],
    outgoing_port_range: Tuple[int, int],
    ip_addresses: tuple[ipaddress.IPv4Address | ipaddress.IPv6Address, ...],
    data_interface: ipaddress.IPv4Address | None,
    data_interface6: ipaddress.IPv6Address | None,
    import_node: Optional[IO],
    export_node: Optional[str],
    root: str,
    verbose: bool,
    development_appsync: Optional[Tuple[str, str]],
) -> None:
    from .node_setup import node_setup as ns

    assert deployment_key.node_key is not None
    assert deployment_key.client_id is not None
    assert deployment_key.secret is not None

    ns(
        ctx,
        str(deployment_key.client_id),
        deployment_key.secret,
        deployment_key.node_key,
        incoming_port_range,
        outgoing_port_range,
        ip_addresses,
        data_interface,
        data_interface6,
        import_node,
        export_node,
        root,
        verbose,
        development_appsync,
    )
