from __future__ import annotations

import json
import os
import typing as t
import uuid
from dataclasses import dataclass, fields
from pathlib import Path

from globus.manager.utils.keychain import DataKeyError, NodeKey

# Manage the handling of deployment keys.
#
# Version 1: Initial 5.4 release to 5.4.60
# - The deployment key file only contained the NodeKey and its fields were
#   directly exposed in the root of the serialized content.
#   IE. '{"d": ..., "dp", ...}'
#
# Version 2: 5.4.61 to 5.4.65
# - The deployment key included the NodeKey, client_id and secret. The NodeKey
#   fields were still directly exposed in the root of the serialized content.
#
# Version 3: 5.4.66+
# - The NodeKey contents were moved under the 'node_key' key in the serialized
#   content.


@dataclass
class DeploymentKey:
    """
    Contents of a GCSv5.4 deployment key. Any value may be NULL.
    """

    node_key: NodeKey | None = None
    client_id: uuid.UUID | str | None = None
    secret: str | None = None

    def serialize(self):
        """
        Returns a JSON-formatted string of the DeploymentKey contents. Always serialized
        to the latest version.
        """
        data = {
            "node_key": self.node_key.export(private_key=True, as_dict=True)
            if self.node_key is not None
            else None,
            "client_id": str(self.client_id) if self.client_id else None,
            "secret": self.secret,
        }
        return json.dumps(data)

    @classmethod
    def deserialize(cls, serialized_data: str):
        """
        Returns an instance of DeploymentKey initialized with the deserialized
        contents of the JSON-formatted string 'serialized_data'.
        """
        try:
            serialized_contents = json.loads(serialized_data)
        except json.JSONDecodeError as e:
            raise DataKeyError(f"The deployment key format is not correct ({e})")

        # Update older versions by moving NodeKey fields into the 'node_key' subkey.
        if "node_key" not in serialized_contents:
            for key in list(serialized_contents.keys()):
                if key not in {f.name for f in fields(cls)}:
                    node_key_contents = serialized_contents.get("node_key", {})
                    node_key_contents.update({key: serialized_contents.pop(key)})
                    serialized_contents["node_key"] = node_key_contents

        node_key: t.Optional[NodeKey] = None
        if serialized_contents.get("node_key"):
            try:
                node_key = NodeKey(**serialized_contents.get("node_key"))
            except Exception as e:
                raise DataKeyError(
                    f"Invalid deployment key. 'node_key' can not be deserialized: {e}"
                )

        client_id: t.Optional[uuid.UUID] = None
        client_id_str = serialized_contents.get("client_id")
        if client_id_str:
            try:
                client_id = uuid.UUID(client_id_str)
            except Exception:
                raise DataKeyError(
                    "Invalid deployment key. "
                    f"The client_id value '{client_id_str}' is not a valid UUID."
                )

        return cls(
            node_key=node_key,
            client_id=client_id,
            secret=serialized_contents.get("secret"),
        )


def load_deployment_key(
    path: Path | str,
) -> DeploymentKey:
    """
    Load the contents, if any, of 'path' and return a deserialized DeploymentKey.
    """
    with open(path, "r") as f:
        return DeploymentKey.deserialize(f.read())


def store_deployment_key(
    path: Path | str,
    deployment_key: DeploymentKey,
    umask: int = 0o77,
) -> None:
    """
    Store the serialized format of the deployment key to path. Make sure the umask
    is correctly set.
    """
    old_umask = os.umask(umask)

    try:
        with open(path, "w") as f:
            f.write(deployment_key.serialize())
    finally:
        os.umask(old_umask)
