from __future__ import annotations

import collections.abc as abc
import functools
import textwrap
import typing as t
from dataclasses import dataclass
from pathlib import Path

import click

from .deployment_key import DeploymentKey, load_deployment_key

T = t.TypeVar("T")


def _add_option_arg(
    arg_name: str,
    value: abc.Callable[[], t.Any],
) -> abc.Callable[[abc.Callable[..., T]], abc.Callable[..., T]]:
    """
    Returns a decorator that will add the field arg_name to kwargs before
    calling the wrapped function.
    """

    def _decorator(wrapped_func: abc.Callable[..., T]) -> abc.Callable[..., T]:
        @functools.wraps(wrapped_func)
        def _wrapper(*args, **kwargs) -> T:
            return wrapped_func(*args, **{**kwargs, arg_name: value()})

        return _wrapper

    return _decorator


def _remove_option_arg(
    arg_name: str,
) -> abc.Callable[[abc.Callable[..., T]], abc.Callable[..., T]]:
    """
    Returns a decorator that will remove the field arg_name from kwargs before
    calling the wrapped function.
    """

    def _decorator(wrapped_func: abc.Callable[..., T]) -> abc.Callable[..., T]:
        @functools.wraps(wrapped_func)
        def _wrapper(*args, **kwargs) -> T:
            kwargs.pop(arg_name)
            return wrapped_func(*args, **kwargs)

        return _wrapper

    return _decorator


class DeploymentKeyOption(click.ParamType):
    """
    Option type that converts a path to an instance of DeploymentKey and stores
    the path value.
    """

    # The meta used in usage messages.
    name = "deployment_key_path"

    _click_path: click.Path
    _path: str
    _required_fields: abc.Iterable[RequiredField]

    @dataclass
    class RequiredField:
        """
        Specify the name of a required field and the error message to display if the
        field it not available.
        """

        field_name: str
        error_msg: str

    def __init__(
        self,
        click_path: click.Path,
        required_fields: abc.Iterable[RequiredField] = (),
    ) -> None:
        self._click_path = click_path
        self._required_fields = required_fields

    def convert(
        self,
        value: str,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> DeploymentKey:
        # Let click.Path() do it's magic first
        self._path = self._click_path.convert(value, param, ctx)

        # If self._click_path.exists is False, the file many not exist. In that
        # case, just return an empty DeploymentKey.
        if not Path(self._path).exists():
            return DeploymentKey(node_key=None, client_id=None, secret=None)

        # Convert path->DeploymentKey
        try:
            dkey = load_deployment_key(Path(self._path))
        except Exception as e:
            self.fail(str(e), param=param, ctx=ctx)

        # Check for required fields
        rf: DeploymentKeyOption.RequiredField
        for rf in self._required_fields:
            if not getattr(dkey, rf.field_name):
                self.fail(rf.error_msg, param=param, ctx=ctx)

        return dkey

    @property
    def path(self) -> str:
        return self._path


MISSING_NODE_KEY_MSG = "\n".join(
    textwrap.wrap(
        "Your deployment key does not contain a valid node key. "
        "This command can not complete without it. "
        "If you are unable to locate the original deployment key, "
        "contact support@globus.org for assistance."
    )
)

MISSING_CLIENT_ID_MSG = "\n".join(
    textwrap.wrap(
        "The client_id is missing from your deployment key. "
        "Use the command 'endpoint key convert' to update "
        "your deployment key and then rerun this command."
    )
)

MISSING_SECRET_MSG = "\n".join(
    textwrap.wrap(
        "The secret is missing from your deployment key. "
        "Use the command 'endpoint key convert' to update "
        "your deployment key and then rerun this command."
    )
)


def deployment_key_option(
    click_path: click.Path,
    path_arg_name: t.Optional[str] = None,
    required_fields: abc.Iterable[tuple[str, str]] = (),
) -> abc.Callable[[abc.Callable[..., T]], abc.Callable[..., T]]:
    """
    Add the deployment key option. This option expects a path value and converts it
    to an instance of DeploymentKey.

    If path_arg_name is not None, the wrapped function will receive the path in the
    function argument named for the string value of path_arg_name.

    required_fields lists the fields of DeploymentKey which are required to be
    present and the error message to display if the field is missing or None.
    """

    def _decorator(wrapped_func: abc.Callable[..., T]) -> abc.Callable[..., T]:
        dk_opt = DeploymentKeyOption(
            click_path,
            [DeploymentKeyOption.RequiredField(*t) for t in required_fields],
        )

        if path_arg_name:
            wrapped_func = _add_option_arg(path_arg_name, lambda: dk_opt.path)(
                wrapped_func
            )

        return click.option(
            "--deployment-key",
            "-d",
            "deployment_key",
            default="deployment-key.json",
            help="Path for deployment key configuration",
            type=dk_opt,
        )(wrapped_func)

    return _decorator


def _deprecated_option_callback(
    error_msg: str,
) -> abc.Callable[[click.Context, click.Option | click.Parameter, t.Any], None]:
    """
    Returns a decorator to be be used as a click.Option callback which will
    raise click.UsageError(error_msg) if the option has a non-None value.
    """

    def _deprecated_option_callback(
        ctx: click.Context,
        param: click.Option | click.Parameter,
        value: t.Any,
    ) -> None:
        if value is not None:
            raise click.UsageError(error_msg, ctx)

    return _deprecated_option_callback


def deprecated_option(
    long_form: str,
    short_form: str,
    arg_name: str,
    help: str,
    error_msg: str,
    hidden: bool = False,
) -> abc.Callable[[abc.Callable[..., T]], abc.Callable[..., T]]:
    """
    Add an option to the command which will throw UsageError(error_msg) if a non-None
    value was supplied (ie. the option was used). The option's None value will not be
    passed to the wrapped function.
    """

    def _decorator(wrapped_func: abc.Callable[..., T]) -> abc.Callable[..., T]:
        return click.option(
            long_form,
            short_form,
            callback=_deprecated_option_callback(error_msg),
            help=help,
            hidden=hidden,
        )(_remove_option_arg(arg_name)(wrapped_func))

    return _decorator


def _help_msg(option_name: str) -> str:
    return (
        f"The {option_name} option has been removed, see the 'endpoint key convert' "
        "command to update your deployment key"
    )


def _error_msg(option_name: str) -> str:
    return _help_msg(option_name) + " and rerun this command without this option."


def deprecated_client_id_option(
    wrapped_func: abc.Callable,
) -> abc.Callable:
    return deprecated_option(
        "--client-id",
        "-c",
        "client_id",
        _help_msg("--client-id"),
        _error_msg("--client-id"),
    )(wrapped_func)


def deprecated_secret_option(
    wrapped_func: abc.Callable,
) -> abc.Callable:
    return deprecated_option(
        "--secret",
        "-s",
        "secret",
        _help_msg("--secret"),
        _error_msg("--secret"),
        hidden=True,
    )(wrapped_func)
