import logging
import time
from collections import namedtuple

import josepy
from acme import client, crypto_util, messages
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from OpenSSL import crypto

logger = logging.getLogger(__name__)

AcmeCertificate = namedtuple(
    "AcmeCertificate", ["domain_name", "certificate", "chain", "key"]
)


class AcmeClient:
    # TODO: more verbose user agent? full version? client id?
    USER_AGENT = "GlobusConnectServer"
    ACC_KEY_BITS = 2048
    CERT_PKEY_BITS = 2048
    WAIT_SEC = 120

    ACME_DIRECTORY_URL = "https://acme-v02.api.letsencrypt.org/directory"

    def __init__(self, dns_client, acme_key=None, acme_account=None):
        self.dns_client = dns_client
        self._acme_key = None
        self._acme_account = None
        self._acme_client = None

        if acme_key:
            self._acme_key = josepy.JWKRSA.from_json(acme_key)
        else:
            self._acme_key = josepy.JWKRSA(
                key=rsa.generate_private_key(
                    public_exponent=65537,
                    key_size=self.ACC_KEY_BITS,
                    backend=default_backend(),
                )
            )

        if not acme_account:
            self._acme_account = self.acme_client.new_account(
                messages.NewRegistration.from_data(terms_of_service_agreed=True)
            )
        else:
            self._acme_account = messages.RegistrationResource.from_json(acme_account)

    @property
    def acme_account(self):
        return self._acme_account.to_json()

    @property
    def acme_key(self):
        return self._acme_key.to_json()

    @property
    def acme_client(self):
        if not self._acme_client:
            net = client.ClientNetwork(
                self._acme_key, account=self._acme_account, user_agent=self.USER_AGENT
            )
            directory = messages.Directory.from_json(
                net.get(self.ACME_DIRECTORY_URL).json()
            )
            self._acme_client = client.ClientV2(directory, net=net)
        return self._acme_client

    def get_cert_chain_and_key(self, domain_name):
        """
        Looks up gcs_certificate, certificate_chain, and certificate_ke from
        config If gcs_certificate needs renewal or doesn't exist yet requests a
        new cert from letsencrypt
        """
        # create private key for the certificate
        # we use a new private key
        pkey = crypto.PKey()
        pkey.generate_key(crypto.TYPE_RSA, self.CERT_PKEY_BITS)
        key = crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey)

        # create Certificate Signing Request order
        csr = crypto_util.make_csr(key, [domain_name, "*." + domain_name])
        order = self.acme_client.new_order(csr)

        # get DNS challenges
        dns_challenges = []
        for authz in order.authorizations:
            for i in authz.body.challenges:
                # Find the supported challenge.
                if i.chall.typ == "dns-01":
                    dns_challenges.append(i)
        if len(dns_challenges) != 2:
            raise ValueError("Expected 2 DNS challenges")

        # get validation values to place in TXT records
        txt = []
        for chal in dns_challenges:
            txt.append(chal.validation(self.acme_client.net.key))

        # update TXT records
        logger.info("Creating TXT records with values: {}".format(txt))
        self.dns_client.create_txt(domain_name, txt)

        # wait for DNS propagation
        logger.info(
            "Waiting {} seconds for DNS changes " "to propagate".format(self.WAIT_SEC)
        )
        time.sleep(self.WAIT_SEC)

        # let letsencrypt know we are ready for validation
        for chal in dns_challenges:
            response = chal.response(self.acme_client.net.key)
            self.acme_client.answer_challenge(chal, response)

        # poll and finalize order to get fullchain
        logger.info("Polling and finalizing order")
        finalized_order = self.acme_client.poll_and_finalize(order)

        # split full chain into cert and chain
        fullchain_pem = finalized_order.fullchain_pem
        cert = crypto.dump_certificate(
            crypto.FILETYPE_PEM,
            crypto.load_certificate(crypto.FILETYPE_PEM, fullchain_pem),
        ).decode()
        chain = fullchain_pem[len(cert) :].lstrip()

        key = key.decode()

        return AcmeCertificate(domain_name, cert, chain, key)
