import fcntl
import os
from contextlib import contextmanager

from flask import current_app


class InterprocessLock:
    """
    Interprocess read-write locks using flock()
    """

    def __init__(self, name, lockdir=None):
        current_app.logger.debug("Creating lock %s", name)
        self.name = name
        if lockdir is None:
            lockdir = os.path.join(current_app.config["HOME"], "locks")
        if not os.path.exists(lockdir):
            os.makedirs(lockdir, mode=0o700, exist_ok=True)

        self.path = os.path.join(lockdir, name)
        self.shared_lock_count = 0
        self._fd = None

    @property
    def fd(self):
        if self._fd is None:
            self._fd = os.open(self.path, os.O_RDWR | os.O_CREAT, mode=0o700)

        return self._fd

    @contextmanager
    def shared(self):
        unlock = False
        try:
            fcntl.flock(self.fd, fcntl.LOCK_SH)
            self.shared_lock_count += 1
            unlock = True

            yield
        finally:
            if unlock:
                self.shared_lock_count -= 1
                if self.shared_lock_count == 0:
                    fcntl.flock(self.fd, fcntl.LOCK_UN)

    @contextmanager
    def exclusive(self):
        unlock = False
        try:
            fcntl.flock(self.fd, fcntl.LOCK_EX)
            unlock = True
            yield
        finally:
            if unlock:
                fcntl.flock(self.fd, fcntl.LOCK_UN)

    def __repr__(self):
        return "<InterprocessLock {}>".format(self.name)
