"""
Module to configure an apache server on SuSE based systems
"""
import os
from subprocess import Popen
from typing import List

__apache_sites_available_file = "/etc/apache2/conf.d/{name}.conf.gcs"
__apache_sites_enabled_file = "/etc/apache2/conf.d/{name}.conf"

__gcs_created_site = "/var/lib/globus-connect-server/created_site_{name}"
__gcs_enabled_site = "/var/lib/globus-connect-server/enabled_site_{name}"
__gcs_enabled_mod = "/var/lib/globus-connect-server/enabled_mod_{name}"
__gcs_enabled_conf = "/var/lib/globus-connect-server/enabled_conf_{name}"

service = "apache2"
required_modules = ["headers", "proxy", "proxy_http", "rewrite", "ssl"]
required_confs: List[str] = ["SSL"]


def site_add(name: str, site_config: str, force=False, enable=False):
    site_available = __apache_sites_available_file.format(name=name)
    created_site_path = __gcs_created_site.format(name=name)
    created_new_site = False

    if not os.path.exists(site_available):
        created_new_site = True

    with open(site_available, "w") as f:
        f.write(site_config)

    if force or created_new_site:
        open(created_site_path, "w").close()

    if enable:
        site_enable(name)


def site_remove(name, force=False):
    site_available = __apache_sites_available_file.format(name=name)
    created_site_path = __gcs_created_site.format(name=name)

    site_disable(name)

    if force or os.path.exists(created_site_path):
        try:
            os.remove(site_available)
        except FileNotFoundError:
            pass
        try:
            os.remove(created_site_path)
        except FileNotFoundError:
            pass


def __check_mod_is_enabled(module):
    check = Popen(
        [
            "/bin/sh",
            "-c",
            f""". /etc/sysconfig/apache2;
                for m in $APACHE_MODULES; do
                    if [ "$m" = {module} ]; then
                        exit 0;
                    fi;
                done;
                exit 1
            """,
        ],
        stdin=None,
    )
    check.communicate()
    return check.returncode == 0


def module_enable(module, force=False):
    if force or not __check_mod_is_enabled(module):
        gcs_enabled_mod = __gcs_enabled_mod.format(name=module)
        enabler = Popen(["/usr/sbin/a2enmod", module], stdin=None)
        enabler.communicate()
        if enabler.returncode == 0:
            open(gcs_enabled_mod, "w").close()


def module_disable(module):
    gcs_enabled_mod = __gcs_enabled_mod.format(name=module)
    if os.path.exists(gcs_enabled_mod):
        disabler = Popen(["/usr/sbin/a2dismod", module], stdin=None)
        disabler.communicate()
        if disabler.returncode == 0:
            try:
                os.remove(gcs_enabled_mod)
            except FileNotFoundError:
                pass


def site_enable(name, force=False):
    available_site_path = __apache_sites_available_file.format(name=name)
    enabled_site_path = __apache_sites_enabled_file.format(name=name)
    gcs_enabled_site_path = __gcs_enabled_site.format(name=name)
    enabled_site = False

    if not os.path.exists(enabled_site_path):
        enabled_site = True

    if force and os.path.exists(enabled_site_path):
        try:
            os.remove(enabled_site_path)
        except FileNotFoundError:
            pass

    if not os.path.lexists(enabled_site_path):
        os.symlink(available_site_path, enabled_site_path)

    if enabled_site:
        open(gcs_enabled_site_path, "w").close()


def site_disable(name):
    enabled_site_path = __apache_sites_enabled_file.format(name=name)
    gcs_enabled_site = __gcs_enabled_site.format(name=name)

    if os.path.exists(gcs_enabled_site):
        try:
            os.remove(enabled_site_path)
        except FileNotFoundError:
            pass
        try:
            os.remove(gcs_enabled_site)
        except FileNotFoundError:
            pass


def __check_conf_is_enabled(conf):
    check = Popen(
        [
            "/bin/sh",
            "-c",
            f""". /etc/sysconfig/apache2;
                for m in $APACHE_SERVER_FLAGS; do
                    if [ "$m" = {conf} ]; then
                        exit 0;
                    fi;
                done;
                exit 1
            """,
        ],
        stdin=None,
    )
    check.communicate()
    return check.returncode == 0


def conf_enable(name, force=False):
    if force or not __check_conf_is_enabled(name):
        gcs_enabled_conf = __gcs_enabled_conf.format(name=name)
        enabler = Popen(["/usr/sbin/a2enflag", name], stdin=None)
        enabler.communicate()
        if enabler.returncode == 0:
            open(gcs_enabled_conf, "w").close()


def conf_disable(name):
    gcs_enabled_conf = __gcs_enabled_conf.format(name=name)
    if os.path.exists(gcs_enabled_conf):
        disabler = Popen(["/usr/sbin/a2disflag", name], stdin=None)
        disabler.communicate()
        if disabler.returncode == 0:
            try:
                os.remove(gcs_enabled_conf)
            except FileNotFoundError:
                pass
