import time
from jose import jwt
import json
from base64 import urlsafe_b64encode, urlsafe_b64decode
from oic.utils.sdb import SessionDB, DictSessionBackend, DefaultToken, Token
from oic.utils.time_util import utc_time_sans_frac
from oic.utils.sdb import WrongTokenType, InvalidToken, Error


class CustomSessionDB(SessionDB):
    def __init__(self, domain, password, db, **kwargs):
        SessionDB.__init__(self, db=db, **kwargs)
        self._db = None  # To enforce we're overriding all SessionBackend access
        self._dict_db = db
        self.domain = domain
        self.password = password

    def create_authz_session(self, aevent, areq, id_token=None, oidreq=None, **kwargs):
        # TODO: Review all this
        """
        Create session holding info about the Authorization event.

        :param aevent: An AuthnEvent instance
        :param areq: The AuthorizationRequest instance
        :param id_token: An IDToken instance
        :param oidreq: An OpenIDRequest instance
        :return: The session identifier, which is the database key
        """

        # We need to generate the `sub` here, before `do_sub` is called, so it gets
        # embedded in the code.
        localname = aevent.uid
        sub = "{0}@{1}".format(localname, self.domain)
        _dic = {
            "oauth_state": "authz",
            "code_used": False,
            "authzreq": areq.to_json(),
            "client_id": areq["client_id"],
            "response_type": areq["response_type"],
            "revoked": False,
            "authn_event": aevent.to_json(),
            "sub": sub,
        }

        _dic.update(kwargs)

        try:
            _val = areq["nonce"]
            if _val:
                _dic["nonce"] = _val
        except (AttributeError, KeyError):
            pass

        for key in ["redirect_uri", "scope", "si_redirects"]:
            try:
                _dic[key] = areq[key]
            except KeyError:
                pass

        if id_token:
            _dic["id_token"] = id_token
        if oidreq:
            _dic["oidreq"] = oidreq.to_json()
        code = urlsafe_b64encode(
            jwt.encode(_dic, self.password, algorithm="HS512").encode("utf-8")
        ).decode("utf-8")
        _dic["code"] = code
        self._dict_db[code] = _dic
        return code

    def do_sub(self, sid, client_salt, sector_id="", subject_type="public"):
        """We created the `sub` already, in `create_authz_session`, so just return it."""
        return self._dict_db[sid]["sub"]

    def upgrade_to_token(
        self,
        token=None,
        issue_refresh=False,
        id_token="",
        oidreq=None,
        key=None,
        access_grant="",
    ):
        """
        Promote session to token. Adapted from PyOIDC to support self-contained
        authorization codes.

        This is where non-reuse of authorization codes is assured in PyOIDC,
        which cannot be done with this approach.

        :param token: The access grant (authorization code)
        :param issue_refresh: If a refresh token should be issued. Must be `False`.
        :param id_token: An IDToken instance
        :param oidreq: An OpenIDRequest instance
        :param key: The session key. One of token or key must be given.
        :return: The session information as a dictionary
        """
        if issue_refresh:
            raise NotImplementedError("Not supported")
        if token:
            # Session key == authorization code in this approach
            key = token

        dic = self._dict_db[key]
        _at = self.access_token(sid=key, sinfo=dic)

        # Confirm that the code hasn't expired
        valid_until = json.loads(dic["authn_event"])["valid_until"]
        if time.time() > valid_until:
            # Unfortunately there does not seem to be a better way to propagate this
            # error.
            raise Exception("Access code expired at {}".format(valid_until))

        dic["access_token"] = _at
        dic["access_token_scope"] = "?"
        dic["oauth_state"] = "token"
        dic["token_type"] = "Bearer"

        if id_token:
            dic["id_token"] = id_token
        if oidreq:
            dic["oidreq"] = oidreq

        # Reintroduce the large parameters, which exist in areq
        areq = json.loads(dic["authzreq"])
        if "state" in areq:
            dic["state"] = areq["state"]
        return dic

    def refresh_token(self, rtoken, client_id):
        raise NotImplementedError()  # Not supported

    def update(self, key, attribute, value):
        self._dict_db[key][attribute] = value

    def __getitem__(self, item):
        """
        Return Session item.

        :param item: authz grant code or refresh token
        """
        try:
            return self._dict_db[item]
        except KeyError:
            # We're in the callback, and need to get the authentication information
            # from the `code`
            authn_info = jwt.decode(
                urlsafe_b64decode(item), self.password, algorithms="HS512"
            )
            self._dict_db[item] = authn_info
            return authn_info


class AuthorizationCode(Token):
    # TODO: Review all this
    def __init__(self, secret, lifetime):
        self.type = "A"  # Only intended for authorization codes
        self.lifetime = lifetime
        self.secret = secret

    def __call__(self, sid, *args, **kwargs):
        raise NotImplementedError()

    def key(self, **kwargs):
        """Return a key - the session id."""
        raise NotImplementedError()

    def type_and_key(self, token):
        """
        Return type of Token (A=Access code, T=Token, R=Refresh token) and the session id.

        :param token: A token
        :return: tuple of token type and session id
        """
        return self.type, token

    def get_key(self, token):
        """
        Return session id.

        :param token: A token
        :return: The session id
        """
        raise NotImplementedError()

    def get_type(self, token):
        """
        Return token type.

        :param token: A token
        :return: Type of Token (A=Access code, T=Token, R=Refresh token)
        """
        raise NotImplementedError()

    def expires_at(self, token):
        """
        Return the expiry timestamp of the token.

        :param token: A token
        :return: Timestamp of the token expiry in UTC
        """
        raise NotImplementedError()

    def is_expired(self, token, when=None):
        """Return if token is still valid."""
        if when is None:
            now = utc_time_sans_frac()
        else:
            now = when
        eat = self.expires_at(token)
        return bool(now > eat)

    def invalidate(self, token):
        """Intended for refresh tokens -- not implemented"""
        raise NotImplementedError()

    def valid(self, token):
        try:
            typ, key = self.type_and_key(token)
        except (Error, InvalidToken):
            raise WrongTokenType()
        if typ != self.type:
            raise WrongTokenType()
        if typ == "R":
            return not self.token_storage[key].get("revoked", False)
        else:
            return True


def make_session_db(
    password, authorization_code_lifetime, access_token_lifetime, domain
):
    """
    Create a CustomSessionDB session database.
    :param password: The shared password used by all instances.
    :param authorization_code_lifetime:
    :param access_token_lifetime:
    :param domain:
    :return:
    """
    secret = None  # Not used, but DefaultToken expects it
    base_url = "https://{}".format(domain)

    code_factory = AuthorizationCode(password, authorization_code_lifetime)
    token_factory = DefaultToken(
        secret, password, typ="T", lifetime=access_token_lifetime
    )
    session_db = CustomSessionDB(
        domain,
        password,
        base_url=base_url,
        db=DictSessionBackend(),
        code_factory=code_factory,
        token_factory=token_factory,
        refresh_token_factory=None,  # No need for refresh tokens, therefore not supported
    )
    return session_db
