import json
import logging
import os
import sys

from sqlalchemy.orm.exc import NoResultFound
from werkzeug.middleware.proxy_fix import ProxyFix

from globus.manager.api.app import make_app
from globus.manager.api.models import Configuration, Endpoint
from globus.manager.database import db
from globus.manager.utils.keychain import NodeKey

os.environ["MANAGER_CONFIG"] = "config/production.py"

application = make_app()
db.init_app(application)

# As a convenience, so we can use `flask.current_app`, load the
# Globus Auth client credentials and host endpoint id
# into the app's runtime config.
with application.app_context():
    try:
        globus_client_id = Configuration.query.filter_by(name="client_id").one().value
        globus_client_secret = (
            Configuration.query.filter_by(name="client_secret").one().value
        )
        host_endpoint_id = (
            Configuration.query.filter_by(name="host_endpoint_id").one().value
        )
        globus_environment = (
            Configuration.query.filter_by(name="environment").one().value.lower()
        )
        globus_node_key = NodeKey(
            **json.loads(Configuration.query.filter_by(name="node_key").one().value)
        )
        globus_node_id = Configuration.query.filter_by(name="node_id").one().value
        tutorial_endpoint = Configuration.query.filter_by(
            name="tutorial_endpoint"
        ).one_or_none()
    except NoResultFound:
        msg = (
            "Globus Connect Server is not configured. Run "
            '"globus-connect-server node setup" to enable the services on this node.'
        )
        print(msg, file=sys.stderr)
        application.logger.error(msg)
        sys.exit(1)

    if tutorial_endpoint and tutorial_endpoint.value == "1":
        application.config["TUTORIAL_ENDPOINT"] = True
    else:
        application.config["TUTORIAL_ENDPOINT"] = False

    appsync_url = Configuration.query.filter_by(name="gcs_appsync_url").one_or_none()
    appsync_client_id = Configuration.query.filter_by(
        name="gcs_appsync_client_id"
    ).one_or_none()

    application.config["CLIENT_ID"] = globus_client_id
    application.config["CLIENT_SECRET"] = globus_client_secret
    application.config["HOST_ENDPOINT_ID"] = host_endpoint_id
    application.config["ENVIRONMENT"] = globus_environment
    application.config["NODE_KEY"] = globus_node_key
    application.config["NODE_ID"] = globus_node_id
    if appsync_url:
        application.config["GCS_APPSYNC_URL"] = appsync_url.value
    if appsync_client_id:
        application.config["GCS_APPSYNC_CLIENT_ID"] = appsync_client_id.value

    items = application.appsync_client.query_all(id="KEY_CHAIN")
    for item in items:
        application.appsync_client.process_item(item)
    ep = Endpoint.query.one_or_none()
    if ep:
        subscription_id = getattr(ep, "subscription_id", None)
        application.config["SUBSCRIPTION_ID"] = subscription_id

    application.config["SERVER_NAME"] = ep.globus_dns_domain.domain_name

    if application.debug:
        application.logger.setLevel(logging.DEBUG)
    else:
        application.logger.setLevel(
            os.environ.get("GCS_MANAGER_LOG_LEVEL", logging.INFO)
        )
application = ProxyFix(application, x_proto=1, x_host=1)
